# 0 "malloc.c"
# 0 "<built-in>"
# 0 "<command-line>"
# 1 "malloc.c"

# 1 "nmalloc.inc" 1
/* --------- nmalloc.c ----------- */

/* Copyright (c) 2003, 2007 by Charles B. Falconer
   Licensed under the terms of the GNU LIBRARY GENERAL PUBLIC
   LICENSE and/or the terms of COPYING.DJ, all available at
   <http://www.delorie.com>.

   Bug reports to <mailto:djgpp@delorie.com>

   Revised 2007-01-04 to include calloc.
   Revised 2007-01-28 per bug report by Florian Xaver. *f1*
   Revised 2014, 2015 for use in DJGPP libc by Andris Pavenis <andris.pavenis@iki.fi>
*/

/* A re-implementation of malloc and friends for DJGPP 2.03/2.04
   This includes many bits modeled after DJs original scheme.
   This is NOT portable - it builds in knowledge of int size etc.
   i.e. unsigned ints and pointers are both 32 bits (size 4)

   The system is NOT thread and interrupt safe, although use of a
   suitable critical section call could make it such.  Nothing
   herein executes for any unusual length of time (with NDEBUG).
*/

/* Some critical tuning constants. Search for them:
   MINSBRK    controls minimal access to sbrk
   ALIGN      controls alignment of pointers
   SAVEMEMORY reduces overhead at expense of checkability
   INT_MAX    (system) controls maximum allocation quantum.
*/

/* To avoid unexpected problems, the default has been changed
   so we now require NEWMALLDBG to enable the original action
*/

# 1 "../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 17 "../../../../include/stdint.h"
typedef signed char int_least8_t;
typedef unsigned char uint_least8_t;
typedef signed char int_fast8_t;
typedef unsigned char uint_fast8_t;
typedef signed char int8_t;
typedef unsigned char uint8_t;

typedef signed short int int_least16_t;
typedef unsigned short int uint_least16_t;
typedef signed int int_fast16_t;
typedef unsigned int uint_fast16_t;
typedef signed short int int16_t;
typedef unsigned short int uint16_t;

typedef signed int int_least32_t;
typedef unsigned int uint_least32_t;
typedef signed int int_fast32_t;
typedef unsigned int uint_fast32_t;
typedef signed int int32_t;
typedef unsigned int uint32_t;

__extension__ typedef signed long long int int_least64_t;
__extension__ typedef unsigned long long int uint_least64_t;
__extension__ typedef signed long long int int_fast64_t;
__extension__ typedef unsigned long long int uint_fast64_t;
__extension__ typedef signed long long int int64_t;
__extension__ typedef unsigned long long int uint64_t;

typedef long int intptr_t;
typedef unsigned long int uintptr_t;

__extension__ typedef signed long long int intmax_t;
__extension__ typedef unsigned long long int uintmax_t;

/* ANSI/ISO C99 says these should not be visible in C++ unless
   explicitly requested.  */

//#if !defined(__cplusplus) || defined(__STDC_LIMIT_MACROS)
# 110 "../../../../include/stdint.h"
  /* These are defined by limits.h, so make them conditional.  */
# 124 "../../../../include/stdint.h"
//#endif /* !__cplusplus || __STDC_LIMIT_MACROS */
# 37 "nmalloc.inc" 2
# 1 "../../../../include/libc/djctx.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */




# 1 "../../../../include/assert.h" 1
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
# 28 "../../../../include/assert.h"
void __dj_assert(const char *,const char *,int,const char *);
# 23 "../../../../include/libc/djctx.h" 2

void djregister_ctx_hooks(void (*)(int), void (*)(void), void (*)(void),
    void (*)(int));
# 38 "nmalloc.inc" 2
# 1 "nmalcdef.h" 1
/* Copyright (c) 2003, 2007 by Charles B. Falconer
   Licensed under the terms of the GNU LIBRARY GENERAL PUBLIC
   LICENSE and/or the terms of COPYING.DJ, all available at
   <http://www.delorie.com>.

   Bug reports to <mailto:djgpp@delorie.com>

   Revised 2007-01-04 to include calloc.
   Revised 2007-01-28 per bug report by Florian Xaver. *f1*
   Revised 2014, 2015 for use in DJGPP libc by Andris Pavenis <andris.pavenis@iki.fi>

   This file contains definitions common for nmalloc.c and nmalign.c
*/
# 36 "nmalcdef.h"
   /* the HOOKABLE variant is only for development   */
   /* It allows some other package to define malloc, */
   /* free, realloc, and to call this package.       */
# 59 "nmalcdef.h"
typedef unsigned char byte;
typedef unsigned int ulong;

# 1 "../../../../include/stddef.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 14 "../../../../include/stddef.h"
# 1 "../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 23 "../../../../include/sys/djtypes.h"
/* For GCC 3.00 or later we use its builtin va_list.                       */
# 34 "../../../../include/sys/djtypes.h"
/* but remains a typedef in C */





/* Define internal type to be used for avoiding strict
   aliasing warnings */

typedef short __attribute__((__may_alias__)) __dj_short_a;
typedef int __attribute__((__may_alias__)) __dj_int_a;
typedef long __attribute__((__may_alias__)) __dj_long_a;
typedef long long __attribute__((__may_alias__)) __dj_long_long_a;
typedef unsigned short __attribute__((__may_alias__)) __dj_unsigned_short_a;
typedef unsigned int __attribute__((__may_alias__)) __dj_unsigned_int_a;
typedef unsigned long __attribute__((__may_alias__)) __dj_unsigned_long_a;
typedef unsigned long long __attribute__((__may_alias__)) __dj_unsigned_long_long_a;
typedef float __attribute__((__may_alias__)) __dj_float_a;
typedef double __attribute__((__may_alias__)) __dj_double_a;
typedef long double __attribute__((__may_alias__)) __dj_long_double_a;
# 15 "../../../../include/stddef.h" 2

/* Some programs think they know better... */
# 26 "../../../../include/stddef.h"
//#ifdef __cplusplus
//#define offsetof(s_type, mbr) ((std::size_t) &((s_type *)0)->mbr)
//#else

//#endif


typedef int ptrdiff_t;




typedef long unsigned int size_t;




typedef unsigned short wchar_t;
# 63 "nmalcdef.h" 2
# 1 "../../../../include/stdlib.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 20 "../../../../include/stdlib.h"
# 1 "../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../include/sys/types.h"
# 1 "../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../include/sys/types.h" 2

typedef int blkcnt_t;
typedef int blksize_t;
typedef int dev_t;
typedef unsigned long fsblkcnt_t;
typedef unsigned long fsfilcnt_t;
typedef int ino_t;
typedef int mode_t;
typedef int nlink_t;


typedef int gid_t;



typedef int off_t;



__extension__ typedef long long off64_t;



__extension__ typedef long long offset_t;



typedef int pid_t;







typedef long signed int ssize_t;




typedef int uid_t;




/* Allow including program to override.  */




typedef struct fd_set {
  unsigned char fd_bits [((256) + 7) / 8];
} fd_set;







typedef unsigned int time_t;
# 21 "../../../../include/stdlib.h" 2
# 1 "../../../../include/stddef.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 22 "../../../../include/stdlib.h" 2

/* Some programs think they know better... */
# 38 "../../../../include/stdlib.h"
extern int __dj_mb_cur_max;

typedef struct {
  int quot;
  int rem;
} div_t;

typedef struct {
  long quot;
  long rem;
} ldiv_t;





void abort(void);
int abs(int _i);

int __dj_atexit(void (*_func)(void));
double atof(const char *_s);
int atoi(const char *_s);
long atol(const char *_s);
void * bsearch(const void *_key, const void *_base, size_t _nelem,
  size_t _size, int (*_cmp)(const void *_ck, const void *_ce));
void * calloc(size_t _nelem, size_t _size);
unsigned int calloc32(size_t _nelem, size_t _size);
div_t div(int _numer, int _denom);
void exit(int _status);
void free(void *_ptr);
void free32(unsigned int _addr);
char * getenv(const char *_name);
long labs(long _i);
ldiv_t ldiv(long _numer, long _denom);
void * malloc(size_t _size);
unsigned int malloc32(size_t _size);
int mblen(const char *_s, size_t _n);
size_t mbstowcs(wchar_t *_wcs, const char *_s, size_t _n);
int mbtowc(wchar_t *_pwc, const char *_s, size_t _n);
void qsort(void *_base, size_t _nelem, size_t _size,
       int (*_cmp)(const void *_e1, const void *_e2));
int rand(void);
void * realloc(void *_ptr, size_t _size);
unsigned int realloc32(unsigned int _addr, size_t _size);
void srand(unsigned _seed);
double strtod(const char *_s, char **_endptr);
long strtol(const char *_s, char **_endptr, int _base);
unsigned long strtoul(const char *_s, char **_endptr, int _base);
int system(const char *_s);
size_t wcstombs(char *_s, const wchar_t *_wcs, size_t _n);
int wctomb(char *_s, wchar_t _wchar);




typedef struct {
  long long int quot;
  long long int rem;
} lldiv_t;

void _Exit(int _status);
long long int atoll(const char *_s);
long long int llabs(long long int _i);
lldiv_t lldiv(long long int _numer, long long int _denom);
float strtof(const char *_s, char **_endptr);
long double strtold(const char *_s, char **_endptr);
long long int strtoll(const char *_s, char **_endptr, int _base);
unsigned long long int strtoull(const char *_s, char **_endptr, int _base);





long a64l(const char *_string);
char * l64a(long _value);
char * mktemp(char *_template);
char * mkdtemp(char *_template);
int mkstemp(char *_template);
int putenv(char *_val);
char * realpath(const char *_path, char *_resolved);
int setenv(const char *_var, const char *_val, int _overwrite);
int unsetenv(const char *_var);



void * alloca(size_t _size);
long double _atold(const char *_s);
void cfree(void *_ptr);
double drand48(void);
char * ecvtbuf(double _val, int _nd, int *_dp, int *_sn, char *_bf);
char * ecvt(double _val, int _nd, int *_dp, int *_sn);
double erand48(unsigned short _state[3]);
char * fcvtbuf(double _val, int _nd, int *_dp, int *_sn, char *_bf);
char * fcvt(double _val, int _nd, int *_dp, int *_sn);
char * gcvt(double _val, int _nd, char *_buf);
char * getpass(const char *_prompt);
int getlongpass(const char *_prompt, char *_buffer, int _max_len);
char * itoa(int _value, char *_buffer, int _radix);
long jrand48(unsigned short _state[3]);
void lcong48(unsigned short _param[7]);
unsigned long lrand48(void);
long mrand48(void);
unsigned long nrand48(unsigned short _state[3]);
unsigned short *seed48(unsigned short _state_seed[3]);
void srand48(long _seedval);
long double _strtold(const char *_s, char **_endptr);
void swab(const void *_from, void *_to, ssize_t _nbytes);
void * valloc (size_t _amt);





/* BSD Random Number Generator */
char * initstate (unsigned _seed, char *_arg_state, int _n);
char * setstate(char *_arg_state);
long random(void);
int srandom(int _seed);

/* These vary in expected prototype, so we just don't prototype them.
void		xfree(void *_ptr);
void *		xmalloc(size_t _size);
void *		xrealloc(void *ptr, size_t _size);
*/
# 173 "../../../../include/stdlib.h"
extern int __system_flags;
# 64 "nmalcdef.h" 2
# 1 "../../../../include/unistd.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 19 "../../../../include/unistd.h"
extern char **__dj_environ;
# 30 "../../../../include/unistd.h"
# 1 "../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 31 "../../../../include/unistd.h" 2
# 1 "../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 32 "../../../../include/unistd.h" 2





/* `lockf' is a simpler interface to the locking facilities of `fcntl'.
   LEN is always relative to the current file position.
   The CMD argument is one of the following.  */






/* Some programs think they know better... */
# 112 "../../../../include/unistd.h"
extern char *optarg;
extern int optind, opterr, optopt;

void __exit(int _status);
void _exit(int _status);
int access(const char *_path, int _amode);
unsigned int alarm(unsigned int _seconds);
int chdir(const char *_path);
int chown(const char *_path, uid_t _owner, gid_t _group);
int close(int _fildes);
size_t confstr(int _name, char *_buf, size_t _len);
char * ctermid(char *_s);
int dup(int _fildes);
int dup2(int _fildes, int _fildes2);
int execl(const char *_path, const char *_arg, ...);
int execle(const char *_path, const char *_arg, ...);
int execlp(const char *_file, const char *_arg, ...);
int execv(const char *_path, char *const _argv[]);
int execve(const char *_path, char *const _argv[], char *const _envp[]);
int execvp(const char *_file, char *const _argv[]);
int fchdir(int _fd);
pid_t fork(void);
long fpathconf(int _fildes, int _name);
char * getcwd(char *_buf, size_t _size);
gid_t getegid(void);
uid_t geteuid(void);
gid_t getgid(void);
int getgroups(int _gidsetsize, gid_t *_grouplist);
char * getlogin(void);
int getopt(int _argc, char *const _argv[], const char *_optstring);
pid_t getpgrp(void);
pid_t getpid(void);
pid_t getppid(void);
uid_t getuid(void);
int isatty(int _fildes);
int link(const char *_existing, const char *_new);
off_t lseek(int _fildes, off_t _offset, int _whence);
long pathconf(const char *_path, int _name);
int pause(void);
int pipe(int _fildes[2]);
ssize_t pwrite(int _fildes, const void *_buf, size_t _nbyte, off_t _offset);
ssize_t read(int _fildes, void *_buf, size_t _nbyte);
int rmdir(const char *_path);
int setgid(gid_t _gid);
int setpgid(pid_t _pid, pid_t _pgid);
pid_t setsid(void);
int setuid(uid_t _uid);
unsigned int sleep(unsigned int _seconds);
long sysconf(int _name);
pid_t tcgetpgrp(int _fildes);
int tcsetpgrp(int _fildes, pid_t _pgrp_id);
char * ttyname(int _fildes);
int unlink(const char *_path);
ssize_t write(int _fildes, const void *_buf, size_t _nbyte);



/* additional access() checks */


char * basename(const char *_fn);
int brk(void *_heaptop);
char * dirname(const char *_fn);
int __file_exists(const char *_fn);
int fchown(int fd, uid_t owner, gid_t group);
int fsync(int _fd);
int ftruncate(int, off_t);
int getdtablesize(void);
int gethostname(char *buf, int size);
int getpagesize(void);
char * getwd(char *__buffer);
int lchown(const char * file, int owner, int group);
int lockf(int _fildes, int _cmd, off_t _len);
int llockf(int _fildes, int _cmd, offset_t _len);
offset_t llseek(int _fildes, offset_t _offset, int _whence);
int nice(int _increment);
int readlink(const char * __file, char * __buffer, size_t __size);
void * djsbrk(int _delta);
int symlink (const char *, const char *);
int sync(void);
int truncate(const char*, off_t);
unsigned int usleep(unsigned int _useconds);

pid_t vfork(void);
# 65 "nmalcdef.h" 2
# 1 "../../../../include/signal.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 15 "../../../../include/signal.h"
# 1 "../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 16 "../../../../include/signal.h" 2

/* 256 software interrupts + 32 exceptions = 288 */
# 35 "../../../../include/signal.h"
typedef int sig_atomic_t;

int raise(int _sig);
void (*signal(int _sig, void (*_func)(int)))(int);
# 51 "../../../../include/signal.h"
/* SIGINT is ansi */
# 62 "../../../../include/signal.h"
typedef struct {
  unsigned long __bits[10]; /* max 320 signals */
} sigset_t;

struct sigaction {
  int sa_flags;
  void (*sa_handler)(int);
  sigset_t sa_mask;
};

int kill(pid_t _pid, int _sig);
int sigaction(int _sig, const struct sigaction *_act, struct sigaction *_oact);
int sigaddset(sigset_t *_set, int _signo);
int sigdelset(sigset_t *_set, int _signo);
int sigemptyset(sigset_t *_set);
int sigfillset(sigset_t *_set);
int sigismember(const sigset_t *_set, int _signo);
int sigpending(sigset_t *_set);
int sigprocmask(int _how, const sigset_t *_set, sigset_t *_oset);
int sigsuspend(const sigset_t *_set);
# 91 "../../../../include/signal.h"
void __djgpp_traceback_exit(int _sig);



extern char *sys_siglist[];

void psignal(int _sig, const char *_msg);
# 66 "nmalcdef.h" 2
# 1 "../../../../include/string.h" 1
/* Copyright (C) 2017 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 17 "../../../../include/string.h"
# 1 "../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 18 "../../../../include/string.h" 2

/* Some programs think they know better... */
# 34 "../../../../include/string.h"
void * memchr(const void *_s, int _c, size_t _n);
int memcmp(const void *_s1, const void *_s2, size_t _n);
void * memcpy(void * _dest, const void * _src, size_t _n);
void * memmove(void *_s1, const void *_s2, size_t _n);
void * memset(void *_s, int _c, size_t _n);
char * strcat(char * _s1, const char * _s2);
char * strchr(const char *_s, int _c);
int strcmp(const char *_s1, const char *_s2);
int strcoll(const char *_s1, const char *_s2);
char * strcpy(char * _s1, const char * _s2);
size_t strcspn(const char *_s1, const char *_s2);
char * strerror(int _errcode);
size_t strlen(const char *_s);
char * strncat(char * _s1, const char * _s2, size_t _n);
int strncmp(const char *_s1, const char *_s2, size_t _n);
char * strncpy(char * _s1, const char * _s2, size_t _n);
char * strpbrk(const char *_s1, const char *_s2);
char * strrchr(const char *_s, int _c);
size_t strspn(const char *_s1, const char *_s2);
char * strstr(const char *_s1, const char *_s2);
char * strtok(char * _s1, const char * _s2);
size_t strxfrm(char * _s1, const char * _s2, size_t _n);
# 64 "../../../../include/string.h"
int strerror_r(int _errnum, char *_strerrbuf, size_t _buflen);
char * strtok_r(char * _s1, const char * _s2, char ** _s3);



# 1 "../../../../include/sys/movedata.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../include/sys/movedata.h"
# 1 "../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../include/sys/movedata.h" 2
# 1 "../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../include/sys/movedata.h" 2

# 1 "../../../../include/sys/fmemcpy.h" 1



# 1 "../../../../include/dpmi.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 40 "../../../../include/sys/cdefs.h"
/* Support gcc's __attribute__ facility.  */




/* Ensure that always traditional GNU extern inline semantics are used
   (aka -fgnu89-inline) even if ISO C99 semantics have been specified.  */
# 9 "../../../../include/dpmi.h" 2
# 1 "../../../../include/sys/nearptr.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */




# 1 "../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 23 "../../../../include/libc/djthunks.h" 2

void crt1_startup(int handle);

uint8_t *djaddr2ptr(uint32_t addr);
uint8_t *djaddr2ptr2(uint32_t addr, uint32_t len);
uint32_t djptr2addr(const uint8_t *ptr);
uint32_t djthunk_get(const char *name);
uint32_t djthunk_get_h(int handle, const char *name);
void *djsbrk(int increment);
# 20 "../../../../include/libc/asmobj.h" 2
# 8 "../../../../include/sys/nearptr.h" 2
# 24 "../../../../include/sys/nearptr.h"
/* Functions to enable "near" pointer access to DOS memory under DPMI
   CW Sandmann 7-95  NO WARRANTY: WARNING, since these functions disable
   memory protection, they MAY DESTROY EVERYTHING ON YOUR COMPUTER! */

int __djgpp_nearptr_enable(void); /* Returns 0 if feature not avail */
void __djgpp_nearptr_disable(void); /* Enables protection */
extern int *____djgpp_selector_limit(void);/*
*/
# 30 "../../../../include/sys/nearptr.h"
       #define __djgpp_selector_limit (*____djgpp_selector_limit())/*
*/
# 30 "../../../../include/sys/nearptr.h"
                                       ; /* Limit on CS and on DS if prot */
extern unsigned *____djgpp_base_address(void);/*
*/
# 31 "../../../../include/sys/nearptr.h"
       #define __djgpp_base_address (*____djgpp_base_address())/*
*/
# 31 "../../../../include/sys/nearptr.h"
                                          ; /* Used in calculation below */
# 10 "../../../../include/dpmi.h" 2
# 1 "../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../include/libc/asmobj.h" 2
# 11 "../../../../include/dpmi.h" 2
# 1 "../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 12 "../../../../include/dpmi.h" 2
# 26 "../../../../include/dpmi.h"
//#ifndef _POSIX_SOURCE

extern unsigned short *____dpmi_error(void);/*
*/
# 28 "../../../../include/dpmi.h"
       #define __dpmi_error (*____dpmi_error())/*
*/
# 28 "../../../../include/dpmi.h"
                                        ;

typedef struct {
  unsigned short offset16;
  unsigned short segment;
} __dpmi_raddr;

typedef struct {
  unsigned int offset32;
  unsigned short selector;
} __dpmi_paddr;

typedef struct {
  unsigned int handle; /* 0, 2 */
  unsigned int size; /* or count */ /* 4, 6 */
  unsigned int address; /* 8, 10 */
} __dpmi_meminfo;

typedef union {
  struct {
    unsigned int edi;
    unsigned int esi;
    unsigned int ebp;
    unsigned int res;
    unsigned int ebx;
    unsigned int edx;
    unsigned int ecx;
    unsigned int eax;
  } d;
  struct {
    unsigned short di, di_hi;
    unsigned short si, si_hi;
    unsigned short bp, bp_hi;
    unsigned short res, res_hi;
    unsigned short bx, bx_hi;
    unsigned short dx, dx_hi;
    unsigned short cx, cx_hi;
    unsigned short ax, ax_hi;
    unsigned short flags;
    unsigned short es;
    unsigned short ds;
    unsigned short fs;
    unsigned short gs;
    unsigned short ip;
    unsigned short cs;
    unsigned short sp;
    unsigned short ss;
  } x;
  struct {
    unsigned char edi[4];
    unsigned char esi[4];
    unsigned char ebp[4];
    unsigned char res[4];
    unsigned char bl, bh, ebx_b2, ebx_b3;
    unsigned char dl, dh, edx_b2, edx_b3;
    unsigned char cl, ch, ecx_b2, ecx_b3;
    unsigned char al, ah, eax_b2, eax_b3;
  } h;
} __attribute__((packed)) __dpmi_regs;

typedef struct {
  unsigned char major;
  unsigned char minor;
  unsigned short flags;
  unsigned char cpu;
  unsigned char master_pic;
  unsigned char slave_pic;
} __dpmi_version_ret;

typedef struct {
  unsigned int largest_available_free_block_in_bytes;
  unsigned int maximum_unlocked_page_allocation_in_pages;
  unsigned int maximum_locked_page_allocation_in_pages;
  unsigned int linear_address_space_size_in_pages;
  unsigned int total_number_of_unlocked_pages;
  unsigned int total_number_of_free_pages;
  unsigned int total_number_of_physical_pages;
  unsigned int free_linear_address_space_in_pages;
  unsigned int size_of_paging_file_partition_in_pages;
  unsigned int reserved[3];
} __dpmi_free_mem_info;

typedef struct {
  unsigned int total_allocated_bytes_of_physical_memory_host;
  unsigned int total_allocated_bytes_of_virtual_memory_host;
  unsigned int total_available_bytes_of_virtual_memory_host;
  unsigned int total_allocated_bytes_of_virtual_memory_vcpu;
  unsigned int total_available_bytes_of_virtual_memory_vcpu;
  unsigned int total_allocated_bytes_of_virtual_memory_client;
  unsigned int total_available_bytes_of_virtual_memory_client;
  unsigned int total_locked_bytes_of_memory_client;
  unsigned int max_locked_bytes_of_memory_client;
  unsigned int highest_linear_address_available_to_client;
  unsigned int size_in_bytes_of_largest_free_memory_block;
  unsigned int size_of_minimum_allocation_unit_in_bytes;
  unsigned int size_of_allocation_alignment_unit_in_bytes;
  unsigned int reserved[19];
} __dpmi_memory_info;

typedef struct {
  unsigned int data16[2];
  unsigned int code16[2];
  unsigned short ip;
  unsigned short reserved;
  unsigned int data32[2];
  unsigned int code32[2];
  unsigned int eip;
} __dpmi_callback_info;

typedef struct {
  unsigned int size_requested;
  unsigned int size;
  unsigned int handle;
  unsigned int address;
  unsigned int name_offset;
  unsigned short name_selector;
  unsigned short reserved1;
  unsigned int reserved2;
} __dpmi_shminfo;

/* Unless otherwise noted, all functions return -1 on error, setting __dpmi_error to the DPMI error code */

void __dpmi_yield(void); /* INT 0x2F AX=1680 */

int __dpmi_allocate_ldt_descriptors(int _count); /* DPMI 0.9 AX=0000 */
int __dpmi_free_ldt_descriptor(int _descriptor); /* DPMI 0.9 AX=0001 */
int __dpmi_segment_to_descriptor(int _segment); /* DPMI 0.9 AX=0002 */
int __dpmi_get_selector_increment_value(void); /* DPMI 0.9 AX=0003 */
int __dpmi_get_segment_base_address(int _selector, unsigned int *_addr); /* DPMI 0.9 AX=0006 */
int __dpmi_set_segment_base_address(int _selector, unsigned int _address); /* DPMI 0.9 AX=0007 */
unsigned int __dpmi_get_segment_limit(int _selector); /* LSL instruction  */
int __dpmi_set_segment_limit(int _selector, unsigned int _limit); /* DPMI 0.9 AX=0008 */
int __dpmi_get_descriptor_access_rights(int _selector); /* LAR instruction  */
int __dpmi_set_descriptor_access_rights(int _selector, int _rights); /* DPMI 0.9 AX=0009 */
int __dpmi_create_alias_descriptor(int _selector); /* DPMI 0.9 AX=000a */
int __dpmi_get_descriptor(int _selector, unsigned char _buffer[8]); /* DPMI 0.9 AX=000b */
int __dpmi_set_descriptor(int _selector, unsigned char _buffer[8]); /* DPMI 0.9 AX=000c */
int __dpmi_allocate_specific_ldt_descriptor(int _selector); /* DPMI 0.9 AX=000d */


int __dpmi_get_multiple_descriptors(int _count, void *_buffer); /* DPMI 1.0 AX=000e */
int __dpmi_set_multiple_descriptors(int _count, void *_buffer); /* DPMI 1.0 AX=000f */

int __dpmi_allocate_dos_memory(int _paragraphs, int *_ret_selector_or_max); /* DPMI 0.9 AX=0100 */
int __dpmi_free_dos_memory(int _selector); /* DPMI 0.9 AX=0101 */
int __dpmi_resize_dos_memory(int _selector, int _newpara, int *_ret_max); /* DPMI 0.9 AX=0102 */

int __dpmi_get_real_mode_interrupt_vector(int _vector, __dpmi_raddr *_address); /* DPMI 0.9 AX=0200 */
int __dpmi_set_real_mode_interrupt_vector(int _vector, __dpmi_raddr *_address); /* DPMI 0.9 AX=0201 */
int __dpmi_get_processor_exception_handler_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0202 */
int __dpmi_set_processor_exception_handler_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0203 */
int __dpmi_get_protected_mode_interrupt_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0204 */
int __dpmi_set_protected_mode_interrupt_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0205 */

int __dpmi_get_extended_exception_handler_vector_pm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0210 */
int __dpmi_get_extended_exception_handler_vector_rm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0211 */
int __dpmi_set_extended_exception_handler_vector_pm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0212 */
int __dpmi_set_extended_exception_handler_vector_rm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0213 */

int __dpmi_simulate_real_mode_interrupt(int _vector, __dpmi_regs *_regs); /* DPMI 0.9 AX=0300 */
int __dpmi_int(int _vector, __dpmi_regs *_regs); /* like above, but sets ss sp fl */ /* DPMI 0.9 AX=0300 */
extern short __dpmi_int_ss, __dpmi_int_sp, __dpmi_int_flags; /* default to zero */
int __dpmi_simulate_real_mode_procedure_retf(__dpmi_regs *_regs); /* DPMI 0.9 AX=0301 */
int __dpmi_simulate_real_mode_procedure_retf_stack(__dpmi_regs *_regs, int stack_words_to_copy, const void *stack_data); /* DPMI 0.9 AX=0301 */
int __dpmi_simulate_real_mode_procedure_iret(__dpmi_regs *_regs); /* DPMI 0.9 AX=0302 */
int __dpmi_allocate_real_mode_callback(unsigned int _handler, unsigned int _regs, __dpmi_raddr *_ret); /* DPMI 0.9 AX=0303 */
int __dpmi_free_real_mode_callback(__dpmi_raddr *_addr); /* DPMI 0.9 AX=0304 */
int __dpmi_get_state_save_restore_addr( __dpmi_raddr *_rm, __dpmi_paddr *_pm); /* DPMI 0.9 AX=0305 */
int __dpmi_get_raw_mode_switch_addr( __dpmi_raddr *_rm, __dpmi_paddr *_pm); /* DPMI 0.9 AX=0306 */

int __dpmi_get_version( __dpmi_version_ret *_ret); /* DPMI 0.9 AX=0400 */

int __dpmi_get_capabilities( int *_flags, char vendor_info[128]); /* DPMI 1.0 AX=0401 */

int __dpmi_get_free_memory_information( __dpmi_free_mem_info *_info); /* DPMI 0.9 AX=0500 */
int __dpmi_allocate_memory(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0501 */
int __dpmi_free_memory(unsigned int _handle); /* DPMI 0.9 AX=0502 */
int __dpmi_resize_memory(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0503 */

int __dpmi_allocate_linear_memory(__dpmi_meminfo *_info, int _commit); /* DPMI 1.0 AX=0504 */
int __dpmi_resize_linear_memory(__dpmi_meminfo *_info, int _commit); /* DPMI 1.0 AX=0505 */
int __dpmi_get_page_attributes(__dpmi_meminfo *_info, short *_buffer); /* DPMI 1.0 AX=0506 */
int __dpmi_set_page_attributes(__dpmi_meminfo *_info, short *_buffer); /* DPMI 1.0 AX=0507 */
int __dpmi_map_device_in_memory_block(__dpmi_meminfo *_info, unsigned int _physaddr); /* DPMI 1.0 AX=0508 */
int __dpmi_map_conventional_memory_in_memory_block(__dpmi_meminfo *_info, unsigned int _linaddr); /* DPMI 1.0 AX=0509 */
int __dpmi_get_memory_block_size_and_base(__dpmi_meminfo *_info); /* DPMI 1.0 AX=050a */
int __dpmi_get_memory_information( __dpmi_memory_info *_buffer); /* DPMI 1.0 AX=050b */

int __dpmi_lock_linear_region(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0600 */
int __dpmi_unlock_linear_region(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0601 */
int __dpmi_mark_real_mode_region_as_pageable(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0602 */
int __dpmi_relock_real_mode_region(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0603 */
int __dpmi_get_page_size( unsigned int *_size); /* DPMI 0.9 AX=0604 */

int __dpmi_mark_page_as_demand_paging_candidate(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0702 */
int __dpmi_discard_page_contents(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0703 */

int __dpmi_physical_address_mapping(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0800 */
int __dpmi_free_physical_address_mapping(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0801 */

/* These next four functions return the old state */
int __dpmi_get_and_disable_virtual_interrupt_state(void); /* DPMI 0.9 AX=0900 */
int __dpmi_get_and_enable_virtual_interrupt_state(void); /* DPMI 0.9 AX=0901 */
int __dpmi_get_and_set_virtual_interrupt_state(int _old_state); /* DPMI 0.9 AH=09   */
int __dpmi_get_virtual_interrupt_state(void); /* DPMI 0.9 AX=0902 */

int __dpmi_get_vendor_specific_api_entry_point(const char *_id, __dpmi_paddr *_api); /* DPMI 0.9 AX=0a00 */

int __dpmi_set_debug_watchpoint(__dpmi_meminfo *_info, int _type); /* DPMI 0.9 AX=0b00 */
int __dpmi_clear_debug_watchpoint(unsigned int _handle); /* DPMI 0.9 AX=0b01 */
int __dpmi_get_state_of_debug_watchpoint(unsigned int _handle, int *_status); /* DPMI 0.9 AX=0b02 */
int __dpmi_reset_debug_watchpoint(unsigned int _handle); /* DPMI 0.9 AX=0b03 */

int __dpmi_install_resident_service_provider_callback(__dpmi_callback_info *_info); /* DPMI 1.0 AX=0c00 */
int __dpmi_terminate_and_stay_resident(int return_code, int paragraphs_to_keep); /* DPMI 1.0 AX=0c01 */

int __dpmi_allocate_shared_memory(__dpmi_shminfo *_info); /* DPMI 1.0 AX=0d00 */
int __dpmi_free_shared_memory(unsigned int _handle); /* DPMI 1.0 AX=0d01 */
int __dpmi_serialize_on_shared_memory(unsigned int _handle, int _flags); /* DPMI 1.0 AX=0d02 */
int __dpmi_free_serialization_on_shared_memory(unsigned int _handle, int _flags); /* DPMI 1.0 AX=0d03 */

int __dpmi_get_coprocessor_status(void); /* DPMI 1.0 AX=0e00 */
int __dpmi_set_coprocessor_emulation(int _flags); /* DPMI 1.0 AX=0e01 */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Backwards compatibility stuff						       */



typedef struct {
  unsigned int available_memory;
  unsigned int available_pages;
  unsigned int available_lockable_pages;
  unsigned int linear_space;
  unsigned int unlocked_pages;
  unsigned int available_physical_pages;
  unsigned int total_physical_pages;
  unsigned int free_linear_space;
  unsigned int max_pages_in_paging_file;
  unsigned int reserved[3];
} _go32_dpmi_meminfo;







typedef struct {
  unsigned int size;
  unsigned int pm_offset;
  unsigned short pm_selector;
  unsigned short rm_offset;
  unsigned short rm_segment;
} _go32_dpmi_seginfo;

/* returns zero if success, else dpmi error and info->size is max size */
int _go32_dpmi_allocate_dos_memory(_go32_dpmi_seginfo *info);
 /* set size to bytes/16, call, use rm_segment.  Do not
	   change anthing but size until the memory is freed.
	   If error, max size is returned in size as bytes/16. */
int _go32_dpmi_free_dos_memory(_go32_dpmi_seginfo *info);
 /* set new size to bytes/16, call.  If error, max size
	   is returned in size as bytes/16 */
int _go32_dpmi_resize_dos_memory(_go32_dpmi_seginfo *info);
 /* uses pm_selector to free memory */

/* These both use the rm_segment:rm_offset fields only */
int _go32_dpmi_get_real_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
int _go32_dpmi_set_real_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);

/* These do NOT wrap the function in pm_offset in an iret handler.
   You must provide an assembler interface yourself, or alloc one below.
   You may NOT LONGjmp out of an interrupt handler. */
int _go32_dpmi_get_protected_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
 /* puts vector in pm_selector:pm_offset. */
int _go32_dpmi_set_protected_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
 /* sets vector from pm_offset and pm_selector */

/********** HELPER FUNCTIONS **********/

int _go32_dpmi_chain_protected_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
 /* sets up wrapper that calls function in pm_offset, chaining to old
	   handler when it returns */

/* These generate assember IRET-style wrappers for functions and set up stack */
int _go32_dpmi_allocate_iret_wrapper(_go32_dpmi_seginfo *info);
 /* Put function ptr in pm_offset, call, returns wrapper entry in pm_offset. */
int _go32_dpmi_free_iret_wrapper(_go32_dpmi_seginfo *info);
 /* assumes pm_offset points to wrapper, frees it */

/* RMCB functions, automatically restructure the real-mode stack for the
   proper return type and set up correct PM stack.  The callback
   (info->pm_offset) is called as (*pmcb)(_go32_dpmi_registers *regs); */
int _go32_dpmi_allocate_real_mode_callback_retf(_go32_dpmi_seginfo *info, unsigned int regs);
 /* points callback at pm_offset, returns seg:ofs of callback addr
	   in rm_segment:rm_offset.  Do not change any fields until freed.
	   Interface is added to simulate far return */
int _go32_dpmi_allocate_real_mode_callback_iret(_go32_dpmi_seginfo *info, unsigned int regs);
 /* same, but simulates iret */
int _go32_dpmi_free_real_mode_callback(_go32_dpmi_seginfo *info);
 /* frees callback */

/* The following two variables may be used to change the default stack size
   for interrupts and rmcb wrappers to a user defined size from the default
   of 32Kbytes.  Each RMCB and chain/iret wrapper gets its own stack. */

extern unsigned int _go32_interrupt_stack_size;
extern unsigned int _go32_rmcb_stack_size;

/* Convenience functions, the return value is *bytes* */
unsigned int _go32_dpmi_remaining_physical_memory(void);
unsigned int _go32_dpmi_remaining_virtual_memory(void);

/* locks memory from a specified offset within the code/data selector */
int _go32_dpmi_lock_code( unsigned int _lockaddr, unsigned int _locksize);
int _go32_dpmi_lock_data( unsigned int _lockaddr, unsigned int _locksize);

int __djgpp_set_page_attributes(void *our_addr, unsigned int num_bytes,
           unsigned short attributes);
int __djgpp_map_physical_memory(void *our_addr, unsigned int num_bytes,
           uint64_t phys_addr);

//#endif /* !_POSIX_SOURCE */
# 5 "../../../../include/sys/fmemcpy.h" 2

void fmemcpy1(__dpmi_paddr dst, const void *src, unsigned len);
void fmemcpy2(void *dst, __dpmi_paddr src, unsigned len);
void fmemcpy12(__dpmi_paddr dst, __dpmi_paddr src, unsigned len);
# 28 "../../../../include/sys/movedata.h" 2
# 36 "../../../../include/sys/movedata.h"
/* This header is intended to be included only by other
   headers, like <go32.h> and <string.h>.  You may
   include this directly, but it will be non-portable. */

/* These lengths are in bytes, optimized for speed */
void dosmemget(unsigned long _offset, size_t _length, void *_buffer);
void dosmemput(const void *_buffer, size_t _length, unsigned long _offset);

/* The lengths here are in TRANSFERS, not bytes! */
void _dosmemgetb(unsigned long _offset, size_t _xfers, void *_buffer);
void _dosmemgetw(unsigned long _offset, size_t _xfers, void *_buffer);
void _dosmemgetl(unsigned long _offset, size_t _xfers, void *_buffer);
void _dosmemputb(const void *_buffer, size_t _xfers, unsigned long _offset);
void _dosmemputw(const void *_buffer, size_t _xfers, unsigned long _offset);
void _dosmemputl(const void *_buffer, size_t _xfers, unsigned long _offset);
# 70 "../../../../include/string.h" 2

int bcmp(const void *_ptr1, const void *_ptr2, size_t _length);
void bcopy(const void *_a, void *_b, size_t _len);
void bzero(void *ptr, size_t _len);
int ffs(int _mask);
char * index(const char *_string, int _c);
void * memccpy(void *_to, const void *_from, int _c, size_t _n);
int memicmp(const void *_s1, const void *_s2, size_t _n);
void * memmem(const void *l, size_t l_len, const void *s, size_t s_len);
char * rindex(const char *_string, int _c);
char * stpcpy(char *_dest, const char *_src);
char * stpncpy(char *_dest, const char *_src, size_t _n);
char * strdup(const char *_s);
char * strndup(const char *_s, size_t _n);
size_t strlcat(char *_dest, const char *_src, size_t _size);
size_t strlcpy(char *_dest, const char *_src, size_t _size);
char * strlwr(char *_s);
int strcasecmp(const char *_s1, const char *_s2);
int stricmp(const char *_s1, const char *_s2);
int strncasecmp(const char *_s1, const char *_s2, size_t _n);
int strnicmp(const char *_s1, const char *_s2, size_t _n);
size_t strnlen(const char *_s, size_t _n);
char * strsep(char **_stringp, const char *_delim);
char * strupr(char *_s);
# 67 "nmalcdef.h" 2
# 1 "../../../../include/limits.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 30 "../../../../include/limits.h"
  /* These are also defined by stdint.h.  */
# 74 "../../../../include/limits.h"
/* #define ARG_MAX			4096 -- depends on tb size; use sysconf */

/* #define OPEN_MAX		20 - DOS can change this */
/* #define STREAM_MAX		20 - DOS can change this */
# 91 "../../../../include/limits.h"
/* constants used in Solaris */



/* gnuc ones */
# 68 "nmalcdef.h" 2
# 1 "../../../../include/malloc.h" 1
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 22 "../../../../include/malloc.h"
# 1 "../../../../include/stdlib.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 23 "../../../../include/malloc.h" 2
void *memalign(size_t _align, size_t _amt);
# 69 "nmalcdef.h" 2
# 1 "../../../../include/libc/sysquery.h" 1
/* -------- sysquery.h ------------ */

/* Copyright (c) 2003 by Charles B. Falconer
   Licensed under the terms of the GNU LIBRARY GENERAL PUBLIC
   LICENSE and/or the terms of COPYING.DJ, all available at
   <http://www.delorie.com>.

   Bug reports to <mailto:djgpp@delorie.com>

   Revised 2015 for use in DJGPP libs by Andris Pavenis <andris.pavenis@iki.fi>
*/







/* The settable hooks, identifiers */
/* HKCOUNT is illegal value */
enum m_hook_kind {malloc_HK, malloc_fail_HK,
                  free_HK, free_null_HK,
                  realloc_HK, realloc_exit_HK,
                  HKCOUNT};

/* Depending on kind, some params may be meaningless */
typedef void (*M_HOOKFN)(size_t sz, void *bk);

/* returns previous value of the appropriate function */
typedef M_HOOKFN (*set_m_hook)(enum m_hook_kind which,
                               M_HOOKFN newhook);

/* This allows a clean connection to debugging software      */
/* NOTE: ANY value equivalent to -1 means data not available */
/*       for the unsigned chars this means 0xffh.            */
struct _sysquery {
   unsigned char data, gdlo, sz, prvf, nxtf, nxt, prv, ohead;
   void * nilp; /* dummy NULL, &freeptrs[1] */
   void * anchors; /* of memory chains */
   set_m_hook hookset; /* how to set a hook */
};

/* This can return the above values, hopefully in a register    */
/* NONE is used in nextfree, prevfree as the equivalent of NULL */
/* With the unclean knowledge that nil is actually a pointer to */
/* freehdrs[1], and that lastsbrk is actually freehdrs[0], the  */
/* entire malloc structure is open to debuggery.                */
struct _sysquery _sysmalloc(void);






/* -------- sysquery.h ------------ */
# 70 "nmalcdef.h" 2

/* system dependant magic.  Only STDIN, STDERR used */
enum {STDIN = 0, STDOUT, STDERR, STDAUX, STDPRN}; /* handles */

/* Incorporation into the system should require deleting the
   following <nmalloc.h>, changing all references to nmalloc
   to malloc, nfree to free, nrealloc to realloc.  Change the
   single call to fakesbrk to sbrk.  Also normally set all
   DEBUGx values above to 0 in place of 1.  Later many
   routines will be made inline.  For debugging compilations
   are done with "/Dinline= " switch.  For production use
   the "/D_NDEBUG=1" switch, which does all the above except
   the inlining. But see NEWMALLDBG above.
*/
# 92 "nmalcdef.h"
/* ============================================================
   Macros and storage for debugging, works before init on DJGPP
   WARNING - GNU extensions used here!!
   Note: many messages are designed for easy search with grep
         and also serve as comments.
*/
# 156 "nmalcdef.h"
/* This is intended to allow finding the header area from
   the address of the immediately adjacent memblocks.
   The guardxx avoid destruction by an off by one pointer
   and serve no real logical purpose.  Note that in some
   cases sz may be recovered from next or next may be
   recovered from sz.
*/
typedef struct memblock {
   struct memblock *prev; /* 1st, protect against overrun */
   struct memblock *next; /* makes this less clobberable  */
   ulong sz; /* of this memblock */
   /* An allocated block has the next two (1) fields NULL */
   /* A free block has them both non-NULL  */
   struct memblock *nextfree;
   struct memblock *prevfree; /* actually data w/SAVEMEMORY */



   /* here lies the actual assigned storage                 */
   /* so the following must be addressed by adding offset   */
   /* storage must always be a multiple of 8 in size        */
   /* thus these items are fictional, i.e. for zero data    */
} memblock, *memblockp;

/* Notice that with SAVEMEMORY the prevfree field only
   exists for free blocks; it reuses data space.  This
   is why we cannot allow 0 sized blocks.
*/






/* conversion and access macros */






/* We can never use an allocation smaller than this */


/* Alternate form of NULL to distinguish free lists
   This is self protection, because freehdrs[1] is otherwise
   unused.  freehdrs[0] is reserved to hold lastsbrk.  In turn
   this means that ALIGN must be >= 4.
*/


/* Magic constants.  MINSBRK must be MINSAVE or larger */
enum {NFLISTS = (int)(8 * sizeof(size_t)), MINSBRK = 1024};

static __inline ulong roundup(size_t sz)
{
   ulong size;

   size = ((sz + (16 -1)) & ~(16 -1)) + (((size_t) &((memblock *)0)->prevfree));
   return size;
} /* roundup */

//memblockp __nmalloc_split(memblockp *mp, ulong sz);
# 39 "nmalloc.inc" 2

/* ============== Globals ============= */

struct nm_state {
/* Headers of lists holding free blocks of 2**2 thru 2**31 size */
/* freehdr[n] holds items sized 2**n thru 2**(n+1) - 1          */
  memblockp freehdrs[NFLISTS]; /* yes, low indices are waste */


/* keep track of the bases of each new sbrk block */

  int lastsbrkbgn; /* zeroed on load */
  void *sbrkbgn[100 /* I have never seen more than 5 needed */]; /* NULLS on load */

/* This holds pointers to hooks, initialized to NULLs */
/* see enum m_hook_kind for actual identifiers in sysquery.h */
  M_HOOKFN hookptr[HKCOUNT];

};

static struct nm_state *nms;
static const struct nm_state ctx_nm_state_init = {}; static void _dummy_cinit(struct nm_state *dummy) {} static struct nm_state nm_state_contexts[100]; static void nm_state_init(int idx) { struct nm_state *x = &nm_state_contexts[idx]; ((void)((nms != x)||(__dj_assert("nms != x","nmalloc.inc",60,__func__),0))); *x = ctx_nm_state_init; _dummy_cinit(x); } static void nm_state_deinit(void) { nms = ((void*)0); } static void nm_state_save(void) { ; } static void nm_state_restore(int idx) { struct nm_state *x = &nm_state_contexts[idx]; if (nms == x) return; nms = x; ; } __attribute__((constructor)) static void static_nm_state_init(void) { djregister_ctx_hooks(nm_state_init, nm_state_deinit, nm_state_save, nm_state_restore); }
# 100 "nmalloc.inc"
/* 1------------------1 */
# 157 "nmalloc.inc"
/* 1------------------1 */
# 177 "nmalloc.inc"
/* 1------------------1 */

/* inserts bases of sbrk chains in sbrkbgn array  */
/* This ensures we can find all controlled memory */
/* gets called when we find an unexpected sbrk.   */
/* Note that if the sbrk was unaligned bk has now */
/* been aligned, and we have no record of wastage */
/* As long as nothing is returned to sbrk this is */
/* not a problem.  This only for the malldbg pkg. */
static void recordnewsbrk(memblockp bk)
{
   int i;

   if (nms->lastsbrkbgn < 100 /* I have never seen more than 5 needed */ - 1) {
      /* This check for a previous entry is probably not
         needed, but it is a rare occurance, so safety */
      for (i = 0; i < nms->lastsbrkbgn; i++) {
         if (bk == nms->sbrkbgn[i]) return;
      }
      nms->sbrkbgn[nms->lastsbrkbgn++] = bk;
   }
/* else we abandon trying to keep track */
} /* recordnewsbrk */

/* 1------------------1 */

static inline int size2bucket(ulong sz)
{
   int b;

   for (b = 0; sz; sz >>= 1, b++) continue;
   return b;
} /* size2bucket */

/* 1------------------1 */

static void badcallabort(const char *msg, int lgh, memblockp m)
{



   write(STDERR, msg, lgh);
   write(STDERR, ": memory fouled\n", 16);




   (void)m; /* anti unused warning */

   raise(288);
} /* badcallabort */

/* 1------------------1 */
# 238 "nmalloc.inc"
/* 1------------------1 */

/* Unlike rmvfromfree, this extracts a block that */
/* may be buried deep within the free list by     */
/* unlinking.  m is already known a free block    */
static void extractfree(memblockp m)
{
   int b;
   memblockp mnxtf, mprvf;

   if (m) {
      b = size2bucket(m->sz);
      ;

      /* ease further tests */
      if ((memblockp)&nms->freehdrs[1] == (mnxtf = m->nextfree)) m->nextfree = ((void*)0);
      if ((memblockp)&nms->freehdrs[1] == (mprvf = m->prevfree)) m->prevfree = ((void*)0);

      if (m->nextfree) m->nextfree->prevfree = mprvf;

      if (m->prevfree) m->prevfree->nextfree = mnxtf;
      else if (m->nextfree) nms->freehdrs[b] = mnxtf;
      else nms->freehdrs[b] = ((void*)0);

      /* mark the block non-free */
      m->nextfree = m->prevfree = ((void*)0);

      ;
      ;
   }
} /* extractfree */

/* 1------------------1 */

static inline memblockp combinelo(memblockp m)
{
   memblockp m1;

   m1 = m;
   if ((m->prev && (m->prev != (memblockp)&nms->freehdrs[1]) && m->prev->nextfree && m->prev->prevfree)) {
      if (m->prev->next != m) {
         badcallabort("combinelo", 9, m);
         exit(1); /* prevent user trapping SIGABRT */
      }
      m1 = m->prev;
      extractfree(m1);
      if (((void*)0) != (m1->next = m->next))
         m1->next->prev = m1;
      m1->sz += m->sz;
   }
   return m1;
} /* combinelo */

/* 1------------------1 */

/* used to combine with lastsbrk, so no ISFREE test */
/* because lastsbrk is not kept in the free lists   */
static memblockp combinehi(memblockp m)
{
   memblockp m1;

   if (m && m->next) {
      ;
      ;
      if (m->next->prev != m) {
         badcallabort("combinehi", 9, m);
         exit(1); /* prevent user trapping SIGABRT */
      }
      m1 = m->next;
      if (m1 != nms->freehdrs[0]) extractfree(m1);
      if (((void*)0) != (m->next = m->next->next))
         m->next->prev = m;
      m->sz += m1->sz;
      ;
   }
   return m;
} /* combinehi */

/* 1------------------1 */

/* This takes care of marking the block as free */
static void mv2freelist(memblockp m)
{
   int b;

   if (m) {
      if ((m->next && (m->next != (memblockp)&nms->freehdrs[1]) && m->next->nextfree && m->next->prevfree)) m = combinehi(m);
      b = size2bucket(m->sz);
      ; ;
      if (nms->freehdrs[0] && (m->next == nms->freehdrs[0])) {
         ;
         nms->freehdrs[0] = combinehi(m);
         nms->freehdrs[0]->nextfree = nms->freehdrs[0]->prevfree = ((void*)0);
         ;
         return;
      }
      else if (nms->freehdrs[b]) {
         m->nextfree = nms->freehdrs[b];
         nms->freehdrs[b]->prevfree = m;
      }
      else {
         m->nextfree = (memblockp)&nms->freehdrs[1];
      }
      m->prevfree = (memblockp)&nms->freehdrs[1];
      if (nms->freehdrs[b]) nms->freehdrs[b]->prevfree = m;
      nms->freehdrs[b] = m;
      ;
   }
} /* mv2freelist */

/* 1------------------1 */

/* this always marks the block as non-free */
static inline void rmvfromfree(memblockp m)
{
   int b;

   if (m) {
      b = size2bucket(m->sz);
      ; ;
      if (m != nms->freehdrs[b]) {
         ;
         badcallabort("rmvfromfree", 11, m);
         exit(1); /* prevent user trapping SIGABRT */
      }
      else {
         if ((memblockp)&nms->freehdrs[1] == m->nextfree)
            nms->freehdrs[b] = ((void*)0);
         else {
            nms->freehdrs[b] = m->nextfree;
            nms->freehdrs[b]->prevfree = (memblockp)&nms->freehdrs[1];
         }

         m->nextfree = ((void*)0);



         ;
         ;
      }
   }
} /* rmvfromfree */

/* 1------------------1 */

static int searchfree(ulong szneed)
{
   int b;

   b = size2bucket(szneed);
   ;

   if (! nms->freehdrs[b] || (nms->freehdrs[b]->sz < szneed)) {
      do {
         b++;
      } while ((b < NFLISTS) && ! nms->freehdrs[b]);
   }
   /* if found we will break off a piece and housekeep */
   if (b < NFLISTS)
      ;
   else {
      b = 0;
      ;
   }
   return b;
} /* searchfree */

/* 1------------------1 */

/* The higher portion is returned in *mp,     */
/* the lower portion via the function return. */
/* and the lower portion is marked non-free   */
static memblockp __nmalloc_split(memblockp *mp, ulong sz)
{
   memblockp m1, m;

   m = *mp;
   m1 = (memblockp)((char *)m + sz);
   if (m->sz < (sz + (((size_t) &((memblock *)0)->prevfree)))) {
      badcallabort("memblockpsz", 11, m);
      exit(1); /* prevent user trapping SIGABRT */
   }
   memcpy(m1, m, (((size_t) &((memblock *)0)->prevfree)));
   m1->prev = m;
   m1->sz = m->sz - sz;
   m->next = m1;
   m->sz = sz;
   m->nextfree = ((void*)0);




   *mp = m1;
   if (m1->next) {
      if (m1->next->prev != m) {
         badcallabort("memblockpnxt", 12, m1);
         exit(1); /* prevent user trapping SIGABRT */
      }
      m1->next->prev = m1;
   }
   ;
   return m;
} /* split */

/* 1------------------1 */

/* Get the memory, see if it extends the present lastsbrk
   If not, put the old lastsbrk into the appropriate freelist
      and replace lastsbrk by the new, setting the headers up
   else update the size markers in lastsbrk.  When done either
   lastsbrk can supply the memory szextra, or NULL is returned.
   A revised lastsbrk block is marked as non-free.
*/
static memblockp extendsbrk(ulong szxtra)
{
   memblockp m;
   byte *expected;
   int alignerr;
   int aligndelta;

   ;

   /* we have to ensure that the new lastsbrk always has    */
   /* room to both realign and to leave a header when split */
   szxtra += (2 * 16 + (((size_t) &((memblock *)0)->prevfree)));
   if (szxtra < MINSBRK) szxtra = MINSBRK;

   if (nms->freehdrs[0])
      expected = ((byte*)nms->freehdrs[0]) + nms->freehdrs[0]->sz;
   else expected = ((void*)0);

   if ((aligndelta = (uintptr_t)expected & (16 -1))) {
      /* lastsbrk end was misaligned, try to align end of this */
      szxtra += 16 - aligndelta;
      aligndelta = 0;
   }

   m = (memblockp)djsbrk(szxtra);
   if ((uintptr_t)-1 == (uintptr_t)m) return ((void*)0);
   else {
      if ((byte*)m == expected) { /* Extending size of lastsbrk */
        


                                                            ;
         nms->freehdrs[0]->sz += szxtra;
         m = nms->freehdrs[0];
      }
      else {
         /* Here we have to check & fix alignment */
        


                                                            ;
         if ((alignerr = ((16 -1) & (uintptr_t)m))) {
/*f1*/ m = (memblockp)((char*)m +
                            (aligndelta = 16 - alignerr));
            ;
         }
         m->sz = szxtra - aligndelta; /* discard alignerr bytes */
         m->prev = m->next = ((void*)0);

         m->nextfree = ((void*)0);




         mv2freelist(nms->freehdrs[0]);
         nms->freehdrs[0] = m;
         recordnewsbrk(m); /* save in list of chains */
      }
   }
   return m;
} /* extendsbrk */

/* 1------------------1 */

/* The mechanism:
   All available memory is kept on the free list, and all adjacent
   blocks, assigned or free, are linked by next/prev fields in order
   of address.  Freehdrs[n] holds the first of a list of free blocks
   of sizes between 2**n and 2**n+1. A pointer to the free portion
   of the block last acquired via sbrk is held in lastsbrk.

   All blocks on the freelist are marked by having a non-NULL value
   in the nextfree or prevfree fields.  The special value NONE is
   used to replace NULL to terminate these lists.  Because of the
   misalignment possibilities it is necessary to keep accurate byte
   count lengths in the sz component of lastsbrk.

   1.  An allocation is made from the first fit freehdrs list. Note
   that there MAY be a usable piece in the next lower freehdr, but
   that is ignored because we do not want to search possibly long
   lists.  The block is removed from the freelist, and any excess
   space is broken off (if large enough to be usable) and assigned
   to the appropriate free list.

   2.  If no suitable free block is found, allocation is attempted
   from the last block created by an sbrk call.  Such a block must
   be large enough to maintain an sbrk pointer after splitting off
   the desired allocation.

   3.  If this fails a new block is created (or extended) via an
   sbrk call.  If possible, the previous lastsbrk block is extended.
   If extension is not possible the remains of the old block alone
   is placed in the freelist.  This (non-extension) case results in
   the prev field of the lastsbrk block being NULL.  The next field
   of the lastsbrk block is always NULL. In this case only it is
   necessary to check and correct memory alignment.

   Insertion is always done into the start of any given freelist.
   Thus there is no list walking needed.  Similarly, any block is
   always removed from the head of the appropriate freelist.

   It is assumed that sbrk will never return a lower address than
   did a previous sbrk.  I am not sure if this affects anything. I
   believe it does not.
*/
static void *nmalloc(size_t size)
{
   memblockp m = ((void*)0), m1;
   ulong szneed;
   int b;
   void *p = ((void*)0);
   size_t sz = size; /* preserve arg for hooks */

   /* compute the rounded up size needed */
   if (!sz) sz++; /* avoid any 0 space allocation */
   szneed = roundup(sz);
   ;
   ;

   /* Check for oversize allocation request */
   if (szneed < ((ulong)(2147483647 - 65536))) {
      /* search the free lists for one */
      b = searchfree(szneed);

      if (b) {
         rmvfromfree(m1 = nms->freehdrs[b]);
         if (m1->sz < szneed + (16 + (((size_t) &((memblock *)0)->prevfree))))
            m = m1;
         else {
            m = __nmalloc_split(&m1, szneed);
            mv2freelist(m1);
         }
      }
      else if (nms->freehdrs[0] &&
              (nms->freehdrs[0]->sz >= (szneed + (((size_t) &((memblock *)0)->prevfree))))) {
         m = __nmalloc_split(&nms->freehdrs[0], szneed);
      }
      /* if not found get more from system */
      else if ((m1 = extendsbrk(szneed))) {
         if (m1->sz < szneed + (16 + (((size_t) &((memblock *)0)->prevfree)))) {
            m = m1;
            ;
         }
         else {
            m = __nmalloc_split(&nms->freehdrs[0], szneed);
         }
      }
   /* else abject_failure(); */
   /* abject_failure COULD check the first possible freehdrs */
   /* list as a last chance to find some suitable memory     */

      if (m) p = (void*)((byte*)(m) + (((size_t) &((memblock *)0)->prevfree)));
      else {
         ;
         p = ((void*)0);
      }
   }
/* else m and p are NULL for oversize; */







   if (nms->hookptr[malloc_HK]) nms->hookptr[malloc_HK](size, p);
   if (!p && nms->hookptr[malloc_fail_HK])
      nms->hookptr[malloc_fail_HK](size, ((void*)0));

   return p;
} /* nmalloc */

/* 1------------------1 */

static void dofree(memblockp m)
{
   /* mark the block free */
   m->nextfree = m->prevfree = (memblockp)&nms->freehdrs[1];

   /* try to combine with lower or higher blocks in memory */
   if ((m->next && (m->next != (memblockp)&nms->freehdrs[1]) && m->next->nextfree && m->next->prevfree)) m = combinehi(m);
   if ((m->prev && (m->prev != (memblockp)&nms->freehdrs[1]) && m->prev->nextfree && m->prev->prevfree)) m = combinelo(m);

   if (nms->freehdrs[0] && (nms->freehdrs[0] == m->prev) )
      ;
   else mv2freelist(m);
} /* dofree */

/* 1------------------1 */

static void nfree(void *ptr)
{
   memblockp m;

   if (nms->hookptr[free_HK]) nms->hookptr[free_HK](0, ptr);

   if (ptr) {
      m = (memblockp)((byte*)(ptr) - (((size_t) &((memblock *)0)->prevfree)));
      ; ;
      if ((m && (m != (memblockp)&nms->freehdrs[1]) && m->nextfree && m->prevfree) || /* bad, refreeing block */
          (!nms->freehdrs[0] || m->nextfree) ) { /* block is fouled */
         badcallabort("free", 4, m);
         return; /* he can trap this SIGABRT */
      }
      dofree(m);



   }
   else if (nms->hookptr[free_null_HK])
      nms->hookptr[free_null_HK](0, ((void*)0));
} /* nfree */

/* 1------------------1 */

static memblockp mv2lastsbrk(memblockp m, ulong szneed)
{
   memblockp m1;

   m1 = __nmalloc_split(&nms->freehdrs[0], szneed);

   /* Now m1 is the proposed new block, of the right size */
   /* links are already revised so copy data from m to m2 */
   memcpy((void*)((byte*)(m1) + (((size_t) &((memblock *)0)->prevfree))), (void*)((byte*)(m) + (((size_t) &((memblock *)0)->prevfree))), m->sz - (((size_t) &((memblock *)0)->prevfree)));

   dofree(m);
   return m1;
} /* mv2lastsbrk */

/* 1------------------1 */

static void *nrealloc(void *ptr, size_t size)
{
   memblockp m, m1, m2;
   void *p;
   ulong szneed;
   int b;
   size_t sz = size;

   if (nms->hookptr[realloc_HK]) nms->hookptr[realloc_HK](sz, ptr);

   if (!ptr) {
      p = nmalloc(sz);
      if (nms->hookptr[realloc_exit_HK])
         nms->hookptr[realloc_exit_HK](size, p);
      return p;
   }

   m = m1 = (memblockp)((byte*)(ptr) - (((size_t) &((memblock *)0)->prevfree)));
   if (!sz) sz++; /* avoid any 0 space allocation */
   szneed = roundup(sz);
   ;
   ;
   if ((m && (m != (memblockp)&nms->freehdrs[1]) && m->nextfree && m->prevfree) || /* bad, realloc of free block */
       (!nms->freehdrs[0] || m->nextfree) ) { /* storage fouled */
      badcallabort("realloc", 7, m);
      p = ((void*)0);
      goto exeunt; /* he can trap this SIGABRT */
   }
   ;

   /* if decreasing simply reduce size and move excess to free */
   if (szneed <= m->sz) {
      ;
      if ((m->sz - szneed) >= (16 + (((size_t) &((memblock *)0)->prevfree)))) {
         m = __nmalloc_split(&m1, szneed);
         mv2freelist(m1);
      }
   /* else just return old pointer, i.e. NOP */
   }
   else if (szneed > ((ulong)(2147483647 - 65536))) {
      /* reject excessive size request */
      p = ((void*)0); goto exeunt;
   }
   else if ((m->next && (m->next != (memblockp)&nms->freehdrs[1]) && m->next->nextfree && m->next->prevfree) &&
            (szneed <= (m->sz + m->next->sz)) ) {
      /* the 'next' block is free and adequate so use it */
      ;
      m = m1 = combinehi(m);
      /* now split off the excess, if any */
      if ((m->sz - szneed) >= (16 + (((size_t) &((memblock *)0)->prevfree)))) {
         m = __nmalloc_split(&m1, szneed);
         mv2freelist(m1);
      }
   /* else m is the oversized return block */
   }
   else if ((nms->freehdrs[0] == m->next) &&
            ((szneed + (16 + (((size_t) &((memblock *)0)->prevfree)))) <= (m->sz + nms->freehdrs[0]->sz)) ) {
      /* lastsbrk is adequate and adjacent so use it */
      ;
      m = m1 = combinehi(m1);
      m = __nmalloc_split(&m1, szneed);
      nms->freehdrs[0] = m1;
   }
   else if ((m->prev && (m->prev != (memblockp)&nms->freehdrs[1]) && m->prev->nextfree && m->prev->prevfree) &&
            (szneed <= (m->sz + m->prev->sz)) ) {
      /* the 'prev' block is free and adequate so use it */
      ;
      m1 = m->prev;
      extractfree(m1);
      m1->sz += m->sz; /* revise the links */
      if ((m1->next = m->next)) m1->next->prev = m1;
      /* we are now done with m links, except sz */

      /* This involves copying the data, overlapping */
      memmove((void*)((byte*)(m1) + (((size_t) &((memblock *)0)->prevfree))), (void*)((byte*)(m) + (((size_t) &((memblock *)0)->prevfree))), m->sz - (((size_t) &((memblock *)0)->prevfree)));

      m = m1; /* done with the old m value */
      /* Is there something leftover */
      if ((m->sz - szneed) >= (16 + (((size_t) &((memblock *)0)->prevfree)))) {
         m = __nmalloc_split(&m1, szneed);
         mv2freelist(m1);
      }
   }
   else if ((b = searchfree(szneed))) {
      /* An adequate free block exists, copy over, free old */
      ;
      rmvfromfree(m1 = nms->freehdrs[b]);
      if (m1->sz < szneed + (16 + (((size_t) &((memblock *)0)->prevfree)))) {
         m2 = m1;
      }
      else {
         m2 = __nmalloc_split(&m1, szneed);
         mv2freelist(m1);
      }
      /* Now m2 is the proposed new block, of the right size */
      /* links are already revised so copy data from m to m2 */
      memcpy((void*)((byte*)(m2) + (((size_t) &((memblock *)0)->prevfree))), (void*)((byte*)(m) + (((size_t) &((memblock *)0)->prevfree))), m->sz - (((size_t) &((memblock *)0)->prevfree)));

      dofree(m);
      m = m2;
   }
   else if (nms->freehdrs[0] &&
            ((szneed + (16 + (((size_t) &((memblock *)0)->prevfree)))) <= nms->freehdrs[0]->sz) ) {
      ;
      m = mv2lastsbrk(m, szneed);
   }
   /* else malloc new size, copy data, and free old */
   else if ((m1 = extendsbrk(szneed))) {
      if (nms->freehdrs[0] == m->next) {
         ;
         /* last chance to avoid copying */
         m = m1 = combinehi(m);
         m = __nmalloc_split(&m1, szneed);
         nms->freehdrs[0] = m1;
      }
      else {
         /* At this point lastsbrk is adequate size */
         /* split off, copy over, and free old      */
         ;
         m = mv2lastsbrk(m, szneed);
      }
   }
   else m = ((void*)0); /* failure */

   if (m) p = (void*)((byte*)(m) + (((size_t) &((memblock *)0)->prevfree)));
   else {
      ;
      p = ((void*)0);
   }







exeunt: /* label used on realloc of free block */
              /* and on trap of oversize request */
   if (!p && ptr && nms->hookptr[malloc_fail_HK])
      nms->hookptr[malloc_fail_HK](size, ptr);
   if (nms->hookptr[realloc_exit_HK])
      nms->hookptr[realloc_exit_HK](size, p);

   return p;
} /* nrealloc */

/* 1------------------1 */

/* calloc included here to ensure that it handles the
   same range of sizes (s * n) as does malloc.  The
   multiplication n*s can wrap, yielding a too small
   value, so we must ensure calloc rejects this.
*/
static void *ncalloc(size_t n, size_t s)
{
   void *result;
   size_t sz;

   result = ((void*)0);
   if (!n || (((size_t)-1) / n) > s) {
      sz = n * s;
      if ((result = nmalloc(sz))) memset(result, 0, sz);
      }
   return result;
} /* ncalloc */
# 3 "malloc.c" 2

void *malloc(size_t size)
{
    return nmalloc(size);
}

void free(void *ptr)
{
    return nfree(ptr);
}

void *realloc(void *ptr, size_t size)
{
    return nrealloc(ptr, size);
}

void *calloc(size_t n, size_t s)
{
    return ncalloc(n, s);
}
