# 0 "ioctl.c"
# 0 "<built-in>"
# 0 "<command-line>"
# 1 "ioctl.c"
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2010 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/*
** File    : /djgpp/src/libc/posix/fcntl/ioctl.c
** Author  : Tom Demmer (demmer@lstm.ruhr-uni-bochum.de)
** SYNOPSIS: ioctl() for djgpp.
** Aim:
**   Provide an ioctl function that does what the naive DOS C programmer
**   expects.
**   Provide an ioctl function that does what the naive UNIX C programmer
**   expects.
**   Mangle and shredder them, roll them into one function and expect it
**   to work both ways.
**   If you really do that, you _ARE_ naive.
**
**   Generally, we follow this approach:
**   Commands for ioctl are 8 bit long because they must fit into AL.
**   The other 8 bit are usually unused. We put some flags inside, e.g.
**   if we need a transfer buffer, which register the DOS ioctl call wants
**   to return on saturdays and sun shine and such. The low byte of the
**   high word takes some more flags.
**
**
**
**   The UNIX ioctls have the command in the lower word (16-Bit).
**   In the high word, the upper three bits tell us if the parameters
**   are in or out parameters. Bit 15 means IN, Bit 14 means OUT, Bit
**   13 means VOID. Because there _must_ be one of them set, we can
**   distinguish UNIX from DOS ioctls. The other tell us the size of the
**   Input/Output parameters.
**   The high byte of the lower word will be used to specify the class
**   to which the command belongs, e.g. `T' for tty. Currently not
**   implemented.
**
**
**   It looks like this:
**   DOS:
**   000fffff ffffffff XFFiibRR CCCCCCCC
**   UNIX:
**   IOVxxxxx xSSSSSSS CCCCCCCC CCCCCCCC
**
**   With
**      C: Command part
**      X: Use Xfer buffer
**      F: What to return
**      f: Still free. Will be used for the generioc ioctl request.
**      i: Number of additional input args (dx,si,di)
**         I think CX is always used or can be safely set to 0
**      b: Brain Dead Flag for generic ioctl
**      R: reserved (might become the number of args someday)
**
**      S: sizeof input/output parameter
**      I: Input  flag
**      O: Output flag
**      V: Void Flag (no parameters)
**      x: Currently unused.
**
**   Right now the function supports the DOS ioctl calls up to 0x440b.
**   If you want to add another function, it goes like this:
**   #define DOS_PLAIN_your_func  0xff // e.g
**   #define DOS_your_func (DOS_PLAIN_your_func | **                          [DOS_XFER]| **                          [DOS_RETAX/DOS_RETDX/DOS_RETDISI] |**                          [DOS_XINARGS(xtra-args)])
gs)])
**
**
** What is completely missing is the UNIX ioctl handler. I don't have the
** slightest idea about what to do here. Right now it just checks for an FSE
** and calls that if it exist. Otherwise we just return -1.
**
**
$Id: ioctl.c,v 1.10 2015/05/08 16:38:44 andris Exp $
$Log: ioctl.c,v $
Revision 1.10  2015/05/08 16:38:44  andris
src/libc/compat/ioctl/ioctl.c: add missing va_end (patch from Ozkan Sezer)

Revision 1.9  2015/05/02 07:32:15  andris
Update file copyright lines aftr running src/copyright.pl

Revision 1.8  2008/04/07 20:35:57  juan.guerrero
Use __FSEXT_func_wrapper instead of calling functions directly
with wrong type of arguments.
Patch by Ozkan Sezer, 2008-03-26.

Revision 1.7  2003/05/10 15:26:11  richdawe
__tb_size is defined in <go32.h> now.

Revision 1.6  2002/10/17 23:00:24  richdawe
Update copyright messages

Revision 1.5  2001/08/09 03:59:13  snowball
Add bare-bones implementation of TIOCSWINSZ.

Revision 1.4  2001/08/08 17:03:19  snowball
Implement TIOCGWINSZ for retrieving the screen height and width.

Revision 1.3  2001/06/20 16:54:33  ams
Do not mix signed and unsigned.

Revision 1.2  1998/06/28 17:25:20  dj
import djgpp 2.02

 * Revision 0.5  1996/07/30  09:42:23  DEMMER
 * Minor code cleanups. Final beta.
 *
 * Revision 0.4  1996/07/29  13:03:29  DEMMER
 * Added va_end(). Probably uneeded for Intel machines. Anyway.
 *
 * Revision 0.3  1996/07/29  12:44:55  DEMMER
 * Split the header stuff from the source
 * Changed encoding bits
 *
 * Revision 0.2  1996/07/04  11:17:37  DEMMER
 * Revised flag scheme.
 * Correct some minor bugs
 *
 * Revision 0.1  1996/07/03  15:42:01  DEMMER
 * Cleaned up flag stuff.
 * Added UNIX ioctl test routines
 *
 * Revision 0.0  1996/07/03  13:53:23  DEMMER
 * Initial version
 *
*/





# 1 "../../../../include/libc/stubs.h" 1
/* Copyright (C) 2007 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 18 "../../../../include/libc/stubs.h"
/* POSIX functions (for when compiling an ANSI function) */
# 89 "../../../../include/libc/stubs.h"
# 1 "../../../../include/stddef.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 14 "../../../../include/stddef.h"
# 1 "../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 23 "../../../../include/sys/djtypes.h"
/* For GCC 3.00 or later we use its builtin va_list.                       */
# 34 "../../../../include/sys/djtypes.h"
/* but remains a typedef in C */





/* Define internal type to be used for avoiding strict
   aliasing warnings */

typedef short __attribute__((__may_alias__)) __dj_short_a;
typedef int __attribute__((__may_alias__)) __dj_int_a;
typedef long __attribute__((__may_alias__)) __dj_long_a;
typedef long long __attribute__((__may_alias__)) __dj_long_long_a;
typedef unsigned short __attribute__((__may_alias__)) __dj_unsigned_short_a;
typedef unsigned int __attribute__((__may_alias__)) __dj_unsigned_int_a;
typedef unsigned long __attribute__((__may_alias__)) __dj_unsigned_long_a;
typedef unsigned long long __attribute__((__may_alias__)) __dj_unsigned_long_long_a;
typedef float __attribute__((__may_alias__)) __dj_float_a;
typedef double __attribute__((__may_alias__)) __dj_double_a;
typedef long double __attribute__((__may_alias__)) __dj_long_double_a;
# 15 "../../../../include/stddef.h" 2

/* Some programs think they know better... */
# 26 "../../../../include/stddef.h"
//#ifdef __cplusplus
//#define offsetof(s_type, mbr) ((std::size_t) &((s_type *)0)->mbr)
//#else

//#endif


typedef int ptrdiff_t;




typedef long unsigned int size_t;




typedef unsigned short wchar_t;
# 90 "../../../../include/libc/stubs.h" 2
# 1 "../../../../include/dpmi.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 40 "../../../../include/sys/cdefs.h"
/* Support gcc's __attribute__ facility.  */




/* Ensure that always traditional GNU extern inline semantics are used
   (aka -fgnu89-inline) even if ISO C99 semantics have been specified.  */
# 9 "../../../../include/dpmi.h" 2
# 1 "../../../../include/sys/nearptr.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */




# 1 "../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 17 "../../../../include/stdint.h"
typedef signed char int_least8_t;
typedef unsigned char uint_least8_t;
typedef signed char int_fast8_t;
typedef unsigned char uint_fast8_t;
typedef signed char int8_t;
typedef unsigned char uint8_t;

typedef signed short int int_least16_t;
typedef unsigned short int uint_least16_t;
typedef signed int int_fast16_t;
typedef unsigned int uint_fast16_t;
typedef signed short int int16_t;
typedef unsigned short int uint16_t;

typedef signed int int_least32_t;
typedef unsigned int uint_least32_t;
typedef signed int int_fast32_t;
typedef unsigned int uint_fast32_t;
typedef signed int int32_t;
typedef unsigned int uint32_t;

__extension__ typedef signed long long int int_least64_t;
__extension__ typedef unsigned long long int uint_least64_t;
__extension__ typedef signed long long int int_fast64_t;
__extension__ typedef unsigned long long int uint_fast64_t;
__extension__ typedef signed long long int int64_t;
__extension__ typedef unsigned long long int uint64_t;

typedef long int intptr_t;
typedef unsigned long int uintptr_t;

__extension__ typedef signed long long int intmax_t;
__extension__ typedef unsigned long long int uintmax_t;

/* ANSI/ISO C99 says these should not be visible in C++ unless
   explicitly requested.  */

//#if !defined(__cplusplus) || defined(__STDC_LIMIT_MACROS)
# 110 "../../../../include/stdint.h"
  /* These are defined by limits.h, so make them conditional.  */
# 124 "../../../../include/stdint.h"
//#endif /* !__cplusplus || __STDC_LIMIT_MACROS */
# 23 "../../../../include/libc/djthunks.h" 2

void crt1_startup(int handle);

uint8_t *djaddr2ptr(uint32_t addr);
uint8_t *djaddr2ptr2(uint32_t addr, uint32_t len);
uint32_t djptr2addr(const uint8_t *ptr);
uint32_t djthunk_get(const char *name);
uint32_t djthunk_get_h(int handle, const char *name);
void *djsbrk(int increment);
# 20 "../../../../include/libc/asmobj.h" 2
# 8 "../../../../include/sys/nearptr.h" 2
# 24 "../../../../include/sys/nearptr.h"
/* Functions to enable "near" pointer access to DOS memory under DPMI
   CW Sandmann 7-95  NO WARRANTY: WARNING, since these functions disable
   memory protection, they MAY DESTROY EVERYTHING ON YOUR COMPUTER! */

int __djgpp_nearptr_enable(void); /* Returns 0 if feature not avail */
void __djgpp_nearptr_disable(void); /* Enables protection */
extern int *____djgpp_selector_limit(void);/*
*/
# 30 "../../../../include/sys/nearptr.h"
       #define __djgpp_selector_limit (*____djgpp_selector_limit())/*
*/
# 30 "../../../../include/sys/nearptr.h"
                                       ; /* Limit on CS and on DS if prot */
extern unsigned *____djgpp_base_address(void);/*
*/
# 31 "../../../../include/sys/nearptr.h"
       #define __djgpp_base_address (*____djgpp_base_address())/*
*/
# 31 "../../../../include/sys/nearptr.h"
                                          ; /* Used in calculation below */
# 10 "../../../../include/dpmi.h" 2
# 1 "../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../include/libc/asmobj.h" 2
# 11 "../../../../include/dpmi.h" 2
# 1 "../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 12 "../../../../include/dpmi.h" 2
# 26 "../../../../include/dpmi.h"
//#ifndef _POSIX_SOURCE

extern unsigned short *____dpmi_error(void);/*
*/
# 28 "../../../../include/dpmi.h"
       #define __dpmi_error (*____dpmi_error())/*
*/
# 28 "../../../../include/dpmi.h"
                                        ;

typedef struct {
  unsigned short offset16;
  unsigned short segment;
} __dpmi_raddr;

typedef struct {
  unsigned int offset32;
  unsigned short selector;
} __dpmi_paddr;

typedef struct {
  unsigned int handle; /* 0, 2 */
  unsigned int size; /* or count */ /* 4, 6 */
  unsigned int address; /* 8, 10 */
} __dpmi_meminfo;

typedef union {
  struct {
    unsigned int edi;
    unsigned int esi;
    unsigned int ebp;
    unsigned int res;
    unsigned int ebx;
    unsigned int edx;
    unsigned int ecx;
    unsigned int eax;
  } d;
  struct {
    unsigned short di, di_hi;
    unsigned short si, si_hi;
    unsigned short bp, bp_hi;
    unsigned short res, res_hi;
    unsigned short bx, bx_hi;
    unsigned short dx, dx_hi;
    unsigned short cx, cx_hi;
    unsigned short ax, ax_hi;
    unsigned short flags;
    unsigned short es;
    unsigned short ds;
    unsigned short fs;
    unsigned short gs;
    unsigned short ip;
    unsigned short cs;
    unsigned short sp;
    unsigned short ss;
  } x;
  struct {
    unsigned char edi[4];
    unsigned char esi[4];
    unsigned char ebp[4];
    unsigned char res[4];
    unsigned char bl, bh, ebx_b2, ebx_b3;
    unsigned char dl, dh, edx_b2, edx_b3;
    unsigned char cl, ch, ecx_b2, ecx_b3;
    unsigned char al, ah, eax_b2, eax_b3;
  } h;
} __attribute__((packed)) __dpmi_regs;

typedef struct {
  unsigned char major;
  unsigned char minor;
  unsigned short flags;
  unsigned char cpu;
  unsigned char master_pic;
  unsigned char slave_pic;
} __dpmi_version_ret;

typedef struct {
  unsigned int largest_available_free_block_in_bytes;
  unsigned int maximum_unlocked_page_allocation_in_pages;
  unsigned int maximum_locked_page_allocation_in_pages;
  unsigned int linear_address_space_size_in_pages;
  unsigned int total_number_of_unlocked_pages;
  unsigned int total_number_of_free_pages;
  unsigned int total_number_of_physical_pages;
  unsigned int free_linear_address_space_in_pages;
  unsigned int size_of_paging_file_partition_in_pages;
  unsigned int reserved[3];
} __dpmi_free_mem_info;

typedef struct {
  unsigned int total_allocated_bytes_of_physical_memory_host;
  unsigned int total_allocated_bytes_of_virtual_memory_host;
  unsigned int total_available_bytes_of_virtual_memory_host;
  unsigned int total_allocated_bytes_of_virtual_memory_vcpu;
  unsigned int total_available_bytes_of_virtual_memory_vcpu;
  unsigned int total_allocated_bytes_of_virtual_memory_client;
  unsigned int total_available_bytes_of_virtual_memory_client;
  unsigned int total_locked_bytes_of_memory_client;
  unsigned int max_locked_bytes_of_memory_client;
  unsigned int highest_linear_address_available_to_client;
  unsigned int size_in_bytes_of_largest_free_memory_block;
  unsigned int size_of_minimum_allocation_unit_in_bytes;
  unsigned int size_of_allocation_alignment_unit_in_bytes;
  unsigned int reserved[19];
} __dpmi_memory_info;

typedef struct {
  unsigned int data16[2];
  unsigned int code16[2];
  unsigned short ip;
  unsigned short reserved;
  unsigned int data32[2];
  unsigned int code32[2];
  unsigned int eip;
} __dpmi_callback_info;

typedef struct {
  unsigned int size_requested;
  unsigned int size;
  unsigned int handle;
  unsigned int address;
  unsigned int name_offset;
  unsigned short name_selector;
  unsigned short reserved1;
  unsigned int reserved2;
} __dpmi_shminfo;

/* Unless otherwise noted, all functions return -1 on error, setting __dpmi_error to the DPMI error code */

void __dpmi_yield(void); /* INT 0x2F AX=1680 */

int __dpmi_allocate_ldt_descriptors(int _count); /* DPMI 0.9 AX=0000 */
int __dpmi_free_ldt_descriptor(int _descriptor); /* DPMI 0.9 AX=0001 */
int __dpmi_segment_to_descriptor(int _segment); /* DPMI 0.9 AX=0002 */
int __dpmi_get_selector_increment_value(void); /* DPMI 0.9 AX=0003 */
int __dpmi_get_segment_base_address(int _selector, unsigned int *_addr); /* DPMI 0.9 AX=0006 */
int __dpmi_set_segment_base_address(int _selector, unsigned int _address); /* DPMI 0.9 AX=0007 */
unsigned int __dpmi_get_segment_limit(int _selector); /* LSL instruction  */
int __dpmi_set_segment_limit(int _selector, unsigned int _limit); /* DPMI 0.9 AX=0008 */
int __dpmi_get_descriptor_access_rights(int _selector); /* LAR instruction  */
int __dpmi_set_descriptor_access_rights(int _selector, int _rights); /* DPMI 0.9 AX=0009 */
int __dpmi_create_alias_descriptor(int _selector); /* DPMI 0.9 AX=000a */
int __dpmi_get_descriptor(int _selector, unsigned char _buffer[8]); /* DPMI 0.9 AX=000b */
int __dpmi_set_descriptor(int _selector, unsigned char _buffer[8]); /* DPMI 0.9 AX=000c */
int __dpmi_allocate_specific_ldt_descriptor(int _selector); /* DPMI 0.9 AX=000d */


int __dpmi_get_multiple_descriptors(int _count, void *_buffer); /* DPMI 1.0 AX=000e */
int __dpmi_set_multiple_descriptors(int _count, void *_buffer); /* DPMI 1.0 AX=000f */

int __dpmi_allocate_dos_memory(int _paragraphs, int *_ret_selector_or_max); /* DPMI 0.9 AX=0100 */
int __dpmi_free_dos_memory(int _selector); /* DPMI 0.9 AX=0101 */
int __dpmi_resize_dos_memory(int _selector, int _newpara, int *_ret_max); /* DPMI 0.9 AX=0102 */

int __dpmi_get_real_mode_interrupt_vector(int _vector, __dpmi_raddr *_address); /* DPMI 0.9 AX=0200 */
int __dpmi_set_real_mode_interrupt_vector(int _vector, __dpmi_raddr *_address); /* DPMI 0.9 AX=0201 */
int __dpmi_get_processor_exception_handler_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0202 */
int __dpmi_set_processor_exception_handler_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0203 */
int __dpmi_get_protected_mode_interrupt_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0204 */
int __dpmi_set_protected_mode_interrupt_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0205 */

int __dpmi_get_extended_exception_handler_vector_pm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0210 */
int __dpmi_get_extended_exception_handler_vector_rm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0211 */
int __dpmi_set_extended_exception_handler_vector_pm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0212 */
int __dpmi_set_extended_exception_handler_vector_rm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0213 */

int __dpmi_simulate_real_mode_interrupt(int _vector, __dpmi_regs *_regs); /* DPMI 0.9 AX=0300 */
int __dpmi_int(int _vector, __dpmi_regs *_regs); /* like above, but sets ss sp fl */ /* DPMI 0.9 AX=0300 */
extern short __dpmi_int_ss, __dpmi_int_sp, __dpmi_int_flags; /* default to zero */
int __dpmi_simulate_real_mode_procedure_retf(__dpmi_regs *_regs); /* DPMI 0.9 AX=0301 */
int __dpmi_simulate_real_mode_procedure_retf_stack(__dpmi_regs *_regs, int stack_words_to_copy, const void *stack_data); /* DPMI 0.9 AX=0301 */
int __dpmi_simulate_real_mode_procedure_iret(__dpmi_regs *_regs); /* DPMI 0.9 AX=0302 */
int __dpmi_allocate_real_mode_callback(unsigned int _handler, unsigned int _regs, __dpmi_raddr *_ret); /* DPMI 0.9 AX=0303 */
int __dpmi_free_real_mode_callback(__dpmi_raddr *_addr); /* DPMI 0.9 AX=0304 */
int __dpmi_get_state_save_restore_addr( __dpmi_raddr *_rm, __dpmi_paddr *_pm); /* DPMI 0.9 AX=0305 */
int __dpmi_get_raw_mode_switch_addr( __dpmi_raddr *_rm, __dpmi_paddr *_pm); /* DPMI 0.9 AX=0306 */

int __dpmi_get_version( __dpmi_version_ret *_ret); /* DPMI 0.9 AX=0400 */

int __dpmi_get_capabilities( int *_flags, char vendor_info[128]); /* DPMI 1.0 AX=0401 */

int __dpmi_get_free_memory_information( __dpmi_free_mem_info *_info); /* DPMI 0.9 AX=0500 */
int __dpmi_allocate_memory(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0501 */
int __dpmi_free_memory(unsigned int _handle); /* DPMI 0.9 AX=0502 */
int __dpmi_resize_memory(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0503 */

int __dpmi_allocate_linear_memory(__dpmi_meminfo *_info, int _commit); /* DPMI 1.0 AX=0504 */
int __dpmi_resize_linear_memory(__dpmi_meminfo *_info, int _commit); /* DPMI 1.0 AX=0505 */
int __dpmi_get_page_attributes(__dpmi_meminfo *_info, short *_buffer); /* DPMI 1.0 AX=0506 */
int __dpmi_set_page_attributes(__dpmi_meminfo *_info, short *_buffer); /* DPMI 1.0 AX=0507 */
int __dpmi_map_device_in_memory_block(__dpmi_meminfo *_info, unsigned int _physaddr); /* DPMI 1.0 AX=0508 */
int __dpmi_map_conventional_memory_in_memory_block(__dpmi_meminfo *_info, unsigned int _linaddr); /* DPMI 1.0 AX=0509 */
int __dpmi_get_memory_block_size_and_base(__dpmi_meminfo *_info); /* DPMI 1.0 AX=050a */
int __dpmi_get_memory_information( __dpmi_memory_info *_buffer); /* DPMI 1.0 AX=050b */

int __dpmi_lock_linear_region(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0600 */
int __dpmi_unlock_linear_region(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0601 */
int __dpmi_mark_real_mode_region_as_pageable(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0602 */
int __dpmi_relock_real_mode_region(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0603 */
int __dpmi_get_page_size( unsigned int *_size); /* DPMI 0.9 AX=0604 */

int __dpmi_mark_page_as_demand_paging_candidate(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0702 */
int __dpmi_discard_page_contents(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0703 */

int __dpmi_physical_address_mapping(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0800 */
int __dpmi_free_physical_address_mapping(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0801 */

/* These next four functions return the old state */
int __dpmi_get_and_disable_virtual_interrupt_state(void); /* DPMI 0.9 AX=0900 */
int __dpmi_get_and_enable_virtual_interrupt_state(void); /* DPMI 0.9 AX=0901 */
int __dpmi_get_and_set_virtual_interrupt_state(int _old_state); /* DPMI 0.9 AH=09   */
int __dpmi_get_virtual_interrupt_state(void); /* DPMI 0.9 AX=0902 */

int __dpmi_get_vendor_specific_api_entry_point(const char *_id, __dpmi_paddr *_api); /* DPMI 0.9 AX=0a00 */

int __dpmi_set_debug_watchpoint(__dpmi_meminfo *_info, int _type); /* DPMI 0.9 AX=0b00 */
int __dpmi_clear_debug_watchpoint(unsigned int _handle); /* DPMI 0.9 AX=0b01 */
int __dpmi_get_state_of_debug_watchpoint(unsigned int _handle, int *_status); /* DPMI 0.9 AX=0b02 */
int __dpmi_reset_debug_watchpoint(unsigned int _handle); /* DPMI 0.9 AX=0b03 */

int __dpmi_install_resident_service_provider_callback(__dpmi_callback_info *_info); /* DPMI 1.0 AX=0c00 */
int __dpmi_terminate_and_stay_resident(int return_code, int paragraphs_to_keep); /* DPMI 1.0 AX=0c01 */

int __dpmi_allocate_shared_memory(__dpmi_shminfo *_info); /* DPMI 1.0 AX=0d00 */
int __dpmi_free_shared_memory(unsigned int _handle); /* DPMI 1.0 AX=0d01 */
int __dpmi_serialize_on_shared_memory(unsigned int _handle, int _flags); /* DPMI 1.0 AX=0d02 */
int __dpmi_free_serialization_on_shared_memory(unsigned int _handle, int _flags); /* DPMI 1.0 AX=0d03 */

int __dpmi_get_coprocessor_status(void); /* DPMI 1.0 AX=0e00 */
int __dpmi_set_coprocessor_emulation(int _flags); /* DPMI 1.0 AX=0e01 */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Backwards compatibility stuff						       */



typedef struct {
  unsigned int available_memory;
  unsigned int available_pages;
  unsigned int available_lockable_pages;
  unsigned int linear_space;
  unsigned int unlocked_pages;
  unsigned int available_physical_pages;
  unsigned int total_physical_pages;
  unsigned int free_linear_space;
  unsigned int max_pages_in_paging_file;
  unsigned int reserved[3];
} _go32_dpmi_meminfo;







typedef struct {
  unsigned int size;
  unsigned int pm_offset;
  unsigned short pm_selector;
  unsigned short rm_offset;
  unsigned short rm_segment;
} _go32_dpmi_seginfo;

/* returns zero if success, else dpmi error and info->size is max size */
int _go32_dpmi_allocate_dos_memory(_go32_dpmi_seginfo *info);
 /* set size to bytes/16, call, use rm_segment.  Do not
	   change anthing but size until the memory is freed.
	   If error, max size is returned in size as bytes/16. */
int _go32_dpmi_free_dos_memory(_go32_dpmi_seginfo *info);
 /* set new size to bytes/16, call.  If error, max size
	   is returned in size as bytes/16 */
int _go32_dpmi_resize_dos_memory(_go32_dpmi_seginfo *info);
 /* uses pm_selector to free memory */

/* These both use the rm_segment:rm_offset fields only */
int _go32_dpmi_get_real_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
int _go32_dpmi_set_real_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);

/* These do NOT wrap the function in pm_offset in an iret handler.
   You must provide an assembler interface yourself, or alloc one below.
   You may NOT LONGjmp out of an interrupt handler. */
int _go32_dpmi_get_protected_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
 /* puts vector in pm_selector:pm_offset. */
int _go32_dpmi_set_protected_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
 /* sets vector from pm_offset and pm_selector */

/********** HELPER FUNCTIONS **********/

int _go32_dpmi_chain_protected_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
 /* sets up wrapper that calls function in pm_offset, chaining to old
	   handler when it returns */

/* These generate assember IRET-style wrappers for functions and set up stack */
int _go32_dpmi_allocate_iret_wrapper(_go32_dpmi_seginfo *info);
 /* Put function ptr in pm_offset, call, returns wrapper entry in pm_offset. */
int _go32_dpmi_free_iret_wrapper(_go32_dpmi_seginfo *info);
 /* assumes pm_offset points to wrapper, frees it */

/* RMCB functions, automatically restructure the real-mode stack for the
   proper return type and set up correct PM stack.  The callback
   (info->pm_offset) is called as (*pmcb)(_go32_dpmi_registers *regs); */
int _go32_dpmi_allocate_real_mode_callback_retf(_go32_dpmi_seginfo *info, unsigned int regs);
 /* points callback at pm_offset, returns seg:ofs of callback addr
	   in rm_segment:rm_offset.  Do not change any fields until freed.
	   Interface is added to simulate far return */
int _go32_dpmi_allocate_real_mode_callback_iret(_go32_dpmi_seginfo *info, unsigned int regs);
 /* same, but simulates iret */
int _go32_dpmi_free_real_mode_callback(_go32_dpmi_seginfo *info);
 /* frees callback */

/* The following two variables may be used to change the default stack size
   for interrupts and rmcb wrappers to a user defined size from the default
   of 32Kbytes.  Each RMCB and chain/iret wrapper gets its own stack. */

extern unsigned int _go32_interrupt_stack_size;
extern unsigned int _go32_rmcb_stack_size;

/* Convenience functions, the return value is *bytes* */
unsigned int _go32_dpmi_remaining_physical_memory(void);
unsigned int _go32_dpmi_remaining_virtual_memory(void);

/* locks memory from a specified offset within the code/data selector */
int _go32_dpmi_lock_code( unsigned int _lockaddr, unsigned int _locksize);
int _go32_dpmi_lock_data( unsigned int _lockaddr, unsigned int _locksize);

int __djgpp_set_page_attributes(void *our_addr, unsigned int num_bytes,
           unsigned short attributes);
int __djgpp_map_physical_memory(void *our_addr, unsigned int num_bytes,
           uint64_t phys_addr);

//#endif /* !_POSIX_SOURCE */
# 91 "../../../../include/libc/stubs.h" 2
# 1 "../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../include/libc/asmobj.h" 2
# 92 "../../../../include/libc/stubs.h" 2
int __spawnve(int mode, const char *path, char *const argv[],
              char *const envp[]);
int __chdir (const char *mydirname);
char * __getcwd(char *buf, size_t size);
int __brk(void *_heaptop);
unsigned int __sbrk(int delta);
void *sbrk(int delta);
long double __modfl(long double _x, long double *_pint);

extern unsigned *____djgpp_timer_countdown(void);/*
*/
# 101 "../../../../include/libc/stubs.h"
       #define __djgpp_timer_countdown (*____djgpp_timer_countdown())/*
*/
# 101 "../../../../include/libc/stubs.h"
                                             ;
extern int *____djgpp_timer_hdlr(void);/*
*/
# 102 "../../../../include/libc/stubs.h"
       #define __djgpp_timer_hdlr (*____djgpp_timer_hdlr())/*
*/
# 102 "../../../../include/libc/stubs.h"
                                   ;
# 133 "ioctl.c" 2
# 1 "../../../../include/stdarg.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 18 "../../../../include/stdarg.h"
# 1 "../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 19 "../../../../include/stdarg.h" 2


typedef __builtin_va_list va_list;
# 134 "ioctl.c" 2
# 1 "../../../../include/stdlib.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 20 "../../../../include/stdlib.h"
# 1 "../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../include/sys/types.h"
# 1 "../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../include/sys/types.h" 2

typedef int blkcnt_t;
typedef int blksize_t;
typedef int dev_t;
typedef unsigned long fsblkcnt_t;
typedef unsigned long fsfilcnt_t;
typedef int ino_t;
typedef int mode_t;
typedef int nlink_t;


typedef int gid_t;



typedef int off_t;



__extension__ typedef long long off64_t;



__extension__ typedef long long offset_t;



typedef int pid_t;







typedef long signed int ssize_t;




typedef int uid_t;




/* Allow including program to override.  */




typedef struct fd_set {
  unsigned char fd_bits [((256) + 7) / 8];
} fd_set;







typedef unsigned int time_t;
# 21 "../../../../include/stdlib.h" 2
# 1 "../../../../include/stddef.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 22 "../../../../include/stdlib.h" 2

/* Some programs think they know better... */
# 38 "../../../../include/stdlib.h"
extern int __dj_mb_cur_max;

typedef struct {
  int quot;
  int rem;
} div_t;

typedef struct {
  long quot;
  long rem;
} ldiv_t;





void abort(void);
int abs(int _i);

int __dj_atexit(void (*_func)(void));
double atof(const char *_s);
int atoi(const char *_s);
long atol(const char *_s);
void * bsearch(const void *_key, const void *_base, size_t _nelem,
  size_t _size, int (*_cmp)(const void *_ck, const void *_ce));
void * calloc(size_t _nelem, size_t _size);
unsigned int calloc32(size_t _nelem, size_t _size);
div_t div(int _numer, int _denom);
void exit(int _status);
void free(void *_ptr);
void free32(unsigned int _addr);
char * getenv(const char *_name);
long labs(long _i);
ldiv_t ldiv(long _numer, long _denom);
void * malloc(size_t _size);
unsigned int malloc32(size_t _size);
int mblen(const char *_s, size_t _n);
size_t mbstowcs(wchar_t *_wcs, const char *_s, size_t _n);
int mbtowc(wchar_t *_pwc, const char *_s, size_t _n);
void qsort(void *_base, size_t _nelem, size_t _size,
       int (*_cmp)(const void *_e1, const void *_e2));
int rand(void);
void * realloc(void *_ptr, size_t _size);
unsigned int realloc32(unsigned int _addr, size_t _size);
void srand(unsigned _seed);
double strtod(const char *_s, char **_endptr);
long strtol(const char *_s, char **_endptr, int _base);
unsigned long strtoul(const char *_s, char **_endptr, int _base);
int system(const char *_s);
size_t wcstombs(char *_s, const wchar_t *_wcs, size_t _n);
int wctomb(char *_s, wchar_t _wchar);




typedef struct {
  long long int quot;
  long long int rem;
} lldiv_t;

void _Exit(int _status);
long long int atoll(const char *_s);
long long int llabs(long long int _i);
lldiv_t lldiv(long long int _numer, long long int _denom);
float strtof(const char *_s, char **_endptr);
long double strtold(const char *_s, char **_endptr);
long long int strtoll(const char *_s, char **_endptr, int _base);
unsigned long long int strtoull(const char *_s, char **_endptr, int _base);





long a64l(const char *_string);
char * l64a(long _value);
char * mktemp(char *_template);
char * mkdtemp(char *_template);
int mkstemp(char *_template);
int putenv(char *_val);
char * realpath(const char *_path, char *_resolved);
int setenv(const char *_var, const char *_val, int _overwrite);
int unsetenv(const char *_var);



void * alloca(size_t _size);
long double _atold(const char *_s);
void cfree(void *_ptr);
double drand48(void);
char * ecvtbuf(double _val, int _nd, int *_dp, int *_sn, char *_bf);
char * ecvt(double _val, int _nd, int *_dp, int *_sn);
double erand48(unsigned short _state[3]);
char * fcvtbuf(double _val, int _nd, int *_dp, int *_sn, char *_bf);
char * fcvt(double _val, int _nd, int *_dp, int *_sn);
char * gcvt(double _val, int _nd, char *_buf);
char * getpass(const char *_prompt);
int getlongpass(const char *_prompt, char *_buffer, int _max_len);
char * itoa(int _value, char *_buffer, int _radix);
long jrand48(unsigned short _state[3]);
void lcong48(unsigned short _param[7]);
unsigned long lrand48(void);
long mrand48(void);
unsigned long nrand48(unsigned short _state[3]);
unsigned short *seed48(unsigned short _state_seed[3]);
void srand48(long _seedval);
long double _strtold(const char *_s, char **_endptr);
void swab(const void *_from, void *_to, ssize_t _nbytes);
void * valloc (size_t _amt);





/* BSD Random Number Generator */
char * initstate (unsigned _seed, char *_arg_state, int _n);
char * setstate(char *_arg_state);
long random(void);
int srandom(int _seed);

/* These vary in expected prototype, so we just don't prototype them.
void		xfree(void *_ptr);
void *		xmalloc(size_t _size);
void *		xrealloc(void *ptr, size_t _size);
*/
# 173 "../../../../include/stdlib.h"
extern int __system_flags;
# 135 "ioctl.c" 2
# 1 "../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 136 "ioctl.c" 2
# 1 "../../../../include/sys/fsext.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../include/sys/fsext.h"
# 1 "../../../../include/stdarg.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../include/sys/fsext.h" 2

typedef enum {
  __FSEXT_nop,
  __FSEXT_open,
  __FSEXT_creat,
  __FSEXT_read,
  __FSEXT_write,
  __FSEXT_ready,
  __FSEXT_close,
  __FSEXT_fcntl,
  __FSEXT_ioctl,
  __FSEXT_lseek,
  __FSEXT_link,
  __FSEXT_unlink,
  __FSEXT_dup,
  __FSEXT_dup2,
  __FSEXT_fstat,
  __FSEXT_stat,
  __FSEXT_llseek,
  __FSEXT_readlink,
  __FSEXT_symlink,
  __FSEXT_fchown,
  __FSEXT_chmod,
  __FSEXT_chown,
  __FSEXT_fchmod
} __FSEXT_Fnumber;

/* _ready gets passed a fd and should return a mask of these,
   as if we were emulating "int ready(int fd)" */




/* The return value is nonzero if the function has overridden the
   caller's functionality. */
typedef int (__FSEXT_Function)(__FSEXT_Fnumber _function_number,
          int *_rv, va_list _args);

int __FSEXT_alloc_fd(__FSEXT_Function *_function);
int __FSEXT_set_function(int _fd, __FSEXT_Function *_function);
__FSEXT_Function *__FSEXT_get_function(int _fd);
void *__FSEXT_set_data(int _fd, void *_data);
void *__FSEXT_get_data(int _fd);

int __FSEXT_add_open_handler(__FSEXT_Function *_function);
int __FSEXT_call_open_handlers(__FSEXT_Fnumber _function_number,
          int *rv, va_list _args);
# 137 "ioctl.c" 2
# 1 "../../../../include/libc/fsexthlp.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../include/stdarg.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 8 "../../../../include/libc/fsexthlp.h" 2
# 1 "../../../../include/sys/fsext.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 9 "../../../../include/libc/fsexthlp.h" 2
# 25 "../../../../include/libc/fsexthlp.h"
/* This is a wrapper for an FSEXT function that allows it to be called
 * with a variable number of arguments. */

static inline int
__FSEXT_func_wrapper (__FSEXT_Function *func,
        __FSEXT_Fnumber fnum,
        int *rv,
        ...)
{
  va_list args;
  int ret;

  __builtin_va_start(args,rv);
  ret = func(fnum, rv, args);
  __builtin_va_end(args);

  return(ret);
}

/* This is a wrapper for __FSEXT_call_open_handlers that allows it
 * to be called with a variable number of arguments. */

static inline int
__FSEXT_call_open_handlers_wrapper (__FSEXT_Fnumber fnum, int *rv, ...)
{
  va_list args;
  int ret;

  __builtin_va_start(args,rv);
  ret = __FSEXT_call_open_handlers(fnum, rv, args);
  __builtin_va_end(args);

  return(ret);
}
# 138 "ioctl.c" 2
# 1 "../../../../include/sys/ioctl.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
# 23 "../../../../include/sys/ioctl.h"
/*
**  plain ioctl functions. Do not use them.
**  Never. Really _N_E_V_E_R_. Unless you really know what
**  you are doing.
**
*/
# 47 "../../../../include/sys/ioctl.h"
/*
**  Flags for DOS commands
*/


                                       /* what to return. 8 possible values */

                                       /* bytes to transfer                 */





/*
** DOS ioctls we support:
*/
# 75 "../../../../include/sys/ioctl.h"
/*
These ones do not fit into my scheme, because they _DO_NOT_ put the size
of the xfer buffer in CX. Aaaaargh
*/
# 89 "../../../../include/sys/ioctl.h"
/* UNIX stuff - Most of it disabled.  */

/*
** __WARNING__ :
** This ifdef works for DJGPP, because its io.h
** defines __djgpp_include_io_h_
*/

/*
* Ioctl's have the command encoded in the lower word,
* and the size of any in or out parameters in the upper
* word.  The high 2 bits of the upper word are used
* to encode the in/out status of the parameter; for now
* we restrict parameters to at most 128 bytes.
*/





/* the 0x20000000 is so we can distinguish new ioctl's from old */



/* this should be _IORW, but stdio got there first */



/*
 * window size structure used with TXSETWIN and TXGETWIN.  This is
 * exactly the same as the Berkeley structure and can be used with
 * TIOCSWINSZ and TIOCGWINSZ -- in fact they are defined to be the
 * same.
 */
struct winsize {
 unsigned short ws_row; /* rows, in characters */
 unsigned short ws_col; /* columns, in characters */
 unsigned short ws_xpixel; /* horizontal size, pixels */
 unsigned short ws_ypixel; /* vertical size, pixels */
};
# 351 "../../../../include/sys/ioctl.h"
int ioctl( int fd, int cmd, ...);
# 139 "ioctl.c" 2
# 1 "../../../../include/libc/farptrgs.h" 1
/* special version for libc - uses %gs instead of %fs.  Ignore comments */

/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
/* Copyright (c) 1995 DJ Delorie.  Permission granted to use for any
   purpose, provided this copyright remains attached and unmodified.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

ͻ
		Far Pointer Simulation Functions			
ͼ

This file attempts to make up for the lack of a "far" keyword in GCC.
Although it doesn't provide access to far call APIs (like Windows), it
does allow you to do far pointer data access without the overhead of
movedata() or dosmemget/dosmemput().

You should *always* include this file when using these functions and
compile with optimization enabled.  They don't exist as normal functions
in any library, and they compile down to only a few opcodes when used
this way.  They are almost as fast as native pointer operations, and
about as fast as far pointers can get.

If you don't use optimization, this file becomes prototypes for
farptr.c, which generates real functions for these when not optimizing.
When optimizing, farptr.c compiles to nothing.

There are two types of functions here - standalone and invariant.  The
standalone functions take a selector and offset.  These are used when
you need only a few accesses, time isn't critical, or you don't know
what's in the %gs register.  The invariant ones don't take a selector,
they only take an offset.  These are used inside loops and in
time-critical accesses where the selector doesn't change.  To specify
the selector, use the farsetsel() function.  That selector is used for
all farns*() functions until changed.  You can use _fargetsel() if you
want to temporary change the selector with _farsetsel() and restore
it afterwards.

The farpoke* and farpeek* take selectors.

The farnspoke* and farnspeek* don't (note the `ns' for `no selector').

Warning: These routines all use the %gs register for their accesses.
GCC normally uses only %ds and %es, and libc functions (movedata,
dosmemget, dosmemput) use %gs.  Still, you should be careful about
assumptions concerning whether or not the value you put in %gs will be
preserved across calls to other functions.  If you guess wrong, your
program will crash.  Better safe than sorry.

*/
# 74 "../../../../include/libc/farptrgs.h"
# 1 "../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 75 "../../../../include/libc/farptrgs.h" 2

void _farpokeb_gs(unsigned short, unsigned int, unsigned char);
void _farpokew_gs(unsigned short, unsigned int, unsigned short);
void _farpokel_gs(unsigned short, unsigned int, unsigned int);
unsigned char _farpeekb_gs(unsigned short, unsigned int);
unsigned short _farpeekw_gs(unsigned short, unsigned int);
unsigned int _farpeekl_gs(unsigned short, unsigned int);
void _farsetsel_gs(unsigned short);
unsigned short _fargetsel_gs(void);
void _farnspokeb_gs(unsigned int, unsigned char);
void _farnspokew_gs(unsigned int, unsigned short);
void _farnspokel_gs(unsigned int, unsigned int);
unsigned char _farnspeekb_gs(unsigned int);
unsigned short _farnspeekw_gs(unsigned int);
unsigned int _farnspeekl_gs(unsigned int);
# 140 "ioctl.c" 2


/****************************************************************************/
/****************************************************************************/
/* S T A R T   O F   I M P L E M E N T A T I O N  ***************************/
/****************************************************************************/
/****************************************************************************/


# 1 "../../../../include/errno.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */







# 1 "../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../include/libc/asmobj.h" 2
# 16 "../../../../include/errno.h" 2







extern int *____dj_errno(void);/*
*/
# 23 "../../../../include/errno.h"
       #define __dj_errno (*____dj_errno())/*
*/
# 23 "../../../../include/errno.h"
                             ;
# 77 "../../../../include/errno.h"
/* EILSEQ		41 */

extern char * sys_errlist[];
extern int sys_nerr;
extern const char * __sys_errlist[];
extern int __sys_nerr;
extern int _doserrno;
# 150 "ioctl.c" 2
# 1 "../../../../include/go32.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../include/go32.h"
# 1 "../../../../include/sys/version.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../include/go32.h" 2
# 1 "../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../include/go32.h" 2
# 1 "../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../include/libc/asmobj.h" 2
# 27 "../../../../include/go32.h" 2
# 1 "../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 28 "../../../../include/go32.h" 2






typedef struct {
  unsigned int size_of_this_structure_in_bytes;
  unsigned int linear_address_of_primary_screen;
  unsigned int linear_address_of_secondary_screen;
  unsigned int linear_address_of_transfer_buffer;
  unsigned int size_of_transfer_buffer; /* >= 4k */
  unsigned int pid;
  unsigned char master_interrupt_controller_base;
  unsigned char slave_interrupt_controller_base;
  unsigned short selector_for_linear_memory;
  unsigned int linear_address_of_stub_info_structure;
  unsigned int linear_address_of_original_psp;
  unsigned short run_mode;
  unsigned short run_mode_info;
} __Go32_Info_Block;

extern __Go32_Info_Block *___go32_info_block(void);/*
*/
# 50 "../../../../include/go32.h"
       #define _go32_info_block (*___go32_info_block())/*
*/
# 50 "../../../../include/go32.h"
                                                 ;







# 1 "../../../../include/sys/movedata.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../include/sys/movedata.h"
# 1 "../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../include/sys/movedata.h" 2
# 1 "../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../include/sys/movedata.h" 2

# 1 "../../../../include/sys/fmemcpy.h" 1



# 1 "../../../../include/dpmi.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 5 "../../../../include/sys/fmemcpy.h" 2

void fmemcpy1(__dpmi_paddr dst, const void *src, unsigned len);
void fmemcpy2(void *dst, __dpmi_paddr src, unsigned len);
void fmemcpy12(__dpmi_paddr dst, __dpmi_paddr src, unsigned len);
# 28 "../../../../include/sys/movedata.h" 2
# 36 "../../../../include/sys/movedata.h"
/* This header is intended to be included only by other
   headers, like <go32.h> and <string.h>.  You may
   include this directly, but it will be non-portable. */

/* These lengths are in bytes, optimized for speed */
void dosmemget(unsigned long _offset, size_t _length, void *_buffer);
void dosmemput(const void *_buffer, size_t _length, unsigned long _offset);

/* The lengths here are in TRANSFERS, not bytes! */
void _dosmemgetb(unsigned long _offset, size_t _xfers, void *_buffer);
void _dosmemgetw(unsigned long _offset, size_t _xfers, void *_buffer);
void _dosmemgetl(unsigned long _offset, size_t _xfers, void *_buffer);
void _dosmemputb(const void *_buffer, size_t _xfers, unsigned long _offset);
void _dosmemputw(const void *_buffer, size_t _xfers, unsigned long _offset);
void _dosmemputl(const void *_buffer, size_t _xfers, unsigned long _offset);
# 59 "../../../../include/go32.h" 2
# 1 "../../../../include/sys/segments.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../include/sys/segments.h"
int _my_cs(void);
int _my_ds(void);
int _my_ss(void);
# 60 "../../../../include/go32.h" 2
# 70 "../../../../include/go32.h"
/* returns number of times hit since last call. (zero first time) */
unsigned _go32_was_ctrl_break_hit(void);
void _go32_want_ctrl_break(int yes); /* auto-yes if call above function */
# 151 "ioctl.c" 2
# 1 "../../../../include/dpmi.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 152 "ioctl.c" 2
# 1 "../../../../include/sys/fsext.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 153 "ioctl.c" 2
# 1 "../../../../include/dos.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 12 "../../../../include/dos.h" 2
# 24 "../../../../include/dos.h"
# 1 "../../../../include/pc.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */







# 1 "../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 15 "../../../../include/pc.h" 2
# 27 "../../../../include/pc.h"
unsigned char inportb (unsigned short _port);
unsigned short inportw (unsigned short _port);
uint32_t inportl (unsigned short _port);
void inportsb (unsigned short _port, unsigned char *_buf, unsigned _len);
void inportsw (unsigned short _port, unsigned short *_buf, unsigned _len);
void inportsl (unsigned short _port, uint32_t *_buf, unsigned _len);
void outportb (unsigned short _port, unsigned char _data);
void outportw (unsigned short _port, unsigned short _data);
void outportl (unsigned short _port, uint32_t _data);
void outportsb(unsigned short _port, const unsigned char *_buf, unsigned _len);
void outportsw(unsigned short _port, const unsigned short *_buf, unsigned _len);
void outportsl(unsigned short _port, const uint32_t *_buf, unsigned _len);

unsigned char inp(unsigned short _port);
unsigned short inpw(unsigned short _port);
void outp(unsigned short _port, unsigned char _data);
void outpw(unsigned short _port, unsigned short _data);

int kbhit(void);

int getkey(void); /* ALT's have 0x100 set */
int getxkey(void); /* ALT's have 0x100 set, 0xe0 sets 0x200 */
const unsigned char * __get_extended_key_string(int _xkey_code);
void sound(int _frequency);



extern unsigned char ScreenAttrib;




int ScreenMode(void);
int ScreenRows(void);
int ScreenCols(void);
void ScreenPutChar(int _ch, int _attr, int _x, int _y);
void ScreenGetChar(int *_ch, int *_attr, int _x, int _y);
void ScreenPutString(const char *_ch, int _attr, int _x, int _y);
void ScreenSetCursor(int _row, int _col);
void ScreenGetCursor(int *_row, int *_col);
void ScreenClear(void);
void ScreenUpdate(const void *_virtual_screen);
void ScreenUpdateLine(const void *_virtual_screen_line, int _row);
void ScreenRetrieve(void *_virtual_screen);
void ScreenVisualBell(void);







//#include <inlines/pc.h>
# 25 "../../../../include/dos.h" 2

extern int _8087;

int _detect_80387(void);

struct DWORDREGS {
  unsigned int edi;
  unsigned int esi;
  unsigned int ebp;
  unsigned int cflag;
  unsigned int ebx;
  unsigned int edx;
  unsigned int ecx;
  unsigned int eax;
  unsigned short eflags;
};

struct DWORDREGS_W {
  unsigned int di;
  unsigned int si;
  unsigned int bp;
  unsigned int cflag;
  unsigned int bx;
  unsigned int dx;
  unsigned int cx;
  unsigned int ax;
  unsigned short flags;
};

struct WORDREGS {
  unsigned short di, _upper_di;
  unsigned short si, _upper_si;
  unsigned short bp, _upper_bp;
  unsigned short cflag, _upper_cflag;
  unsigned short bx, _upper_bx;
  unsigned short dx, _upper_dx;
  unsigned short cx, _upper_cx;
  unsigned short ax, _upper_ax;
  unsigned short flags;
};

struct BYTEREGS {
  unsigned short di, _upper_di;
  unsigned short si, _upper_si;
  unsigned short bp, _upper_bp;
  unsigned int cflag;
  unsigned char bl;
  unsigned char bh;
  unsigned short _upper_bx;
  unsigned char dl;
  unsigned char dh;
  unsigned short _upper_dx;
  unsigned char cl;
  unsigned char ch;
  unsigned short _upper_cx;
  unsigned char al;
  unsigned char ah;
  unsigned short _upper_ax;
  unsigned short flags;
};

union REGS { /* Compatible with DPMI structure, except cflag */
  struct DWORDREGS d;






  struct DWORDREGS_W x;


  struct WORDREGS w;
  struct BYTEREGS h;
};

struct SREGS {
  unsigned short es;
  unsigned short ds;
  unsigned short fs;
  unsigned short gs;
  unsigned short cs;
  unsigned short ss;
};

struct ftime {
  unsigned ft_tsec:5; /* 0-29, double to get real seconds */
  unsigned ft_min:6; /* 0-59 */
  unsigned ft_hour:5; /* 0-23 */
  unsigned ft_day:5; /* 1-31 */
  unsigned ft_month:4; /* 1-12 */
  unsigned ft_year:7; /* since 1980 */
};

struct date {
  short da_year;
  char da_day;
  char da_mon;
};

struct time {
  unsigned char ti_min;
  unsigned char ti_hour;
  unsigned char ti_hund;
  unsigned char ti_sec;
};

struct dfree {
  unsigned df_avail;
  unsigned df_total;
  unsigned df_bsec;
  unsigned df_sclus;
};





extern unsigned short _osmajor, _osminor;
extern unsigned short _os_trueversion;
extern const char * _os_flavor;
extern int _doserrno;

unsigned short _get_dos_version(int);

int _get_fat_size(const int _drive);
int _get_fs_type(const int _drive, char *const _result_str);
int _is_cdrom_drive(const int _drive);
int _is_fat32(const int _drive);
int _is_ram_drive(const int _drive);
int _media_type(const int _drive);

int int86(int _ivec, union REGS *in, union REGS *out);
int int86x(int _ivec, union REGS *_in, union REGS *_out, struct SREGS *_seg);
int intdos(union REGS *_in, union REGS *_out);
int intdosx(union REGS *_in, union REGS *_out, struct SREGS *_seg);
int bdos(int _func, unsigned _dx, unsigned _al);
int bdosptr(int _func, void *_dx, unsigned _al);

int _int86(int ivec, union REGS *in, union REGS *out);





int enable(void);
int disable(void);

int getftime(int _handle, struct ftime *_ftimep);
int setftime(int _handle, struct ftime *_ftimep);

int getcbrk(void);
int setcbrk(int _new_value);

void getdate(struct date *);
void gettime(struct time *);
void setdate(struct date *);
void settime(struct time *);

void getdfree(unsigned char _drive, struct dfree *_ptr);

void delay(unsigned _msec);


/*
 *  For compatibility with other DOS C compilers.
 */
# 204 "../../../../include/dos.h"
struct _dosdate_t {
  unsigned char day; /* 1-31 */
  unsigned char month; /* 1-12 */
  unsigned short year; /* 1980-2099 */
  unsigned char dayofweek; /* 0-6, 0=Sunday */
};


struct _dostime_t {
  unsigned char hour; /* 0-23 */
  unsigned char minute; /* 0-59 */
  unsigned char second; /* 0-59 */
  unsigned char hsecond; /* 0-99 */
};


struct _find_t {
  char reserved[21];
  unsigned char attrib;
  unsigned short wr_time;
  unsigned short wr_date;
  unsigned int size;
  char name[256];
} __attribute__((packed));

/* make sure that structure packing is correct */
typedef int _DJCHK_FIND_T[(sizeof(struct _find_t)==286)*3 - 1];

struct _diskfree_t {
  unsigned short total_clusters;
  unsigned short avail_clusters;
  unsigned short sectors_per_cluster;
  unsigned short bytes_per_sector;
};


struct _DOSERROR {
  int exterror;
  char _class;
  char action;
  char locus;
};


struct _DOSERROR_STR {
  char *exterror_str;
  char *class_str;
  char *action_str;
  char *locus_str;
};


unsigned int _dos_creat(const char *_filename, unsigned int _attr, int *_handle);
unsigned int _dos_creatnew(const char *_filename, unsigned int _attr, int *_handle);
unsigned int _dos_open(const char *_filename, unsigned int _mode, int *_handle);
unsigned int _dos_write(int _handle, const void *_buffer, unsigned int _count, unsigned int *_result);
unsigned int _dos_read(int _handle, void *_buffer, unsigned int _count, unsigned int *_result);
unsigned int _dos_close(int _handle);
unsigned int _dos_commit(int _handle);

unsigned int _dos_findfirst(const char *_name, unsigned int _attr, struct _find_t *_result);
unsigned int _dos_findnext(struct _find_t *_result);

void _dos_getdate(struct _dosdate_t *_date);
unsigned int _dos_setdate(struct _dosdate_t *_date);
void _dos_gettime(struct _dostime_t *_time);
unsigned int _dos_settime(struct _dostime_t *_time);

unsigned int _dos_getftime(int _handle, unsigned int *_p_date, unsigned int *_p_time);
unsigned int _dos_setftime(int _handle, unsigned int _date, unsigned int _time);
unsigned int _dos_getfileattr(const char *_filename, unsigned int *_p_attr);
unsigned int _dos_setfileattr(const char *_filename, unsigned int _attr);

void _dos_getdrive(unsigned int *_p_drive);
void _dos_setdrive(unsigned int _drive, unsigned int *_p_drives);
unsigned int _dos_getdiskfree(unsigned int _drive, struct _diskfree_t *_diskspace);

int _dosexterr(struct _DOSERROR *_p_error);

int _dostrerr(struct _DOSERROR *_p_error, struct _DOSERROR_STR *_p_str);
# 154 "ioctl.c" 2
# 1 "../../../../include/libc/dosio.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../include/libc/dosio.h"
/* set to O_BINARY or O_TEXT */
extern char *__file_handle_modes;
extern void __file_handle_set(int fd, int mode);

int __doserr_to_errno(int doserr);

/* puts "path" in the transfer buffer, fixing
   unix-allowed multi-slashes */
int _put_path(const char *_path);
int _put_path2(const char *_path, int _offset);

/* Convenience functions for setting up transfers */




/* Functions used by write and _write.  */
int _write_fill_seek_gap(int fd);
int _write_int(int fd, const char *buffer, unsigned long long count);
# 155 "ioctl.c" 2

static int _dos_ioctl(int fd, int cmd, int argcx,int argdx,int argsi,int argdi,
                      int xarg)
{
/*
** Do an int0x21,0x44xx and such
*/
    __dpmi_regs r = {0};
    int dos_selector = 0;
    int dos_segment = 0;

    /*
    ** Lower word of cmd -> al
    */
    r.x.ax = 0x4400 |(cmd &0xff);
    r.x.bx = fd;
    r.x.cx = argcx;
    r.x.dx = argdx;
    r.x.si = argsi;
    r.x.di = argdi;
    /*
    ** This one must have dl=0
    */
    if(cmd == (0x01| ((1 & 3)<<12) /* How many extra args we expect.    */)) r.x.dx = argdx &0xf;
    /*
    ** Normally, there is a difference between reading and writing.
    ** But some functions require codes in the buffer on input already.
    ** So we transfer before int and after the int.
    ** We always use DX as a pointer to the buffer.
    ** I _do_ like clear APIs.
    */
    if(cmd & 0x8000 /* use xfer buffer                   */){
        if(argcx <= (int)_go32_info_block.size_of_transfer_buffer){ /* Can we use transfer buffer ? */
            dosmemput((void *)(uintptr_t)argdx,argcx, _go32_info_block.linear_address_of_transfer_buffer);
            r.x.ds = (_go32_info_block.linear_address_of_transfer_buffer>>4) &0xffff;
            r.x.dx = _go32_info_block.linear_address_of_transfer_buffer &0xf;
        }
        else{
            /* No, we have to get some DOS mem ourselves */
            dos_segment = __dpmi_allocate_dos_memory((argcx+15)>>4,
                                                     &dos_selector);
            r.x.ds = dos_segment;
            if(-1 == dos_segment){
                __dj_errno = 25;
                return -1;
            }
        }
    }
    /*
    ** At DOS 3.2+ MS decided to change the API to an a bit more
    ** abstruse way. CX now does not hold the number of bytes
    ** to r/w, but is now a function_number:categegory_code.
    ** The size of the XBuffer is determined by the device driver
    ** which can set it to whatever it likes.
    ** We do not have second sight, so you'll have to add this size as a
    ** parameter after all the registers and the buffer pointer.
    */
    if( cmd & 0x0400 /* CX does not hold the number of    */ ){
        if(xarg <= (int)_go32_info_block.size_of_transfer_buffer){ /* Can we use transfer buffer ? */
            dosmemput((void *)(uintptr_t)argdx,xarg, _go32_info_block.linear_address_of_transfer_buffer);
            r.x.ds = (_go32_info_block.linear_address_of_transfer_buffer>>4) &0xffff;
            r.x.dx = _go32_info_block.linear_address_of_transfer_buffer &0xf;
        }
        else{
            /* No, we have to get some DOS mem ourselves */
            dos_segment = __dpmi_allocate_dos_memory((xarg+15)>>4,
                                                      &dos_selector);
            r.x.ds = dos_segment;
            if(-1 == dos_segment){
                __dj_errno = 25;
                return -1;
            }
        }
    }
    /*
    ** Call DOS
    */
    if(-1 == __dpmi_int(0x21,&r)){
        if(dos_selector) __dpmi_free_dos_memory(dos_selector);
        __dj_errno = 14;
        return -1;
    }
    __dj_errno = 0 ;
    /*
    ** DOS error?
    */
    if(r.x.flags &1){
        if(dos_selector) __dpmi_free_dos_memory(dos_selector);
        __dj_errno = __doserr_to_errno(r.x.ax);
        return -1;
    }
    /*
    ** move back to our buffers and find the return value
    */
    if(cmd & 0x8000 /* use xfer buffer                   */){
        if(dos_selector){
            dosmemget(dos_segment<<4,argcx,(void*)(uintptr_t) argdx);
            __dpmi_free_dos_memory(dos_selector);
        }
        else
           dosmemget(_go32_info_block.linear_address_of_transfer_buffer,argcx,(void *)(uintptr_t) argdx);
    }
    if(cmd & 0x0400 /* CX does not hold the number of    */){
        if(dos_selector){
            dosmemget(dos_segment<<4,xarg,(void*)(uintptr_t) argdx);
            __dpmi_free_dos_memory(dos_selector);
        }
        else
           dosmemget(_go32_info_block.linear_address_of_transfer_buffer,xarg,(void *)(uintptr_t) argdx);
    }
    /*
    ** Return the requested value or 0.
    */
    switch(cmd & 0x7000 /* Here we put the bits to tell      */){
        case 0x1000 /* return AX as result */ : return r.x.ax;
        case 0x2000 /* return DX as result */ : return r.x.dx;
        case 0x3000 /* return DI SI as result */ : return (r.x.di << 16) | r.x.si;
        default : return 0;
    }
    /* NOTREACHED */
}


static int _unix_ioctl(int fd, int cmd, va_list args)
{
  __FSEXT_Function *func = __FSEXT_get_function(fd);
  if(func)
  {
    int rv;
    if (__FSEXT_func_wrapper(func, __FSEXT_ioctl, &rv, fd))
       return rv;
  }

  switch (cmd)
  {
    case (int)(0x40000000 /* copy out parameters */|((sizeof(struct winsize)&0x7f /* parameters must be < 128 bytes */)<<16)|('t'<<8)|104) /* get window size */:
    {
      struct winsize *win;

      win = __builtin_va_arg(args,struct winsize *);

      _farsetsel_gs(_go32_info_block.selector_for_linear_memory);
      win->ws_row = _farnspeekb_gs(0x0484) + 1;
      win->ws_col = _farnspeekw_gs(0x044a);
      win->ws_xpixel = 1;
      win->ws_ypixel = 1;
      return 0;
    }

    case (int)(0x80000000 /* copy in parameters */|((sizeof(struct winsize)&0x7f /* parameters must be < 128 bytes */)<<16)|('t'<<8)|103) /* set window size */: /* Do nothing implementation.  */
    {
      struct winsize *win;

      win = __builtin_va_arg(args,struct winsize *);

      _farsetsel_gs(_go32_info_block.selector_for_linear_memory);
      if (win->ws_row == _farnspeekb_gs(0x484) + 1
          && win->ws_col == _farnspeekw_gs(0x44a))
        return 0;
      break;
    }
  }

  /* All else fails.  */
  __dj_errno = 27;
  return -1;
}

/*
**
** The user callable entry point.
**
*/
int ioctl(int fd, int cmd, ...)
{
  int argcx,argdx,argsi,argdi;
  int narg,xarg;
  __FSEXT_Function *func = __FSEXT_get_function(fd);
  int rv;
  va_list args;

  /**
   ** see if this is a file system extension file
   **
   */
  if (func && __FSEXT_func_wrapper(func, __FSEXT_ioctl, &rv, fd))
    return rv;

  __builtin_va_start(args,cmd);

  if(((cmd) & 0xd0000000U))
  {
# 363 "ioctl.c"
     rv = _unix_ioctl(fd, cmd, args);
     __builtin_va_end(args);
     return rv;
  }
  /* Handle a DOS request */
  /* extract arguments */
  narg = (cmd >> 12) & 3;
  argdx=argsi=argdi=xarg=0;
  argcx = __builtin_va_arg(args,int);

  if (narg > 0) argdx = __builtin_va_arg(args,int);
  if (narg > 1) argdi = __builtin_va_arg(args,int);
  if (narg > 2) argsi = __builtin_va_arg(args,int);
  if (cmd & 0x0400 /* CX does not hold the number of    */) xarg = __builtin_va_arg(args,int);
  __builtin_va_end(args);

  return _dos_ioctl(fd,cmd,argcx,argdx,argsi,argdi,xarg);
}
