# 0 "fstat.c"
# 0 "<built-in>"
# 0 "<command-line>"
# 1 "fstat.c"
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
/* This is file FSTAT.C */
/*
 *   Almost a 100% U**X-compatible fstat() substitute.
 *
 * Usage:
 *
 *   That's easy: just put this in libc.a, and then call fstat() as usual.
 *
 * Rationale:
 *
 *   Many Unix-born programs make heavy use of fstat() library
 *   function to make decisions on files' equality, size, access
 *   attributes etc.  In the MS-DOS environment, many implementations
 *   of fstat() are crippled, because DOS makes it very hard to get to
 *   certain pieces of information about files and directories.  Thus
 *   porting a program to DOS is usually an exercise in #ifdef'ing.
 *   This implementation facilitates porting Unix programs to MS-DOS
 *   by providing an fstat() which is much more Unix-compatible than
 *   those of most DOS-based C compilers (e.g., Borland's).
 *   Specifically, the following issues are taken care of:
 *
 *      1. Mode bits are returned for the actual file, files are NOT
 *         reported read-only (as in Borland's library fstat()).
 *      2. Mode bits are set for all 3 groups (user, group, other).
 *      3. Device code (st_dev, st_rdev) is correctly reported (0 = 'A',
 *         1 = 'B' etc.).
 *      4. Character device names (such as /dev/con, lpt1, aux etc.) are
 *         treated as if they were on a special drive called `@:'
 *         (st_dev = -1).  The "character special" mode bit is set
 *         for these devices.
 *      5. The inode number (st_ino) is taken from the starting cluster
 *         number of the file.  If the cluster number is unavailable, it
 *         is invented using the file's name in a manner that minimizes
 *         the possibility of inventing an inode which already belongs
 *         to another file.  See below for details.
 *      6. Executable files are found based on files' extensions and
 *         magic numbers present at their beginning, and their execute
 *         bits are set.
 *
 *   Lossage:
 *
 *      Beautiful as the above sounds, this implementation does fail
 *      under certain circumstances.  The following is a list of known
 *      problems:
 *
 *      1. Files open on networked drives mounted by Novell Netware
 *         before revision 4.x cannot be traced using DOS System File
 *         Table.  Therefore, name, extension, file attributes and the
 *         drive letter are not available for these.  Until somebody
 *         tells me how this information can be obtained under Novell,
 *         nothing could be done here.  For the time being, these files
 *         will get st_dev of -2.
 *      2. For files which reside on networked drives, the inode number
 *         is always invented, because network redirectors usually do
 *         not bring that info with them.
 *      3. Empty files do not have a starting cluster number, because
 *         DOS doesn't allocate one until you actually write something
 *         to a file.  For these the inode is also invented.
 *      4. If the st_ino field is a 16 bit number, the invented inode
 *         numbers are from 65535 and down, assuming that most disks have
 *         unused portions near their end.  Valid cluster numbers are 16-bit
 *         unsigned integers, so a possibility of a clash exists, although
 *         the last 80 or more cluster numbers are unused on all drives
 *         I've seen.  If the st_ino is 32 bit, then invented inodes are
 *         all greater than 64k, which totally eliminates a possibility
 *         of a clash with an actual cluster number.
 *      5. As this implementation relies heavily on undocumented DOS
 *         features, it will fail to get actual file info in environments
 *         other than native DOS, such as DR-DOS, OS/2 etc.  For these,
 *         the function will return whatever info is available with
 *         conventional DOS calls, which is no less than any other
 *         implementation could do.  This fstat() might also fail for
 *         future DOS versions, if the layout of DOS System File Table
 *         is changed; however, this seems unlikely.
 *
 * Copyright (c) 1994-1996 Eli Zaretskii <eliz@is.elta.co.il>
 *
 * This software may be used freely so long as this copyright notice is
 * left intact.  There is no warranty on this software.
 *
 */

/*
 * Tested with DJGPP port of GNU C compiler, versions 1.11maint5 and 1.12m2,
 * under MS-DOS 3.3, 4.01, 5.0, 6.20 (with and without DoubleSpace) and
 * with networked drives under XFS 1.76, Novell Netware 3.22, and
 * TSoft NFS 0.24Beta.
 *
 */

# 1 "../../../../../include/libc/stubs.h" 1
/* Copyright (C) 2007 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 18 "../../../../../include/libc/stubs.h"
/* POSIX functions (for when compiling an ANSI function) */
# 89 "../../../../../include/libc/stubs.h"
# 1 "../../../../../include/stddef.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 14 "../../../../../include/stddef.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 23 "../../../../../include/sys/djtypes.h"
/* For GCC 3.00 or later we use its builtin va_list.                       */
# 34 "../../../../../include/sys/djtypes.h"
/* but remains a typedef in C */





/* Define internal type to be used for avoiding strict
   aliasing warnings */

typedef short __attribute__((__may_alias__)) __dj_short_a;
typedef int __attribute__((__may_alias__)) __dj_int_a;
typedef long __attribute__((__may_alias__)) __dj_long_a;
typedef long long __attribute__((__may_alias__)) __dj_long_long_a;
typedef unsigned short __attribute__((__may_alias__)) __dj_unsigned_short_a;
typedef unsigned int __attribute__((__may_alias__)) __dj_unsigned_int_a;
typedef unsigned long __attribute__((__may_alias__)) __dj_unsigned_long_a;
typedef unsigned long long __attribute__((__may_alias__)) __dj_unsigned_long_long_a;
typedef float __attribute__((__may_alias__)) __dj_float_a;
typedef double __attribute__((__may_alias__)) __dj_double_a;
typedef long double __attribute__((__may_alias__)) __dj_long_double_a;
# 15 "../../../../../include/stddef.h" 2

/* Some programs think they know better... */
# 26 "../../../../../include/stddef.h"
//#ifdef __cplusplus
//#define offsetof(s_type, mbr) ((std::size_t) &((s_type *)0)->mbr)
//#else

//#endif


typedef int ptrdiff_t;




typedef long unsigned int size_t;




typedef unsigned short wchar_t;
# 90 "../../../../../include/libc/stubs.h" 2
# 1 "../../../../../include/dpmi.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 40 "../../../../../include/sys/cdefs.h"
/* Support gcc's __attribute__ facility.  */




/* Ensure that always traditional GNU extern inline semantics are used
   (aka -fgnu89-inline) even if ISO C99 semantics have been specified.  */
# 9 "../../../../../include/dpmi.h" 2
# 1 "../../../../../include/sys/nearptr.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */




# 1 "../../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 17 "../../../../../include/stdint.h"
typedef signed char int_least8_t;
typedef unsigned char uint_least8_t;
typedef signed char int_fast8_t;
typedef unsigned char uint_fast8_t;
typedef signed char int8_t;
typedef unsigned char uint8_t;

typedef signed short int int_least16_t;
typedef unsigned short int uint_least16_t;
typedef signed int int_fast16_t;
typedef unsigned int uint_fast16_t;
typedef signed short int int16_t;
typedef unsigned short int uint16_t;

typedef signed int int_least32_t;
typedef unsigned int uint_least32_t;
typedef signed int int_fast32_t;
typedef unsigned int uint_fast32_t;
typedef signed int int32_t;
typedef unsigned int uint32_t;

__extension__ typedef signed long long int int_least64_t;
__extension__ typedef unsigned long long int uint_least64_t;
__extension__ typedef signed long long int int_fast64_t;
__extension__ typedef unsigned long long int uint_fast64_t;
__extension__ typedef signed long long int int64_t;
__extension__ typedef unsigned long long int uint64_t;

typedef long int intptr_t;
typedef unsigned long int uintptr_t;

__extension__ typedef signed long long int intmax_t;
__extension__ typedef unsigned long long int uintmax_t;

/* ANSI/ISO C99 says these should not be visible in C++ unless
   explicitly requested.  */

//#if !defined(__cplusplus) || defined(__STDC_LIMIT_MACROS)
# 110 "../../../../../include/stdint.h"
  /* These are defined by limits.h, so make them conditional.  */
# 124 "../../../../../include/stdint.h"
//#endif /* !__cplusplus || __STDC_LIMIT_MACROS */
# 23 "../../../../../include/libc/djthunks.h" 2

void crt1_startup(int handle);

uint8_t *djaddr2ptr(uint32_t addr);
uint8_t *djaddr2ptr2(uint32_t addr, uint32_t len);
uint32_t djptr2addr(const uint8_t *ptr);
uint32_t djthunk_get(const char *name);
uint32_t djthunk_get_h(int handle, const char *name);
void *djsbrk(int increment);
# 20 "../../../../../include/libc/asmobj.h" 2
# 8 "../../../../../include/sys/nearptr.h" 2
# 24 "../../../../../include/sys/nearptr.h"
/* Functions to enable "near" pointer access to DOS memory under DPMI
   CW Sandmann 7-95  NO WARRANTY: WARNING, since these functions disable
   memory protection, they MAY DESTROY EVERYTHING ON YOUR COMPUTER! */

int __djgpp_nearptr_enable(void); /* Returns 0 if feature not avail */
void __djgpp_nearptr_disable(void); /* Enables protection */
extern int *____djgpp_selector_limit(void);/*
*/
# 30 "../../../../../include/sys/nearptr.h"
       #define __djgpp_selector_limit (*____djgpp_selector_limit())/*
*/
# 30 "../../../../../include/sys/nearptr.h"
                                       ; /* Limit on CS and on DS if prot */
extern unsigned *____djgpp_base_address(void);/*
*/
# 31 "../../../../../include/sys/nearptr.h"
       #define __djgpp_base_address (*____djgpp_base_address())/*
*/
# 31 "../../../../../include/sys/nearptr.h"
                                          ; /* Used in calculation below */
# 10 "../../../../../include/dpmi.h" 2
# 1 "../../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../../include/libc/asmobj.h" 2
# 11 "../../../../../include/dpmi.h" 2
# 1 "../../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 12 "../../../../../include/dpmi.h" 2
# 26 "../../../../../include/dpmi.h"
//#ifndef _POSIX_SOURCE

extern unsigned short *____dpmi_error(void);/*
*/
# 28 "../../../../../include/dpmi.h"
       #define __dpmi_error (*____dpmi_error())/*
*/
# 28 "../../../../../include/dpmi.h"
                                        ;

typedef struct {
  unsigned short offset16;
  unsigned short segment;
} __dpmi_raddr;

typedef struct {
  unsigned int offset32;
  unsigned short selector;
} __dpmi_paddr;

typedef struct {
  unsigned int handle; /* 0, 2 */
  unsigned int size; /* or count */ /* 4, 6 */
  unsigned int address; /* 8, 10 */
} __dpmi_meminfo;

typedef union {
  struct {
    unsigned int edi;
    unsigned int esi;
    unsigned int ebp;
    unsigned int res;
    unsigned int ebx;
    unsigned int edx;
    unsigned int ecx;
    unsigned int eax;
  } d;
  struct {
    unsigned short di, di_hi;
    unsigned short si, si_hi;
    unsigned short bp, bp_hi;
    unsigned short res, res_hi;
    unsigned short bx, bx_hi;
    unsigned short dx, dx_hi;
    unsigned short cx, cx_hi;
    unsigned short ax, ax_hi;
    unsigned short flags;
    unsigned short es;
    unsigned short ds;
    unsigned short fs;
    unsigned short gs;
    unsigned short ip;
    unsigned short cs;
    unsigned short sp;
    unsigned short ss;
  } x;
  struct {
    unsigned char edi[4];
    unsigned char esi[4];
    unsigned char ebp[4];
    unsigned char res[4];
    unsigned char bl, bh, ebx_b2, ebx_b3;
    unsigned char dl, dh, edx_b2, edx_b3;
    unsigned char cl, ch, ecx_b2, ecx_b3;
    unsigned char al, ah, eax_b2, eax_b3;
  } h;
} __attribute__((packed)) __dpmi_regs;

typedef struct {
  unsigned char major;
  unsigned char minor;
  unsigned short flags;
  unsigned char cpu;
  unsigned char master_pic;
  unsigned char slave_pic;
} __dpmi_version_ret;

typedef struct {
  unsigned int largest_available_free_block_in_bytes;
  unsigned int maximum_unlocked_page_allocation_in_pages;
  unsigned int maximum_locked_page_allocation_in_pages;
  unsigned int linear_address_space_size_in_pages;
  unsigned int total_number_of_unlocked_pages;
  unsigned int total_number_of_free_pages;
  unsigned int total_number_of_physical_pages;
  unsigned int free_linear_address_space_in_pages;
  unsigned int size_of_paging_file_partition_in_pages;
  unsigned int reserved[3];
} __dpmi_free_mem_info;

typedef struct {
  unsigned int total_allocated_bytes_of_physical_memory_host;
  unsigned int total_allocated_bytes_of_virtual_memory_host;
  unsigned int total_available_bytes_of_virtual_memory_host;
  unsigned int total_allocated_bytes_of_virtual_memory_vcpu;
  unsigned int total_available_bytes_of_virtual_memory_vcpu;
  unsigned int total_allocated_bytes_of_virtual_memory_client;
  unsigned int total_available_bytes_of_virtual_memory_client;
  unsigned int total_locked_bytes_of_memory_client;
  unsigned int max_locked_bytes_of_memory_client;
  unsigned int highest_linear_address_available_to_client;
  unsigned int size_in_bytes_of_largest_free_memory_block;
  unsigned int size_of_minimum_allocation_unit_in_bytes;
  unsigned int size_of_allocation_alignment_unit_in_bytes;
  unsigned int reserved[19];
} __dpmi_memory_info;

typedef struct {
  unsigned int data16[2];
  unsigned int code16[2];
  unsigned short ip;
  unsigned short reserved;
  unsigned int data32[2];
  unsigned int code32[2];
  unsigned int eip;
} __dpmi_callback_info;

typedef struct {
  unsigned int size_requested;
  unsigned int size;
  unsigned int handle;
  unsigned int address;
  unsigned int name_offset;
  unsigned short name_selector;
  unsigned short reserved1;
  unsigned int reserved2;
} __dpmi_shminfo;

/* Unless otherwise noted, all functions return -1 on error, setting __dpmi_error to the DPMI error code */

void __dpmi_yield(void); /* INT 0x2F AX=1680 */

int __dpmi_allocate_ldt_descriptors(int _count); /* DPMI 0.9 AX=0000 */
int __dpmi_free_ldt_descriptor(int _descriptor); /* DPMI 0.9 AX=0001 */
int __dpmi_segment_to_descriptor(int _segment); /* DPMI 0.9 AX=0002 */
int __dpmi_get_selector_increment_value(void); /* DPMI 0.9 AX=0003 */
int __dpmi_get_segment_base_address(int _selector, unsigned int *_addr); /* DPMI 0.9 AX=0006 */
int __dpmi_set_segment_base_address(int _selector, unsigned int _address); /* DPMI 0.9 AX=0007 */
unsigned int __dpmi_get_segment_limit(int _selector); /* LSL instruction  */
int __dpmi_set_segment_limit(int _selector, unsigned int _limit); /* DPMI 0.9 AX=0008 */
int __dpmi_get_descriptor_access_rights(int _selector); /* LAR instruction  */
int __dpmi_set_descriptor_access_rights(int _selector, int _rights); /* DPMI 0.9 AX=0009 */
int __dpmi_create_alias_descriptor(int _selector); /* DPMI 0.9 AX=000a */
int __dpmi_get_descriptor(int _selector, unsigned char _buffer[8]); /* DPMI 0.9 AX=000b */
int __dpmi_set_descriptor(int _selector, unsigned char _buffer[8]); /* DPMI 0.9 AX=000c */
int __dpmi_allocate_specific_ldt_descriptor(int _selector); /* DPMI 0.9 AX=000d */


int __dpmi_get_multiple_descriptors(int _count, void *_buffer); /* DPMI 1.0 AX=000e */
int __dpmi_set_multiple_descriptors(int _count, void *_buffer); /* DPMI 1.0 AX=000f */

int __dpmi_allocate_dos_memory(int _paragraphs, int *_ret_selector_or_max); /* DPMI 0.9 AX=0100 */
int __dpmi_free_dos_memory(int _selector); /* DPMI 0.9 AX=0101 */
int __dpmi_resize_dos_memory(int _selector, int _newpara, int *_ret_max); /* DPMI 0.9 AX=0102 */

int __dpmi_get_real_mode_interrupt_vector(int _vector, __dpmi_raddr *_address); /* DPMI 0.9 AX=0200 */
int __dpmi_set_real_mode_interrupt_vector(int _vector, __dpmi_raddr *_address); /* DPMI 0.9 AX=0201 */
int __dpmi_get_processor_exception_handler_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0202 */
int __dpmi_set_processor_exception_handler_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0203 */
int __dpmi_get_protected_mode_interrupt_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0204 */
int __dpmi_set_protected_mode_interrupt_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0205 */

int __dpmi_get_extended_exception_handler_vector_pm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0210 */
int __dpmi_get_extended_exception_handler_vector_rm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0211 */
int __dpmi_set_extended_exception_handler_vector_pm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0212 */
int __dpmi_set_extended_exception_handler_vector_rm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0213 */

int __dpmi_simulate_real_mode_interrupt(int _vector, __dpmi_regs *_regs); /* DPMI 0.9 AX=0300 */
int __dpmi_int(int _vector, __dpmi_regs *_regs); /* like above, but sets ss sp fl */ /* DPMI 0.9 AX=0300 */
extern short __dpmi_int_ss, __dpmi_int_sp, __dpmi_int_flags; /* default to zero */
int __dpmi_simulate_real_mode_procedure_retf(__dpmi_regs *_regs); /* DPMI 0.9 AX=0301 */
int __dpmi_simulate_real_mode_procedure_retf_stack(__dpmi_regs *_regs, int stack_words_to_copy, const void *stack_data); /* DPMI 0.9 AX=0301 */
int __dpmi_simulate_real_mode_procedure_iret(__dpmi_regs *_regs); /* DPMI 0.9 AX=0302 */
int __dpmi_allocate_real_mode_callback(unsigned int _handler, unsigned int _regs, __dpmi_raddr *_ret); /* DPMI 0.9 AX=0303 */
int __dpmi_free_real_mode_callback(__dpmi_raddr *_addr); /* DPMI 0.9 AX=0304 */
int __dpmi_get_state_save_restore_addr( __dpmi_raddr *_rm, __dpmi_paddr *_pm); /* DPMI 0.9 AX=0305 */
int __dpmi_get_raw_mode_switch_addr( __dpmi_raddr *_rm, __dpmi_paddr *_pm); /* DPMI 0.9 AX=0306 */

int __dpmi_get_version( __dpmi_version_ret *_ret); /* DPMI 0.9 AX=0400 */

int __dpmi_get_capabilities( int *_flags, char vendor_info[128]); /* DPMI 1.0 AX=0401 */

int __dpmi_get_free_memory_information( __dpmi_free_mem_info *_info); /* DPMI 0.9 AX=0500 */
int __dpmi_allocate_memory(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0501 */
int __dpmi_free_memory(unsigned int _handle); /* DPMI 0.9 AX=0502 */
int __dpmi_resize_memory(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0503 */

int __dpmi_allocate_linear_memory(__dpmi_meminfo *_info, int _commit); /* DPMI 1.0 AX=0504 */
int __dpmi_resize_linear_memory(__dpmi_meminfo *_info, int _commit); /* DPMI 1.0 AX=0505 */
int __dpmi_get_page_attributes(__dpmi_meminfo *_info, short *_buffer); /* DPMI 1.0 AX=0506 */
int __dpmi_set_page_attributes(__dpmi_meminfo *_info, short *_buffer); /* DPMI 1.0 AX=0507 */
int __dpmi_map_device_in_memory_block(__dpmi_meminfo *_info, unsigned int _physaddr); /* DPMI 1.0 AX=0508 */
int __dpmi_map_conventional_memory_in_memory_block(__dpmi_meminfo *_info, unsigned int _linaddr); /* DPMI 1.0 AX=0509 */
int __dpmi_get_memory_block_size_and_base(__dpmi_meminfo *_info); /* DPMI 1.0 AX=050a */
int __dpmi_get_memory_information( __dpmi_memory_info *_buffer); /* DPMI 1.0 AX=050b */

int __dpmi_lock_linear_region(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0600 */
int __dpmi_unlock_linear_region(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0601 */
int __dpmi_mark_real_mode_region_as_pageable(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0602 */
int __dpmi_relock_real_mode_region(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0603 */
int __dpmi_get_page_size( unsigned int *_size); /* DPMI 0.9 AX=0604 */

int __dpmi_mark_page_as_demand_paging_candidate(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0702 */
int __dpmi_discard_page_contents(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0703 */

int __dpmi_physical_address_mapping(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0800 */
int __dpmi_free_physical_address_mapping(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0801 */

/* These next four functions return the old state */
int __dpmi_get_and_disable_virtual_interrupt_state(void); /* DPMI 0.9 AX=0900 */
int __dpmi_get_and_enable_virtual_interrupt_state(void); /* DPMI 0.9 AX=0901 */
int __dpmi_get_and_set_virtual_interrupt_state(int _old_state); /* DPMI 0.9 AH=09   */
int __dpmi_get_virtual_interrupt_state(void); /* DPMI 0.9 AX=0902 */

int __dpmi_get_vendor_specific_api_entry_point(const char *_id, __dpmi_paddr *_api); /* DPMI 0.9 AX=0a00 */

int __dpmi_set_debug_watchpoint(__dpmi_meminfo *_info, int _type); /* DPMI 0.9 AX=0b00 */
int __dpmi_clear_debug_watchpoint(unsigned int _handle); /* DPMI 0.9 AX=0b01 */
int __dpmi_get_state_of_debug_watchpoint(unsigned int _handle, int *_status); /* DPMI 0.9 AX=0b02 */
int __dpmi_reset_debug_watchpoint(unsigned int _handle); /* DPMI 0.9 AX=0b03 */

int __dpmi_install_resident_service_provider_callback(__dpmi_callback_info *_info); /* DPMI 1.0 AX=0c00 */
int __dpmi_terminate_and_stay_resident(int return_code, int paragraphs_to_keep); /* DPMI 1.0 AX=0c01 */

int __dpmi_allocate_shared_memory(__dpmi_shminfo *_info); /* DPMI 1.0 AX=0d00 */
int __dpmi_free_shared_memory(unsigned int _handle); /* DPMI 1.0 AX=0d01 */
int __dpmi_serialize_on_shared_memory(unsigned int _handle, int _flags); /* DPMI 1.0 AX=0d02 */
int __dpmi_free_serialization_on_shared_memory(unsigned int _handle, int _flags); /* DPMI 1.0 AX=0d03 */

int __dpmi_get_coprocessor_status(void); /* DPMI 1.0 AX=0e00 */
int __dpmi_set_coprocessor_emulation(int _flags); /* DPMI 1.0 AX=0e01 */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Backwards compatibility stuff						       */



typedef struct {
  unsigned int available_memory;
  unsigned int available_pages;
  unsigned int available_lockable_pages;
  unsigned int linear_space;
  unsigned int unlocked_pages;
  unsigned int available_physical_pages;
  unsigned int total_physical_pages;
  unsigned int free_linear_space;
  unsigned int max_pages_in_paging_file;
  unsigned int reserved[3];
} _go32_dpmi_meminfo;







typedef struct {
  unsigned int size;
  unsigned int pm_offset;
  unsigned short pm_selector;
  unsigned short rm_offset;
  unsigned short rm_segment;
} _go32_dpmi_seginfo;

/* returns zero if success, else dpmi error and info->size is max size */
int _go32_dpmi_allocate_dos_memory(_go32_dpmi_seginfo *info);
 /* set size to bytes/16, call, use rm_segment.  Do not
	   change anthing but size until the memory is freed.
	   If error, max size is returned in size as bytes/16. */
int _go32_dpmi_free_dos_memory(_go32_dpmi_seginfo *info);
 /* set new size to bytes/16, call.  If error, max size
	   is returned in size as bytes/16 */
int _go32_dpmi_resize_dos_memory(_go32_dpmi_seginfo *info);
 /* uses pm_selector to free memory */

/* These both use the rm_segment:rm_offset fields only */
int _go32_dpmi_get_real_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
int _go32_dpmi_set_real_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);

/* These do NOT wrap the function in pm_offset in an iret handler.
   You must provide an assembler interface yourself, or alloc one below.
   You may NOT LONGjmp out of an interrupt handler. */
int _go32_dpmi_get_protected_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
 /* puts vector in pm_selector:pm_offset. */
int _go32_dpmi_set_protected_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
 /* sets vector from pm_offset and pm_selector */

/********** HELPER FUNCTIONS **********/

int _go32_dpmi_chain_protected_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
 /* sets up wrapper that calls function in pm_offset, chaining to old
	   handler when it returns */

/* These generate assember IRET-style wrappers for functions and set up stack */
int _go32_dpmi_allocate_iret_wrapper(_go32_dpmi_seginfo *info);
 /* Put function ptr in pm_offset, call, returns wrapper entry in pm_offset. */
int _go32_dpmi_free_iret_wrapper(_go32_dpmi_seginfo *info);
 /* assumes pm_offset points to wrapper, frees it */

/* RMCB functions, automatically restructure the real-mode stack for the
   proper return type and set up correct PM stack.  The callback
   (info->pm_offset) is called as (*pmcb)(_go32_dpmi_registers *regs); */
int _go32_dpmi_allocate_real_mode_callback_retf(_go32_dpmi_seginfo *info, unsigned int regs);
 /* points callback at pm_offset, returns seg:ofs of callback addr
	   in rm_segment:rm_offset.  Do not change any fields until freed.
	   Interface is added to simulate far return */
int _go32_dpmi_allocate_real_mode_callback_iret(_go32_dpmi_seginfo *info, unsigned int regs);
 /* same, but simulates iret */
int _go32_dpmi_free_real_mode_callback(_go32_dpmi_seginfo *info);
 /* frees callback */

/* The following two variables may be used to change the default stack size
   for interrupts and rmcb wrappers to a user defined size from the default
   of 32Kbytes.  Each RMCB and chain/iret wrapper gets its own stack. */

extern unsigned int _go32_interrupt_stack_size;
extern unsigned int _go32_rmcb_stack_size;

/* Convenience functions, the return value is *bytes* */
unsigned int _go32_dpmi_remaining_physical_memory(void);
unsigned int _go32_dpmi_remaining_virtual_memory(void);

/* locks memory from a specified offset within the code/data selector */
int _go32_dpmi_lock_code( unsigned int _lockaddr, unsigned int _locksize);
int _go32_dpmi_lock_data( unsigned int _lockaddr, unsigned int _locksize);

int __djgpp_set_page_attributes(void *our_addr, unsigned int num_bytes,
           unsigned short attributes);
int __djgpp_map_physical_memory(void *our_addr, unsigned int num_bytes,
           uint64_t phys_addr);

//#endif /* !_POSIX_SOURCE */
# 91 "../../../../../include/libc/stubs.h" 2
# 1 "../../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../../include/libc/asmobj.h" 2
# 92 "../../../../../include/libc/stubs.h" 2
int __spawnve(int mode, const char *path, char *const argv[],
              char *const envp[]);
int __chdir (const char *mydirname);
char * __getcwd(char *buf, size_t size);
int __brk(void *_heaptop);
unsigned int __sbrk(int delta);
void *sbrk(int delta);
long double __modfl(long double _x, long double *_pint);

extern unsigned *____djgpp_timer_countdown(void);/*
*/
# 101 "../../../../../include/libc/stubs.h"
       #define __djgpp_timer_countdown (*____djgpp_timer_countdown())/*
*/
# 101 "../../../../../include/libc/stubs.h"
                                             ;
extern int *____djgpp_timer_hdlr(void);/*
*/
# 102 "../../../../../include/libc/stubs.h"
       #define __djgpp_timer_hdlr (*____djgpp_timer_hdlr())/*
*/
# 102 "../../../../../include/libc/stubs.h"
                                   ;
# 102 "fstat.c" 2
# 1 "../../../../../include/stdarg.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 18 "../../../../../include/stdarg.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 19 "../../../../../include/stdarg.h" 2


typedef __builtin_va_list va_list;
# 103 "fstat.c" 2
# 1 "../../../../../include/string.h" 1
/* Copyright (C) 2017 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 17 "../../../../../include/string.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 18 "../../../../../include/string.h" 2

/* Some programs think they know better... */
# 34 "../../../../../include/string.h"
void * memchr(const void *_s, int _c, size_t _n);
int memcmp(const void *_s1, const void *_s2, size_t _n);
void * memcpy(void * _dest, const void * _src, size_t _n);
void * memmove(void *_s1, const void *_s2, size_t _n);
void * memset(void *_s, int _c, size_t _n);
char * strcat(char * _s1, const char * _s2);
char * strchr(const char *_s, int _c);
int strcmp(const char *_s1, const char *_s2);
int strcoll(const char *_s1, const char *_s2);
char * strcpy(char * _s1, const char * _s2);
size_t strcspn(const char *_s1, const char *_s2);
char * strerror(int _errcode);
size_t strlen(const char *_s);
char * strncat(char * _s1, const char * _s2, size_t _n);
int strncmp(const char *_s1, const char *_s2, size_t _n);
char * strncpy(char * _s1, const char * _s2, size_t _n);
char * strpbrk(const char *_s1, const char *_s2);
char * strrchr(const char *_s, int _c);
size_t strspn(const char *_s1, const char *_s2);
char * strstr(const char *_s1, const char *_s2);
char * strtok(char * _s1, const char * _s2);
size_t strxfrm(char * _s1, const char * _s2, size_t _n);
# 64 "../../../../../include/string.h"
int strerror_r(int _errnum, char *_strerrbuf, size_t _buflen);
char * strtok_r(char * _s1, const char * _s2, char ** _s3);



# 1 "../../../../../include/sys/movedata.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../../include/sys/movedata.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../../include/sys/movedata.h" 2
# 1 "../../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../../include/sys/movedata.h" 2

# 1 "../../../../../include/sys/fmemcpy.h" 1



# 1 "../../../../../include/dpmi.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 5 "../../../../../include/sys/fmemcpy.h" 2

void fmemcpy1(__dpmi_paddr dst, const void *src, unsigned len);
void fmemcpy2(void *dst, __dpmi_paddr src, unsigned len);
void fmemcpy12(__dpmi_paddr dst, __dpmi_paddr src, unsigned len);
# 28 "../../../../../include/sys/movedata.h" 2
# 36 "../../../../../include/sys/movedata.h"
/* This header is intended to be included only by other
   headers, like <go32.h> and <string.h>.  You may
   include this directly, but it will be non-portable. */

/* These lengths are in bytes, optimized for speed */
void dosmemget(unsigned long _offset, size_t _length, void *_buffer);
void dosmemput(const void *_buffer, size_t _length, unsigned long _offset);

/* The lengths here are in TRANSFERS, not bytes! */
void _dosmemgetb(unsigned long _offset, size_t _xfers, void *_buffer);
void _dosmemgetw(unsigned long _offset, size_t _xfers, void *_buffer);
void _dosmemgetl(unsigned long _offset, size_t _xfers, void *_buffer);
void _dosmemputb(const void *_buffer, size_t _xfers, unsigned long _offset);
void _dosmemputw(const void *_buffer, size_t _xfers, unsigned long _offset);
void _dosmemputl(const void *_buffer, size_t _xfers, unsigned long _offset);
# 70 "../../../../../include/string.h" 2

int bcmp(const void *_ptr1, const void *_ptr2, size_t _length);
void bcopy(const void *_a, void *_b, size_t _len);
void bzero(void *ptr, size_t _len);
int ffs(int _mask);
char * index(const char *_string, int _c);
void * memccpy(void *_to, const void *_from, int _c, size_t _n);
int memicmp(const void *_s1, const void *_s2, size_t _n);
void * memmem(const void *l, size_t l_len, const void *s, size_t s_len);
char * rindex(const char *_string, int _c);
char * stpcpy(char *_dest, const char *_src);
char * stpncpy(char *_dest, const char *_src, size_t _n);
char * strdup(const char *_s);
char * strndup(const char *_s, size_t _n);
size_t strlcat(char *_dest, const char *_src, size_t _size);
size_t strlcpy(char *_dest, const char *_src, size_t _size);
char * strlwr(char *_s);
int strcasecmp(const char *_s1, const char *_s2);
int stricmp(const char *_s1, const char *_s2);
int strncasecmp(const char *_s1, const char *_s2, size_t _n);
int strnicmp(const char *_s1, const char *_s2, size_t _n);
size_t strnlen(const char *_s, size_t _n);
char * strsep(char **_stringp, const char *_delim);
char * strupr(char *_s);
# 104 "fstat.c" 2
# 1 "../../../../../include/stdlib.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 20 "../../../../../include/stdlib.h"
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../../include/sys/types.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../../include/sys/types.h" 2

typedef int blkcnt_t;
typedef int blksize_t;
typedef int dev_t;
typedef unsigned long fsblkcnt_t;
typedef unsigned long fsfilcnt_t;
typedef int ino_t;
typedef int mode_t;
typedef int nlink_t;


typedef int gid_t;



typedef int off_t;



__extension__ typedef long long off64_t;



__extension__ typedef long long offset_t;



typedef int pid_t;







typedef long signed int ssize_t;




typedef int uid_t;




/* Allow including program to override.  */




typedef struct fd_set {
  unsigned char fd_bits [((256) + 7) / 8];
} fd_set;







typedef unsigned int time_t;
# 21 "../../../../../include/stdlib.h" 2
# 1 "../../../../../include/stddef.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 22 "../../../../../include/stdlib.h" 2

/* Some programs think they know better... */
# 38 "../../../../../include/stdlib.h"
extern int __dj_mb_cur_max;

typedef struct {
  int quot;
  int rem;
} div_t;

typedef struct {
  long quot;
  long rem;
} ldiv_t;





void abort(void);
int abs(int _i);

int __dj_atexit(void (*_func)(void));
double atof(const char *_s);
int atoi(const char *_s);
long atol(const char *_s);
void * bsearch(const void *_key, const void *_base, size_t _nelem,
  size_t _size, int (*_cmp)(const void *_ck, const void *_ce));
void * calloc(size_t _nelem, size_t _size);
unsigned int calloc32(size_t _nelem, size_t _size);
div_t div(int _numer, int _denom);
void exit(int _status);
void free(void *_ptr);
void free32(unsigned int _addr);
char * getenv(const char *_name);
long labs(long _i);
ldiv_t ldiv(long _numer, long _denom);
void * malloc(size_t _size);
unsigned int malloc32(size_t _size);
int mblen(const char *_s, size_t _n);
size_t mbstowcs(wchar_t *_wcs, const char *_s, size_t _n);
int mbtowc(wchar_t *_pwc, const char *_s, size_t _n);
void qsort(void *_base, size_t _nelem, size_t _size,
       int (*_cmp)(const void *_e1, const void *_e2));
int rand(void);
void * realloc(void *_ptr, size_t _size);
unsigned int realloc32(unsigned int _addr, size_t _size);
void srand(unsigned _seed);
double strtod(const char *_s, char **_endptr);
long strtol(const char *_s, char **_endptr, int _base);
unsigned long strtoul(const char *_s, char **_endptr, int _base);
int system(const char *_s);
size_t wcstombs(char *_s, const wchar_t *_wcs, size_t _n);
int wctomb(char *_s, wchar_t _wchar);




typedef struct {
  long long int quot;
  long long int rem;
} lldiv_t;

void _Exit(int _status);
long long int atoll(const char *_s);
long long int llabs(long long int _i);
lldiv_t lldiv(long long int _numer, long long int _denom);
float strtof(const char *_s, char **_endptr);
long double strtold(const char *_s, char **_endptr);
long long int strtoll(const char *_s, char **_endptr, int _base);
unsigned long long int strtoull(const char *_s, char **_endptr, int _base);





long a64l(const char *_string);
char * l64a(long _value);
char * mktemp(char *_template);
char * mkdtemp(char *_template);
int mkstemp(char *_template);
int putenv(char *_val);
char * realpath(const char *_path, char *_resolved);
int setenv(const char *_var, const char *_val, int _overwrite);
int unsetenv(const char *_var);



void * alloca(size_t _size);
long double _atold(const char *_s);
void cfree(void *_ptr);
double drand48(void);
char * ecvtbuf(double _val, int _nd, int *_dp, int *_sn, char *_bf);
char * ecvt(double _val, int _nd, int *_dp, int *_sn);
double erand48(unsigned short _state[3]);
char * fcvtbuf(double _val, int _nd, int *_dp, int *_sn, char *_bf);
char * fcvt(double _val, int _nd, int *_dp, int *_sn);
char * gcvt(double _val, int _nd, char *_buf);
char * getpass(const char *_prompt);
int getlongpass(const char *_prompt, char *_buffer, int _max_len);
char * itoa(int _value, char *_buffer, int _radix);
long jrand48(unsigned short _state[3]);
void lcong48(unsigned short _param[7]);
unsigned long lrand48(void);
long mrand48(void);
unsigned long nrand48(unsigned short _state[3]);
unsigned short *seed48(unsigned short _state_seed[3]);
void srand48(long _seedval);
long double _strtold(const char *_s, char **_endptr);
void swab(const void *_from, void *_to, ssize_t _nbytes);
void * valloc (size_t _amt);





/* BSD Random Number Generator */
char * initstate (unsigned _seed, char *_arg_state, int _n);
char * setstate(char *_arg_state);
long random(void);
int srandom(int _seed);

/* These vary in expected prototype, so we just don't prototype them.
void		xfree(void *_ptr);
void *		xmalloc(size_t _size);
void *		xrealloc(void *ptr, size_t _size);
*/
# 173 "../../../../../include/stdlib.h"
extern int __system_flags;
# 105 "fstat.c" 2
# 1 "../../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 106 "fstat.c" 2
# 1 "../../../../../include/stdio.h" 1
/* Copyright (C) 2018 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2017 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 21 "../../../../../include/stdio.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 22 "../../../../../include/stdio.h" 2





/* Some programs think they know better... */
# 63 "../../../../../include/stdio.h"
/* Note that the definitions of these fields are NOT guaranteed!  They
   may change with any release without notice!  The fact that they
   are here at all is to comply with ANSI specifictions. */

typedef struct {
  ssize_t _cnt;
  char *_ptr;
  char *_base;
  size_t _bufsiz;
  int _flag;
  int _file;
  char *_name_to_remove;
  size_t _fillsize;
} FILE;

typedef unsigned long fpos_t;

extern FILE __dj_stdin, __dj_stdout, __dj_stderr;




void clearerr(FILE *_stream);
int fclose(FILE *_stream);
int feof(FILE *_stream);
int ferror(FILE *_stream);
int fflush(FILE *_stream);
int fgetc(FILE *_stream);
int fgetpos(FILE *_stream, fpos_t *_pos);
char * fgets(char *_s, int _n, FILE *_stream);
FILE * fopen(const char *_filename, const char *_mode);
int fprintf(FILE *_stream, const char *_format, ...);
int fputc(int _c, FILE *_stream);
int fputs(const char *_s, FILE *_stream);
size_t fread(void *_ptr, size_t _size, size_t _nelem, FILE *_stream);
FILE * freopen(const char *_filename, const char *_mode, FILE *_stream);
int fscanf(FILE *_stream, const char *_format, ...);
int fseek(FILE *_stream, long _offset, int _mode);
int fsetpos(FILE *_stream, const fpos_t *_pos);
long ftell(FILE *_stream);
size_t fwrite(const void *_ptr, size_t _size, size_t _nelem, FILE *_stream);
int getc(FILE *_stream);
int getchar(void);
char * gets(char *_s);
void perror(const char *_s);
int printf(const char *_format, ...);
int putc(int _c, FILE *_stream);
int putchar(int _c);
int puts(const char *_s);
int remove(const char *_filename);
int rename(const char *_old, const char *_new);
void rewind(FILE *_stream);
int scanf(const char *_format, ...);
void setbuf(FILE *_stream, char *_buf);
int setvbuf(FILE *_stream, char *_buf, int _mode, size_t _size);
int sprintf(char *_s, const char *_format, ...);
int sscanf(const char *_s, const char *_format, ...);
FILE * tmpfile(void);
char * tmpnam(char *_s);
int ungetc(int _c, FILE *_stream);
int vfprintf(FILE *_stream, const char *_format, va_list _ap);
int vprintf(const char *_format, va_list _ap);
int vsprintf(char *_s, const char *_format, va_list _ap);




int snprintf(char *str, size_t n, const char *fmt, ...);
int vfscanf(FILE *_stream, const char *_format, va_list _ap);
int vscanf(const char *_format, va_list _ap);
int vsnprintf(char *str, size_t n, const char *fmt, va_list ap);
int vsscanf(const char *_s, const char *_format, va_list _ap);







/* #define STREAM_MAX	20 - DOS can change this */

int dprintf(int _fd, const char *_format, ...) __attribute__ ((__format__ (__printf__, 2, 3)));
int fileno(FILE *_stream);
FILE * fdopen(int _fildes, const char *_type);
int mkstemp(char *_template);
int pclose(FILE *_pf);
FILE * popen(const char *_command, const char *_mode);
char * tempnam(const char *_dir, const char *_prefix);
int vdprintf(int _fd, const char *_format, va_list _ap) __attribute__ ((__format__ (__printf__, 2, 0)));



extern FILE __dj_stdprn, __dj_stdaux;





void _djstat_describe_lossage(FILE *_to_where);
int _doprnt(const char *_fmt, va_list _args, FILE *_f);
int _doscan(FILE *_f, const char *_fmt, va_list _args);
int _doscan_low(FILE *, int (*)(FILE *_get), int (*_unget)(int, FILE *), const char *_fmt, va_list _args);
int fpurge(FILE *_f);
int getw(FILE *_f);
char * mktemp(char *_template);
int putw(int _v, FILE *_f);
void setbuffer(FILE *_f, void *_buf, int _size);
void setlinebuf(FILE *_f);
int _rename(const char *_old, const char *_new); /* Simple (no directory) */
int asprintf(char **_sp, const char *_format, ...) __attribute__((format (__printf__, 2, 3)));
char * asnprintf(char *_s, size_t *_np, const char *_format, ...) __attribute__((format (__printf__, 3, 4)));
int vasprintf(char **_sp, const char *_format, va_list _ap) __attribute__((format (__printf__, 2, 0)));
char * vasnprintf(char *_s, size_t *_np, const char *_format, va_list _ap) __attribute__((format (__printf__, 3, 0)));
# 185 "../../../../../include/stdio.h"
int fseeko(FILE *_stream, off_t _offset, int _mode);
off_t ftello(FILE *_stream);
int fseeko64(FILE *_stream, off64_t _offset, int _mode);
off64_t ftello64(FILE *_stream);
# 107 "fstat.c" 2
# 1 "../../../../../include/errno.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */







# 1 "../../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../../include/libc/asmobj.h" 2
# 16 "../../../../../include/errno.h" 2







extern int *____dj_errno(void);/*
*/
# 23 "../../../../../include/errno.h"
       #define __dj_errno (*____dj_errno())/*
*/
# 23 "../../../../../include/errno.h"
                             ;
# 77 "../../../../../include/errno.h"
/* EILSEQ		41 */

extern char * sys_errlist[];
extern int sys_nerr;
extern const char * __sys_errlist[];
extern int __sys_nerr;
extern int _doserrno;
# 108 "fstat.c" 2
# 1 "../../../../../include/io.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../../include/io.h"
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../../include/io.h" 2

int chsize(int handle, long size);
int _close(int _fd);
int _creat(const char *_path, int _attrib);
int _creatnew(const char *_path, int _attrib, int _mode);
ssize_t crlf2nl(char *_buffer, ssize_t _length);
int _dos_lock(int _fd, long _offset, long _length);
int _dos_lk64(int _fd, long long _offset, long long _length);
long filelength(int _handle);
short _get_dev_info(int _arg);
long long lfilelength(int _handle);
int lock(int _fd, long _offset, long _length);
int lock64(int _fd, long long _offset, long long _length);
int _open(const char *_path, int _oflag);
ssize_t _read(int _fd, void *_buf, size_t _nbyte);
int setmode(int _fd, int _newmode);
off_t tell(int _fd);
int _dos_unlock(int _fd, long _offset, long _length);
int _dos_unlk64(int _fd, long long _offset, long long _length);
int unlock(int _fd, long _offset, long _length);
int unlock64(int _fd, long long _offset, long long _length);
ssize_t _write(int _fd, const void *_buf, size_t _nbyte);
int _chmod(const char *_path, int _func, ...);
void _flush_disk_cache(void);
# 109 "fstat.c" 2
# 1 "../../../../../include/time.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 18 "../../../../../include/time.h"
/* 65536(tics/hour) / 3600(sec/hour) * 5(scale) = 91.02
   The 5 is to make it a whole number (18.2*5=91) so that
   floating point ops aren't required to use it. */


# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../../include/time.h" 2

/* Some programs think they know better... */
# 36 "../../../../../include/time.h"
typedef int clock_t;
# 48 "../../../../../include/time.h"
struct tm {
  int tm_sec;
  int tm_min;
  int tm_hour;
  int tm_mday;
  int tm_mon;
  int tm_year;
  int tm_wday;
  int tm_yday;
  int tm_isdst;
  char *__tm_zone;
  int __tm_gmtoff;
};

char * asctime(const struct tm *_tptr);
clock_t clock(void);
char * ctime(const time_t *_cal);
double difftime(time_t _t1, time_t _t0);
struct tm * gmtime(const time_t *_tod);
struct tm * localtime(const time_t *_tod);
time_t mktime(struct tm *_tptr);
size_t strftime(char * _s, size_t _n, const char * _format, const struct tm * _tptr);
time_t time(time_t *_tod);
# 81 "../../../../../include/time.h"
extern char *tzname[2];

char * asctime_r(const struct tm * __restrict__ _tptr, char * __restrict__ _buf);
char * ctime_r(const time_t *_cal, char *_buf);
struct tm * gmtime_r(const time_t * __restrict__ _tod, struct tm * __restrict__ _tptr);
struct tm * localtime_r(const time_t * __restrict__ _tod, struct tm * __restrict__ _tptr);
void tzset(void);






struct timeval {
  time_t tv_sec;
  long tv_usec;
};

struct timezone {
  int tz_minuteswest;
  int tz_dsttime;
};

# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 105 "../../../../../include/time.h" 2

typedef long long uclock_t;


int gettimeofday(struct timeval *_tp, struct timezone *_tzp);
unsigned long rawclock(void);
int select(int _nfds, fd_set *_readfds, fd_set *_writefds, fd_set *_exceptfds, struct timeval *_timeout);
int settimeofday(struct timeval *_tp, ...);
void tzsetwall(void);
uclock_t uclock(void);


/* GCC-4.8.0 and newer versions defines _rdtsc as macro in ia32intrin.h
   included from x86intrin.h (for MMX/SSE/AVX instructions. */



# 1 "../../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 123 "../../../../../include/time.h" 2

unsigned long long _rdtsc(void);
# 110 "fstat.c" 2
# 1 "../../../../../include/unistd.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 19 "../../../../../include/unistd.h"
extern char **__dj_environ;
# 30 "../../../../../include/unistd.h"
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 31 "../../../../../include/unistd.h" 2
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 32 "../../../../../include/unistd.h" 2





/* `lockf' is a simpler interface to the locking facilities of `fcntl'.
   LEN is always relative to the current file position.
   The CMD argument is one of the following.  */






/* Some programs think they know better... */
# 112 "../../../../../include/unistd.h"
extern char *optarg;
extern int optind, opterr, optopt;

void __exit(int _status);
void _exit(int _status);
int access(const char *_path, int _amode);
unsigned int alarm(unsigned int _seconds);
int chdir(const char *_path);
int chown(const char *_path, uid_t _owner, gid_t _group);
int close(int _fildes);
size_t confstr(int _name, char *_buf, size_t _len);
char * ctermid(char *_s);
int dup(int _fildes);
int dup2(int _fildes, int _fildes2);
int execl(const char *_path, const char *_arg, ...);
int execle(const char *_path, const char *_arg, ...);
int execlp(const char *_file, const char *_arg, ...);
int execv(const char *_path, char *const _argv[]);
int execve(const char *_path, char *const _argv[], char *const _envp[]);
int execvp(const char *_file, char *const _argv[]);
int fchdir(int _fd);
pid_t fork(void);
long fpathconf(int _fildes, int _name);
char * getcwd(char *_buf, size_t _size);
gid_t getegid(void);
uid_t geteuid(void);
gid_t getgid(void);
int getgroups(int _gidsetsize, gid_t *_grouplist);
char * getlogin(void);
int getopt(int _argc, char *const _argv[], const char *_optstring);
pid_t getpgrp(void);
pid_t getpid(void);
pid_t getppid(void);
uid_t getuid(void);
int isatty(int _fildes);
int link(const char *_existing, const char *_new);
off_t lseek(int _fildes, off_t _offset, int _whence);
long pathconf(const char *_path, int _name);
int pause(void);
int pipe(int _fildes[2]);
ssize_t pwrite(int _fildes, const void *_buf, size_t _nbyte, off_t _offset);
ssize_t read(int _fildes, void *_buf, size_t _nbyte);
int rmdir(const char *_path);
int setgid(gid_t _gid);
int setpgid(pid_t _pid, pid_t _pgid);
pid_t setsid(void);
int setuid(uid_t _uid);
unsigned int sleep(unsigned int _seconds);
long sysconf(int _name);
pid_t tcgetpgrp(int _fildes);
int tcsetpgrp(int _fildes, pid_t _pgrp_id);
char * ttyname(int _fildes);
int unlink(const char *_path);
ssize_t write(int _fildes, const void *_buf, size_t _nbyte);



/* additional access() checks */


char * basename(const char *_fn);
int brk(void *_heaptop);
char * dirname(const char *_fn);
int __file_exists(const char *_fn);
int fchown(int fd, uid_t owner, gid_t group);
int fsync(int _fd);
int ftruncate(int, off_t);
int getdtablesize(void);
int gethostname(char *buf, int size);
int getpagesize(void);
char * getwd(char *__buffer);
int lchown(const char * file, int owner, int group);
int lockf(int _fildes, int _cmd, off_t _len);
int llockf(int _fildes, int _cmd, offset_t _len);
offset_t llseek(int _fildes, offset_t _offset, int _whence);
int nice(int _increment);
int readlink(const char * __file, char * __buffer, size_t __size);
void * sbrk(int _delta);
int symlink (const char *, const char *);
int sync(void);
int truncate(const char*, off_t);
unsigned int usleep(unsigned int _useconds);

pid_t vfork(void);
# 111 "fstat.c" 2
# 1 "../../../../../include/fcntl.h" 1
/* Copyright (C) 2016 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2009 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 60 "../../../../../include/fcntl.h"
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 61 "../../../../../include/fcntl.h" 2

struct flock {
  off_t l_len;
  pid_t l_pid;
  off_t l_start;
  short l_type;
  short l_whence;
};

struct flock64 {
  offset_t l_len;
  pid_t l_pid;
  offset_t l_start;
  short l_type;
  short l_whence;
};

extern int _fmode; /* O_TEXT or O_BINARY */

int open(const char *_path, int _oflag, ...);
int creat(const char *_path, mode_t _mode);
int fcntl(int _fildes, int _cmd, ...);



/* Additional non-POSIX flags for open(). */
/* They are present on GNU libc. */
# 105 "../../../../../include/fcntl.h"
extern int __djgpp_share_flags;





/*
 *  For compatibility with other DOS C compilers.
 */
# 127 "../../../../../include/fcntl.h"
/*
 * Support for advanced filesystems (Windows 9x VFAT, NTFS, LFN etc.)
 */
# 138 "../../../../../include/fcntl.h"
unsigned _get_volume_info (const char *_path, int *_max_file_len, int *_max_path_len, char *_filesystype);
char _use_lfn (const char *_path);
char *_lfn_gen_short_fname (const char *_long_fname, char *_short_fname);
int _is_DOS83 (const char *_fname);




unsigned _lfn_get_ftime (int _handle, int _which);

char _preserve_fncase (void);
# 112 "fstat.c" 2
# 1 "../../../../../include/dos.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 12 "../../../../../include/dos.h" 2
# 24 "../../../../../include/dos.h"
# 1 "../../../../../include/pc.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */







# 1 "../../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 15 "../../../../../include/pc.h" 2
# 27 "../../../../../include/pc.h"
unsigned char inportb (unsigned short _port);
unsigned short inportw (unsigned short _port);
uint32_t inportl (unsigned short _port);
void inportsb (unsigned short _port, unsigned char *_buf, unsigned _len);
void inportsw (unsigned short _port, unsigned short *_buf, unsigned _len);
void inportsl (unsigned short _port, uint32_t *_buf, unsigned _len);
void outportb (unsigned short _port, unsigned char _data);
void outportw (unsigned short _port, unsigned short _data);
void outportl (unsigned short _port, uint32_t _data);
void outportsb(unsigned short _port, const unsigned char *_buf, unsigned _len);
void outportsw(unsigned short _port, const unsigned short *_buf, unsigned _len);
void outportsl(unsigned short _port, const uint32_t *_buf, unsigned _len);

unsigned char inp(unsigned short _port);
unsigned short inpw(unsigned short _port);
void outp(unsigned short _port, unsigned char _data);
void outpw(unsigned short _port, unsigned short _data);

int kbhit(void);

int getkey(void); /* ALT's have 0x100 set */
int getxkey(void); /* ALT's have 0x100 set, 0xe0 sets 0x200 */
const unsigned char * __get_extended_key_string(int _xkey_code);
void sound(int _frequency);



extern unsigned char ScreenAttrib;




int ScreenMode(void);
int ScreenRows(void);
int ScreenCols(void);
void ScreenPutChar(int _ch, int _attr, int _x, int _y);
void ScreenGetChar(int *_ch, int *_attr, int _x, int _y);
void ScreenPutString(const char *_ch, int _attr, int _x, int _y);
void ScreenSetCursor(int _row, int _col);
void ScreenGetCursor(int *_row, int *_col);
void ScreenClear(void);
void ScreenUpdate(const void *_virtual_screen);
void ScreenUpdateLine(const void *_virtual_screen_line, int _row);
void ScreenRetrieve(void *_virtual_screen);
void ScreenVisualBell(void);







//#include <inlines/pc.h>
# 25 "../../../../../include/dos.h" 2

extern int _8087;

int _detect_80387(void);

struct DWORDREGS {
  unsigned int edi;
  unsigned int esi;
  unsigned int ebp;
  unsigned int cflag;
  unsigned int ebx;
  unsigned int edx;
  unsigned int ecx;
  unsigned int eax;
  unsigned short eflags;
};

struct DWORDREGS_W {
  unsigned int di;
  unsigned int si;
  unsigned int bp;
  unsigned int cflag;
  unsigned int bx;
  unsigned int dx;
  unsigned int cx;
  unsigned int ax;
  unsigned short flags;
};

struct WORDREGS {
  unsigned short di, _upper_di;
  unsigned short si, _upper_si;
  unsigned short bp, _upper_bp;
  unsigned short cflag, _upper_cflag;
  unsigned short bx, _upper_bx;
  unsigned short dx, _upper_dx;
  unsigned short cx, _upper_cx;
  unsigned short ax, _upper_ax;
  unsigned short flags;
};

struct BYTEREGS {
  unsigned short di, _upper_di;
  unsigned short si, _upper_si;
  unsigned short bp, _upper_bp;
  unsigned int cflag;
  unsigned char bl;
  unsigned char bh;
  unsigned short _upper_bx;
  unsigned char dl;
  unsigned char dh;
  unsigned short _upper_dx;
  unsigned char cl;
  unsigned char ch;
  unsigned short _upper_cx;
  unsigned char al;
  unsigned char ah;
  unsigned short _upper_ax;
  unsigned short flags;
};

union REGS { /* Compatible with DPMI structure, except cflag */
  struct DWORDREGS d;






  struct DWORDREGS_W x;


  struct WORDREGS w;
  struct BYTEREGS h;
};

struct SREGS {
  unsigned short es;
  unsigned short ds;
  unsigned short fs;
  unsigned short gs;
  unsigned short cs;
  unsigned short ss;
};

struct ftime {
  unsigned ft_tsec:5; /* 0-29, double to get real seconds */
  unsigned ft_min:6; /* 0-59 */
  unsigned ft_hour:5; /* 0-23 */
  unsigned ft_day:5; /* 1-31 */
  unsigned ft_month:4; /* 1-12 */
  unsigned ft_year:7; /* since 1980 */
};

struct date {
  short da_year;
  char da_day;
  char da_mon;
};

struct time {
  unsigned char ti_min;
  unsigned char ti_hour;
  unsigned char ti_hund;
  unsigned char ti_sec;
};

struct dfree {
  unsigned df_avail;
  unsigned df_total;
  unsigned df_bsec;
  unsigned df_sclus;
};





extern unsigned short _osmajor, _osminor;
extern unsigned short _os_trueversion;
extern const char * _os_flavor;
extern int _doserrno;

unsigned short _get_dos_version(int);

int _get_fat_size(const int _drive);
int _get_fs_type(const int _drive, char *const _result_str);
int _is_cdrom_drive(const int _drive);
int _is_fat32(const int _drive);
int _is_ram_drive(const int _drive);
int _media_type(const int _drive);

int int86(int _ivec, union REGS *in, union REGS *out);
int int86x(int _ivec, union REGS *_in, union REGS *_out, struct SREGS *_seg);
int intdos(union REGS *_in, union REGS *_out);
int intdosx(union REGS *_in, union REGS *_out, struct SREGS *_seg);
int bdos(int _func, unsigned _dx, unsigned _al);
int bdosptr(int _func, void *_dx, unsigned _al);

int _int86(int ivec, union REGS *in, union REGS *out);





int enable(void);
int disable(void);

int getftime(int _handle, struct ftime *_ftimep);
int setftime(int _handle, struct ftime *_ftimep);

int getcbrk(void);
int setcbrk(int _new_value);

void getdate(struct date *);
void gettime(struct time *);
void setdate(struct date *);
void settime(struct time *);

void getdfree(unsigned char _drive, struct dfree *_ptr);

void delay(unsigned _msec);


/*
 *  For compatibility with other DOS C compilers.
 */
# 204 "../../../../../include/dos.h"
struct _dosdate_t {
  unsigned char day; /* 1-31 */
  unsigned char month; /* 1-12 */
  unsigned short year; /* 1980-2099 */
  unsigned char dayofweek; /* 0-6, 0=Sunday */
};


struct _dostime_t {
  unsigned char hour; /* 0-23 */
  unsigned char minute; /* 0-59 */
  unsigned char second; /* 0-59 */
  unsigned char hsecond; /* 0-99 */
};


struct _find_t {
  char reserved[21];
  unsigned char attrib;
  unsigned short wr_time;
  unsigned short wr_date;
  unsigned int size;
  char name[256];
} __attribute__((packed));

/* make sure that structure packing is correct */
typedef int _DJCHK_FIND_T[(sizeof(struct _find_t)==286)*3 - 1];

struct _diskfree_t {
  unsigned short total_clusters;
  unsigned short avail_clusters;
  unsigned short sectors_per_cluster;
  unsigned short bytes_per_sector;
};


struct _DOSERROR {
  int exterror;
  char _class;
  char action;
  char locus;
};


struct _DOSERROR_STR {
  char *exterror_str;
  char *class_str;
  char *action_str;
  char *locus_str;
};


unsigned int _dos_creat(const char *_filename, unsigned int _attr, int *_handle);
unsigned int _dos_creatnew(const char *_filename, unsigned int _attr, int *_handle);
unsigned int _dos_open(const char *_filename, unsigned int _mode, int *_handle);
unsigned int _dos_write(int _handle, const void *_buffer, unsigned int _count, unsigned int *_result);
unsigned int _dos_read(int _handle, void *_buffer, unsigned int _count, unsigned int *_result);
unsigned int _dos_close(int _handle);
unsigned int _dos_commit(int _handle);

unsigned int _dos_findfirst(const char *_name, unsigned int _attr, struct _find_t *_result);
unsigned int _dos_findnext(struct _find_t *_result);

void _dos_getdate(struct _dosdate_t *_date);
unsigned int _dos_setdate(struct _dosdate_t *_date);
void _dos_gettime(struct _dostime_t *_time);
unsigned int _dos_settime(struct _dostime_t *_time);

unsigned int _dos_getftime(int _handle, unsigned int *_p_date, unsigned int *_p_time);
unsigned int _dos_setftime(int _handle, unsigned int _date, unsigned int _time);
unsigned int _dos_getfileattr(const char *_filename, unsigned int *_p_attr);
unsigned int _dos_setfileattr(const char *_filename, unsigned int _attr);

void _dos_getdrive(unsigned int *_p_drive);
void _dos_setdrive(unsigned int _drive, unsigned int *_p_drives);
unsigned int _dos_getdiskfree(unsigned int _drive, struct _diskfree_t *_diskspace);

int _dosexterr(struct _DOSERROR *_p_error);

int _dostrerr(struct _DOSERROR *_p_error, struct _DOSERROR_STR *_p_str);
# 113 "fstat.c" 2
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 114 "fstat.c" 2
# 1 "../../../../../include/sys/stat.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 44 "../../../../../include/sys/stat.h"
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 45 "../../../../../include/sys/stat.h" 2
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 46 "../../../../../include/sys/stat.h" 2





struct stat {
  time_t st_atime;
  time_t st_ctime;
  dev_t st_dev;
  gid_t st_gid;
  ino_t st_ino;
  mode_t st_mode;
  time_t st_mtime;
  nlink_t st_nlink;
  off_t st_size;
  blksize_t st_blksize;
  uid_t st_uid;
  dev_t st_rdev;
};

int chmod(const char *_path, mode_t _mode);
int fchmod(int _fildes, mode_t _mode);
int fstat(int _fildes, struct stat *_buf);
int mkdir(const char *_path, mode_t _mode);
int mkfifo(const char *_path, mode_t _mode);
int stat(const char *_path, struct stat *_buf);
mode_t umask(mode_t _cmask);



/* POSIX.1 doesn't mention these at all */
# 93 "../../../../../include/sys/stat.h"
void _fixpath(const char *, char *);
char * __canonicalize_path(const char *, char *, size_t);
unsigned short _get_magic(const char *, int);
int _is_executable(const char *, int, const char *);
int lstat(const char * _path, struct stat * _buf);
int mknod(const char *_path, mode_t _mode, dev_t _dev);
char * _truename(const char *, char *);
char * _truename_sfn(const char *, char *);

/* Bit-mapped variable _djstat_flags describes what expensive
   f?stat() features our application needs.  If you don't need a
   feature, set its bit in the variable.  By default, all the
   bits are cleared (i.e., you get the most expensive code).  */







extern unsigned short _djstat_flags;

/* Bit-mapped variable _djstat_fail_bits describes which individual
   undocumented features f?stat() failed to use.  To get a human-
   readable description of the bits, call _djstat_describe_lossage(). */
# 131 "../../../../../include/sys/stat.h"
extern unsigned short _djstat_fail_bits;
# 115 "fstat.c" 2
# 1 "../../../../../include/limits.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 30 "../../../../../include/limits.h"
  /* These are also defined by stdint.h.  */
# 74 "../../../../../include/limits.h"
/* #define ARG_MAX			4096 -- depends on tb size; use sysconf */

/* #define OPEN_MAX		20 - DOS can change this */
/* #define STREAM_MAX		20 - DOS can change this */
# 91 "../../../../../include/limits.h"
/* constants used in Solaris */



/* gnuc ones */
# 116 "fstat.c" 2
# 1 "../../../../../include/libc/fd_props.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 13 "../../../../../include/libc/fd_props.h"
# 1 "../../../../../include/string.h" 1
/* Copyright (C) 2017 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 14 "../../../../../include/libc/fd_props.h" 2
# 24 "../../../../../include/libc/fd_props.h"
/* Delete file when the last descriptor referencing it is closed.  */


/* Tell write and _write to test for file offset greater than EOF.  If so,
   they will fill the gap with zeroes.  */


/* Set when there can't be an EOF gap or it should be left alone.  */


/* Set when the descriptor is used for pipe emulation.  */


/* Set when the descriptor is opened for append only. */


/* Set when the descriptor is used for directory emulation. */


typedef struct fd_properties fd_properties;

struct fd_properties
{
  unsigned char ref_count;
  char *filename;
  unsigned long flags;
  fd_properties *prev;
  fd_properties *next;
};

extern fd_properties ** __fd_properties;

int __set_fd_properties(int _fd, const char * _file, int _oflags);
void __dup_fd_properties(int _from, int _to);
int __clear_fd_properties(int _fd);

static __inline__ int __has_fd_properties(int _fd)
{
  return _fd >= 0 && __fd_properties && __fd_properties[_fd];
}

static __inline__ void __set_fd_flags(int _fd, unsigned long _flags)
{
  __fd_properties[_fd]->flags |= _flags;
}

static __inline__ void __clear_fd_flags(int _fd, unsigned long _flags)
{
  __fd_properties[_fd]->flags &= ~_flags;
}

static __inline__ unsigned long __get_fd_flags(int _fd)
{
  return __has_fd_properties(_fd) ? __fd_properties[_fd]->flags : 0;
}

static __inline__ const char * __get_fd_name(int _fd)
{
  return __has_fd_properties(_fd) ? __fd_properties[_fd]->filename : ((void*)0);
}
# 117 "fstat.c" 2
# 1 "../../../../../include/libc/getdinfo.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
# 20 "../../../../../include/libc/getdinfo.h"
/* For parsing the bits returned by _get_dev_info.  */
# 118 "fstat.c" 2

# 1 "../../../../../include/dpmi.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 120 "fstat.c" 2
# 1 "../../../../../include/go32.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../../include/go32.h"
# 1 "../../../../../include/sys/version.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../../include/go32.h" 2
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../../include/go32.h" 2
# 1 "../../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../../include/libc/asmobj.h" 2
# 27 "../../../../../include/go32.h" 2
# 1 "../../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 28 "../../../../../include/go32.h" 2






typedef struct {
  unsigned int size_of_this_structure_in_bytes;
  unsigned int linear_address_of_primary_screen;
  unsigned int linear_address_of_secondary_screen;
  unsigned int linear_address_of_transfer_buffer;
  unsigned int size_of_transfer_buffer; /* >= 4k */
  unsigned int pid;
  unsigned char master_interrupt_controller_base;
  unsigned char slave_interrupt_controller_base;
  unsigned short selector_for_linear_memory;
  unsigned int linear_address_of_stub_info_structure;
  unsigned int linear_address_of_original_psp;
  unsigned short run_mode;
  unsigned short run_mode_info;
} __Go32_Info_Block;

extern __Go32_Info_Block *___go32_info_block(void);/*
*/
# 50 "../../../../../include/go32.h"
       #define _go32_info_block (*___go32_info_block())/*
*/
# 50 "../../../../../include/go32.h"
                                                 ;







# 1 "../../../../../include/sys/movedata.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 59 "../../../../../include/go32.h" 2
# 1 "../../../../../include/sys/segments.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../../include/sys/segments.h"
int _my_cs(void);
int _my_ds(void);
int _my_ss(void);
# 60 "../../../../../include/go32.h" 2
# 70 "../../../../../include/go32.h"
/* returns number of times hit since last call. (zero first time) */
unsigned _go32_was_ctrl_break_hit(void);
void _go32_want_ctrl_break(int yes); /* auto-yes if call above function */
# 121 "fstat.c" 2
# 1 "../../../../../include/libc/farptrgs.h" 1
/* special version for libc - uses %gs instead of %fs.  Ignore comments */

/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
/* Copyright (c) 1995 DJ Delorie.  Permission granted to use for any
   purpose, provided this copyright remains attached and unmodified.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

ͻ
		Far Pointer Simulation Functions			
ͼ

This file attempts to make up for the lack of a "far" keyword in GCC.
Although it doesn't provide access to far call APIs (like Windows), it
does allow you to do far pointer data access without the overhead of
movedata() or dosmemget/dosmemput().

You should *always* include this file when using these functions and
compile with optimization enabled.  They don't exist as normal functions
in any library, and they compile down to only a few opcodes when used
this way.  They are almost as fast as native pointer operations, and
about as fast as far pointers can get.

If you don't use optimization, this file becomes prototypes for
farptr.c, which generates real functions for these when not optimizing.
When optimizing, farptr.c compiles to nothing.

There are two types of functions here - standalone and invariant.  The
standalone functions take a selector and offset.  These are used when
you need only a few accesses, time isn't critical, or you don't know
what's in the %gs register.  The invariant ones don't take a selector,
they only take an offset.  These are used inside loops and in
time-critical accesses where the selector doesn't change.  To specify
the selector, use the farsetsel() function.  That selector is used for
all farns*() functions until changed.  You can use _fargetsel() if you
want to temporary change the selector with _farsetsel() and restore
it afterwards.

The farpoke* and farpeek* take selectors.

The farnspoke* and farnspeek* don't (note the `ns' for `no selector').

Warning: These routines all use the %gs register for their accesses.
GCC normally uses only %ds and %es, and libc functions (movedata,
dosmemget, dosmemput) use %gs.  Still, you should be careful about
assumptions concerning whether or not the value you put in %gs will be
preserved across calls to other functions.  If you guess wrong, your
program will crash.  Better safe than sorry.

*/
# 74 "../../../../../include/libc/farptrgs.h"
# 1 "../../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 75 "../../../../../include/libc/farptrgs.h" 2

void _farpokeb_gs(unsigned short, unsigned int, unsigned char);
void _farpokew_gs(unsigned short, unsigned int, unsigned short);
void _farpokel_gs(unsigned short, unsigned int, unsigned int);
unsigned char _farpeekb_gs(unsigned short, unsigned int);
unsigned short _farpeekw_gs(unsigned short, unsigned int);
unsigned int _farpeekl_gs(unsigned short, unsigned int);
void _farsetsel_gs(unsigned short);
unsigned short _fargetsel_gs(void);
void _farnspokeb_gs(unsigned int, unsigned char);
void _farnspokew_gs(unsigned int, unsigned short);
void _farnspokel_gs(unsigned int, unsigned int);
unsigned char _farnspeekb_gs(unsigned int);
unsigned short _farnspeekw_gs(unsigned int);
unsigned int _farnspeekl_gs(unsigned int);
# 122 "fstat.c" 2
# 1 "../../../../../include/libc/bss.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 23 "../../../../../include/libc/bss.h"
/* this gets *incremented* at startup, so that if a program dumps
   itself (emacs) you still have a way of telling.  Compare this value
   with a static, and if it's different, you've been restarted.
   First time through on a non-unexec'd program, it's greater than 0 */

extern int __bss_count;
# 123 "fstat.c" 2
# 1 "../../../../../include/libc/symlink.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */

/* Written by Laurynas Biveinis                                */
/* This file contains some internal info related to symlinks   */
/* Note: symlink file format is still in internal include file */
/* because I don't think it's required for user apps           */
# 28 "../../../../../include/libc/symlink.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 29 "../../../../../include/libc/symlink.h" 2






/* Symlink files have fixed length - 510 bytes. Why this value? Why not? */
/* It is big enough to hold longest possible path                        */


/* Semi-internal library function which reads symlink contents given  */
/* either a file name or its handle.  Used by readlink(), fstat() and */
/* user supplied (if any) file fstat handler.                         */

int __internal_readlink(const char * __path, int __fhandle, char * __buf,
                        size_t __max);

/* A prototype for internal library function for fully resolving symlink   */
/* chain. Standard function readlink() solves only one symlink level.      */
/* If path name passed appears to be not a symlink, it is copied to result */
/* string. Return code 1 means success, 0 - failure (to many links - errno */
/* is set too).                                                            */

int __solve_symlinks(const char * __symlink_path, char * __real_path);

/* Internal library function for fully resolving symlinks in `__symlink_path' */
/* except the last its component. Otherwise it is similar to __solve_symlinks */
int __solve_dir_symlinks(const char * __symlink_path, char * __real_path);
# 124 "fstat.c" 2
# 1 "../../../../../include/sys/fsext.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../../include/sys/fsext.h"
# 1 "../../../../../include/stdarg.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../../include/sys/fsext.h" 2

typedef enum {
  __FSEXT_nop,
  __FSEXT_open,
  __FSEXT_creat,
  __FSEXT_read,
  __FSEXT_write,
  __FSEXT_ready,
  __FSEXT_close,
  __FSEXT_fcntl,
  __FSEXT_ioctl,
  __FSEXT_lseek,
  __FSEXT_link,
  __FSEXT_unlink,
  __FSEXT_dup,
  __FSEXT_dup2,
  __FSEXT_fstat,
  __FSEXT_stat,
  __FSEXT_llseek,
  __FSEXT_readlink,
  __FSEXT_symlink,
  __FSEXT_fchown,
  __FSEXT_chmod,
  __FSEXT_chown,
  __FSEXT_fchmod
} __FSEXT_Fnumber;

/* _ready gets passed a fd and should return a mask of these,
   as if we were emulating "int ready(int fd)" */




/* The return value is nonzero if the function has overridden the
   caller's functionality. */
typedef int (__FSEXT_Function)(__FSEXT_Fnumber _function_number,
          int *_rv, va_list _args);

int __FSEXT_alloc_fd(__FSEXT_Function *_function);
int __FSEXT_set_function(int _fd, __FSEXT_Function *_function);
__FSEXT_Function *__FSEXT_get_function(int _fd);
void *__FSEXT_set_data(int _fd, void *_data);
void *__FSEXT_get_data(int _fd);

int __FSEXT_add_open_handler(__FSEXT_Function *_function);
int __FSEXT_call_open_handlers(__FSEXT_Fnumber _function_number,
          int *rv, va_list _args);
# 125 "fstat.c" 2
# 1 "../../../../../include/libc/fsexthlp.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../../include/stdarg.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 8 "../../../../../include/libc/fsexthlp.h" 2
# 1 "../../../../../include/sys/fsext.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 9 "../../../../../include/libc/fsexthlp.h" 2
# 25 "../../../../../include/libc/fsexthlp.h"
/* This is a wrapper for an FSEXT function that allows it to be called
 * with a variable number of arguments. */

static inline int
__FSEXT_func_wrapper (__FSEXT_Function *func,
        __FSEXT_Fnumber fnum,
        int *rv,
        ...)
{
  va_list args;
  int ret;

  __builtin_va_start(args,rv);
  ret = func(fnum, rv, args);
  __builtin_va_end(args);

  return(ret);
}

/* This is a wrapper for __FSEXT_call_open_handlers that allows it
 * to be called with a variable number of arguments. */

static inline int
__FSEXT_call_open_handlers_wrapper (__FSEXT_Fnumber fnum, int *rv, ...)
{
  va_list args;
  int ret;

  __builtin_va_start(args,rv);
  ret = __FSEXT_call_open_handlers(fnum, rv, args);
  __builtin_va_end(args);

  return(ret);
}
# 126 "fstat.c" 2
# 1 "xstat.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
/*
 * Header for internal stat()/fstat()/lstat() assist functions.
 *
 */




# 1 "../../../../../include/stdio.h" 1
/* Copyright (C) 2018 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2017 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 16 "xstat.h" 2
# 1 "../../../../../include/time.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 17 "xstat.h" 2
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 18 "xstat.h" 2
# 1 "../../../../../include/sys/stat.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 19 "xstat.h" 2
# 1 "../../../../../include/dpmi.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 20 "xstat.h" 2

/* Some errno.h headers do not define EFAULT.  Here the value is
   taken from POSIX IEEE 1003.1.  */
# 1 "../../../../../include/errno.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "xstat.h" 2




/* Under MS-DOS, file access permissions are shared by all, except for
   Write permission. */




/* Macro to convert a segment and an offset to a "far offset" suitable
   for _farxxx() functions of DJGPP.  */






extern unsigned short _osmajor, _osminor;
extern const char * _os_flavor;

extern time_t _file_time_stamp(unsigned int);
extern ino_t _invent_inode(const char *, unsigned, unsigned long);
extern unsigned short _get_magic(const char *, int);
extern unsigned short _get_dos_version(int);
extern char * _truename(const char *, char *);
extern int _is_remote_drive(int);
extern int _is_executable(const char *, int, const char *);
extern short _get_dev_info(int);
extern long __filelength(int);
extern int _is_remote_handle(int);
extern void _djstat_describe_lossage(FILE *);
extern int _getftime(int, unsigned int *);
extern blksize_t _get_cached_blksize (const char *path);
# 127 "fstat.c" 2

/* Should we bother about executables at all? */


/* Should we bother about any access bits?  */


/* Do we need SFT info at all? */


/*
 * Lower-level assist functions to get a starting cluster of a file,
 * which will serve as an inode number.  The starting cluster is
 * found in the System File Table entry (an internal structure of
 * DOS) which belongs to our file.
 *
 * Much of the following code is derived from file H2NAME.C, which
 * came with ``Undocumented DOS'', 1st edition.
 */

/* Array of SFT entry sizes as function of DOS version. */
static size_t sft_size_list[] = {0, 0x28, 0x35, 0x3b};

/* Actual size of SFT entry for the version of DOS under
   which we are running.  */
static size_t sft_size;

/* Static array to hold a copy of SFT entry.  Should be at
   least as long as the largest number in sft_size_list[].  */
static unsigned char sft_buf[0x40];

/* Linear address of pointer to Job File Table (JFT) which
   is an array of indices into the SFT which correspond to
   open file handles.  */
static unsigned long htbl_ptr_addr;

/* True version of DOS (not the one simulated by SETVER).  */
static unsigned short dos_major, dos_minor;

/* Segment and offset of first SFT sub-table.  All searches
   start from this address.  */
static unsigned short sft_start_seg, sft_start_off;

/* This holds the failure bits from last call to fstat_init(),
   so we can return them each time fstat_assist() is called.  */
static unsigned short fstat_init_bits;

/* Holds the last seen value of __bss_count, to be safe for
   restarted programs (emacs).  */
static int fstat_count = -1;

/* The address of the PSP of the caller.  */
static unsigned long psp_addr;

/* Initialization routine, called once per program run.
 * Finds DOS version, SFT entry size and addresses of
 * program handle table and first SFT sub-table.
 */
static int
fstat_init(void)
{
  __dpmi_regs regs = {0};
  int sft_ptr_addr;
  unsigned short true_dos_version;

  /* Each DOS program has a table of file handles which are used by
   * DOS open() call.  This table holds, for each handle which is in
   * use, the index into the System File Tables' list which contains
   * data about this handle.  The pointer to that handle table is found
   * at offset 34h in the program's PSP.
   */

  /* Linear address of pointer to the handle table.  We postpone
   * dereferencing this pointer to obtain the address of the handle
   * table until we're actually called for a specific handle, because
   * somebody could in the meanwhile change that address, e.g. by
   * calling INT 21h/AX=67h to enlarge the maximum number of file
   * handles.
   */
  htbl_ptr_addr = psp_addr + 0x34;

  /*
   * Find the pointer to the first subtable in the list of SFT's.
   * It is stored at offset 4 in the DOS List-of-Lists area, a ptr
   * to which is returned by the undocumented DOS function 52h.
   * We don't check FLAGS after 52h returns because Ralph Brown's
   * Interrupt List doesn't say FLAGS are set to indicate a
   * failure.
   */
  regs.h.ah = 0x52;
  __dpmi_int(0x21, &regs);

  /* Linear addres of pointer to SFT list.  */
  sft_ptr_addr = ((int)((((unsigned long)(regs.x.es)) << 4) + (unsigned short)(regs.x.bx + 4)));

  /* SFT entry size depends on DOS version.
     We need exact knowledge about DOS internals, so we need the
     TRUE DOS version (not the simulated one by SETVER), if that's
     available.  */
  true_dos_version = _os_trueversion;
  dos_major = true_dos_version >> 8;
  dos_minor = true_dos_version & 0xff;
  sft_size = sft_size_list[dos_major > 4 ? 3 : dos_major - 1];
  if (!sft_size) /* unsupported DOS version */
    {
      _djstat_fail_bits |= 2 /* Unsupported DOS version */;
      return 0;
    }

  /* Segment and offset of start of SFT list.  */
  sft_start_off = _farpeekw_gs(_go32_info_block.selector_for_linear_memory, sft_ptr_addr);
  sft_start_seg = _farpeekw_gs(_go32_info_block.selector_for_linear_memory, sft_ptr_addr + 2);

  return 1;
}

/* Given a handle, copy contents of System File Table entry which
 * belongs to that handle into a local buffer, and return the index
 * into the SFT array where our entry was found.  In case of failure
 * to use SFT, return -2.  If FHANDLE is illegal, return -1.
 */
static short
get_sft_entry(int fhandle)
{
  unsigned long sft_seg;
  unsigned short sft_off;
  unsigned long htbl_addr;
  short sft_idx, retval;

  __dpmi_regs regs = {0};
  _djstat_fail_bits = fstat_init_bits;

  /* Force initialization if we were restarted (emacs).  */
  if (fstat_count != __bss_count)
    {
      fstat_count = __bss_count;
      dos_major = 0;
    }

  /* Find the PSP address of the current process.  */
  regs.h.ah = 0x62; /* Get PSP address */
  __dpmi_int(0x21, &regs);
  psp_addr = (unsigned long)regs.x.bx << 4;

  /* If first time called, initialize.  */
  if (!dos_major && !fstat_init())
    {
      fstat_init_bits = _djstat_fail_bits;
      return -2;
    }

  /* Test file handle for validity.
   * For DOS 3.x and later, the number of possible file handles
   * is at offset 32h in the PSP; for prior versions, it is 20.
   */
  if (fhandle < 0
      || fhandle >= (_osmajor < 3 ?
       20
       : _farpeekw_gs(_go32_info_block.selector_for_linear_memory, psp_addr + 0x32)))
    return -1;

  /* Linear address of the handle table. */
  htbl_addr = ((int)((((unsigned long)(_farpeekw_gs(_go32_info_block.selector_for_linear_memory, htbl_ptr_addr + 2))) << 4) + (unsigned short)(_farpeekw_gs(_go32_info_block.selector_for_linear_memory, htbl_ptr_addr))))
                                                        ;

  /* Index of the entry for our file handle in the SFT array.  */
  retval = sft_idx = _farpeekb_gs(_go32_info_block.selector_for_linear_memory, htbl_addr + fhandle);

  if (sft_idx < 0) /* invalid file handle or bad handle table */
    {
      _djstat_fail_bits |= 0x400 /* bad SFT index in JFT */;
      return -1;
    }

  /* Given the index into the SFT list, find our SFT entry.
   * The list consists of arrays (sub-tables) of entries, each sub-
   * table preceeded by a header.  The header holds a pointer to the
   * next sub-table in the list and number of entries in this sub-table.
   * The list is searched until the sub-table which contains our
   * target is found, then the sub-table entries are skipped until
   * we arrive at our target.
   */

  /* Segment (shifted 4 bits left) and offset of start of SFT list.  */
  sft_off = sft_start_off;
  sft_seg = ((int)((((unsigned long)(sft_start_seg)) << 4) + (unsigned short)(0)));

  while (sft_off != 0xFFFF)
    {
      unsigned long entry_addr = sft_seg + sft_off;
      short subtable_len = _farpeekw_gs(_go32_info_block.selector_for_linear_memory, entry_addr + 4);

      if (sft_idx < subtable_len)
        { /* Our target is in this sub-table.  Pull in the entire
           * SFT entry for use by fstat_assist().
           */
   fmemcpy2(sft_buf, (__dpmi_paddr){ .selector = _go32_info_block.selector_for_linear_memory, .offset32 = entry_addr + 6 + sft_idx * sft_size, },
     sft_size);
          return retval;
        }
      /* Our target not in this subtable.
       * Subtract the number of entries in this sub-table from the
       * index of our entry, and proceed to next sub-table.
       */
      sft_idx -= subtable_len;
      sft_off = _farpeekw_gs(_go32_info_block.selector_for_linear_memory, entry_addr);
      sft_seg = ((int)((((unsigned long)(_farpeekw_gs(_go32_info_block.selector_for_linear_memory, entry_addr + 2))) << 4) + (unsigned short)(0)));
    }

  /* Get here only by error, which probably means unsupported DOS version. */
  _djstat_fail_bits |= 0x800 /* file entry not found in SFT array */;
  return -2;
}

/* On LFN platforms, we can get all the 3 time-related fields.  */

static void
set_fstat_times (int fhandle, struct stat *stat_buf)
{
  if (_use_lfn(0) /* assume it's the same on ALL drives */)
    {
      time_t access_time;
      unsigned int create_time;

      /* Access time is currently date only (time is zeroed).  */
      access_time = _file_time_stamp (_lfn_get_ftime (fhandle, 2));
      if (access_time > stat_buf->st_atime)
 stat_buf->st_atime = access_time;

      /* Creation time might be zero if the file was created
	 by a DOS program which doesn't support LFN API.  */
      create_time = _lfn_get_ftime (fhandle, 1);
      if (create_time)
 stat_buf->st_ctime = _file_time_stamp (create_time);
    }
}

/* fstat_assist() is where all the actual work is done.
 * It uses SFT entry, if available and its contents are verified.
 * Otherwise, it finds all the available info by conventional
 * DOS calls.
 */

static int
fstat_assist(int fhandle, struct stat *stat_buf)
{
  short have_trusted_values = 1;
  unsigned int dos_ftime;
  char drv_no;
  short dev_info;
  unsigned char is_dev;
  unsigned char is_remote;
  short sft_idx = -1;
  unsigned short sft_fdate, sft_ftime;
  long sft_fsize;
  unsigned short trusted_ftime = 0, trusted_fdate = 0;
  long trusted_fsize = 0;
  int is_link = 0;
  const char *fd_name = ((void*)0);
  const char *filename = "";

  if ((dev_info = _get_dev_info(fhandle)) == -1)
    return -1; /* errno set by _get_dev_info() */

  _djstat_fail_bits = 0;

  /* Get pointer to an SFT entry which holds data for our handle. */
  if ( (_djstat_flags & (0x20 /* fstat() needs write bit? */ | 2 /* get execute bits from file extension? */ | 1 /* should we bother getting inode numbers? */)) == 0 &&
       (sft_idx = get_sft_entry(fhandle)) == -1)
    {
      __dj_errno = 6;
      return -1;
    }

  /* Initialize buffers. */
  memset(stat_buf, 0, sizeof(struct stat));
  dos_ftime = 0;

  /* Get some info about this handle by conventional DOS calls.  These
   * will serve as verification of SFT entry contents and also as
   * fall-back in case SFT method fails.
   */
  if (_getftime(fhandle, &dos_ftime) == 0 &&
      (trusted_fsize = __filelength(fhandle)) != -1L)
    {
      trusted_ftime = dos_ftime & 0xffff;
      trusted_fdate = dos_ftime >> 16;
    }
  else
    have_trusted_values = 0;

  if (dev_info & 0x0080)
    {
      is_dev = 1;
      is_remote = 0; /* device can't be remote */
    }
  else
    {
      is_dev = 0;
      if (dev_info & 0x8000)
 is_remote = 1;
      else
 is_remote = 0;

      if(!have_trusted_values && dev_info == 0 && _os_trueversion == 0x532)
        is_dev = 1; /* Device under NT or Win2K with pre-open/lfn handle. */
    }

  /* First, fill the fields which are constant under DOS. */
  stat_buf->st_uid = getuid();
  stat_buf->st_gid = getgid();
  stat_buf->st_nlink = 1;

  /* Get the file name from the file descriptor properties (fd_props),
   * if possible, and fix it up. */
  fd_name = __get_fd_name(fhandle);
  if (fd_name != ((void*)0))
    filename = fd_name;

  /* Get the block size for the device associated with `fhandle'. */

  if (*filename)
    {
      stat_buf->st_blksize = _get_cached_blksize(filename);
      if (stat_buf->st_blksize == -1)
 return -1; /* errno set by _get_cached_blksize() */
    }
  else
    {
      /* Fall back on transfer buffer size, if we can't determine file name
       * (which gives the drive letter and then the drive's cluster size). */
      stat_buf->st_blksize = _go32_info_block.size_of_transfer_buffer;
    }


  /* If SFT entry for our handle is required and available, we will use it.  */
  if ( (_djstat_flags & (0x20 /* fstat() needs write bit? */ | 2 /* get execute bits from file extension? */ | 1 /* should we bother getting inode numbers? */)) == 0 && sft_idx >= 0)
    {
      /* Determine positions of data items in the SFT. */
      size_t fattr_ofs, name_ofs, ext_ofs, fsize_ofs, fdate_ofs,
             ftime_ofs, clust_ofs;

      switch (dos_major)
        {
          case 2:
              fattr_ofs = 2;
              drv_no = sft_buf[3] - 1; /* 1 = 'A' etc. */
              name_ofs = 4;
              ext_ofs = 0x0b;
              fsize_ofs = 0x13;
              fdate_ofs = 0x17;
              ftime_ofs = 0x19;
              clust_ofs = 0x1c;
              is_remote = 0; /* DOS 2.x didn't have remote files */
              break;

          case 3:
              fattr_ofs = 4;
              drv_no = sft_buf[5] & 0x3f;
              if (dos_minor == 0)
                {
                  name_ofs = 0x21;
                  ext_ofs = 0x29;
                }
              else /* DOS 3.1 - 3.3x */
                {
                  name_ofs = 0x20;
                  ext_ofs = 0x28;
                }
              clust_ofs = 0x0b;
              ftime_ofs = 0x0d;
              fdate_ofs = 0x0f;
              fsize_ofs = 0x11;
              break;

          default: /* DOS 4 and up */
              fattr_ofs = 4;
              drv_no = sft_buf[5] & 0x3f;
              clust_ofs = 0x0b;
              ftime_ofs = 0x0d;
              fdate_ofs = 0x0f;
              fsize_ofs = 0x11;
              name_ofs = 0x20;
              ext_ofs = 0x28;

        }

      if (is_dev)
        {
          /* We have a character device.
           * We will pretend as if they all reside on a special
           * drive `@:', which is illegal in DOS, and just happens
           * to give a neat st_dev (= '@' - 'A') = -1.
           */

          stat_buf->st_dev = -1;




          if ( (_djstat_flags & 1 /* should we bother getting inode numbers? */) == 0 )
            {
              /* Character device names are all at most 8-character long. */
              short i = 8;
              unsigned char *src = sft_buf + name_ofs;
              char dev_name[16], *dst = dev_name + 7;

              strcpy(dev_name, "@:\\dev\\        "); /* pad with 8 blanks */
              while (i-- && *src != ' ') /* copy non-blank chars */
                *dst++ = *src++;

              stat_buf->st_ino = _invent_inode(dev_name, 0, 0);
            }

          /* Should we make printer devices write-only here? */
          stat_buf->st_mode |= (0x2000 | (00400 | 00040 | 00004) | 00200);

          /* We will arrange things so that devices have current time in
           * the access-time and modified-time fields of struct stat.
           */
          stat_buf->st_atime = stat_buf->st_mtime = time(0);

          /* MS-DOS returns the time of boot when _getftime() is called
           * for character devices, but this is undocumented and
           * unsupported by DOS clones (e.g. DR-DOS).  It is also
           * inconsistent with our stat().  Therefore, we will arrange
           * things so that devices have zero (the beginning of times)
           * in creation-time field.
           */
          dos_ftime = 0;
          stat_buf->st_ctime = _file_time_stamp(dos_ftime);

          return 0;
        }

      /* Files are not allowed to fail DOS calls for their time
       * stamps and size.
       */
      else if (have_trusted_values)
        {
   unsigned char *pname = sft_buf + name_ofs;

          /* This is a regular, existing file.  It cannot be a
           * directory, because DOS won't let us open() a directory.
           * Each file under MS-DOS is always readable by everyone.
           */
          stat_buf->st_mode |= (0x0000 | (00400 | 00040 | 00004));

          /* We will be extra careful in trusting SFT data: it must be
           * consistent with date, time and size of the file as known
           * from conventional DOS calls.
           */
          sft_fdate = *((unsigned short *)(sft_buf + fdate_ofs));
          sft_ftime = *((unsigned short *)(sft_buf + ftime_ofs));
          sft_fsize = *((long *)(sft_buf + fsize_ofs));

   /* In addition, it seems that 32-bit File Access, at least
	   * in Windows 95, creates fake SFT entries for some files,
	   * which have bogus cluster numbers and all-blank file name
	   * and extension.  (It is unclear to me when exactly are
	   * these fake SFT entries created.)
	   * So, in addition, we check the file's name in the SFT to
	   * be non-blank, since file names in the FCB format cannot
	   * be all-blank, even on Windows 95.  */
   while (pname < sft_buf + name_ofs + 8 + 3 && *pname == ' ')
     pname++;

          if (pname < sft_buf + name_ofs + 8 + 3 &&
       sft_ftime == trusted_ftime &&
              sft_fdate == trusted_fdate &&
              sft_fsize == trusted_fsize)
            { /* Now we are ready to get the SFT info. */
              char sft_extension[4], *dst = sft_extension + 2;
              unsigned char *src = sft_buf + ext_ofs + 2;
              int i = 3;

              /* Get the file's extension.  It is held in the SFT entry
               * as a blank-padded 3-character string without terminating
               * zero.  Some crazy files have embedded blanks in their
               * extensions, so only TRAILING blanks are insignificant.
               */
              memset(sft_extension, 0, sizeof(sft_extension));;
              while (*src == ' ' && i--) /* skip traling blanks */
                {
                  dst--; src--;
                }

              if (i >= 0)
                while (i--) /* move whatever left */
                  *dst-- = *src--;

              /* Build Unix-style file permission bits. */
              /* Check for symlink at first.  This test will fail if we
               * don't have read access to file.
               */
              if (sft_fsize == 510)
              {
                 int old_errno = __dj_errno;
   char buf[2];
   int bytes_read = __internal_readlink(((void*)0), fhandle, buf, 1);
   if (bytes_read != -1)
   {
      stat_buf->st_mode = 0x8000;
      is_link = 1;
   }
   else
       __dj_errno = old_errno;
              }
              if (!is_link)
              {
                 if ( !(sft_buf[fattr_ofs] & 0x07) ) /* no R, S or H bits set */
                   stat_buf->st_mode |= 00200;

                 /* Execute permission bits.  fstat() cannot be called on
                  * directories under DOS, so only executable programs/batch
                  * files should be considered.
                  */
                 if (_is_executable((const char *)0, fhandle, sft_extension))
                   stat_buf->st_mode |= (00100 | 00010 | 00001);

                 /* DOS 4.x and above seems to know about named pipes. */
                 if (dos_major > 3 && (sft_buf[6] & 0x20))
                   stat_buf->st_mode |= 0x4000;
              }
              /* Device code. */
              stat_buf->st_dev = drv_no;




              /* The file's starting cluster number will serve as its
               * inode number.
               */
              if ( (_djstat_flags & 1 /* should we bother getting inode numbers? */) == 0 && !is_remote)
                stat_buf->st_ino = *((unsigned short *)(sft_buf + clust_ofs));

              /* If the cluster number returns zero (e.g., for empty files,
               * because DOS didn't allocate it a cluster yet) we have to
               * invent the inode using the file's name.  We will use the
               * index into the SFT as part of unique identifier for our
               * file, so a possibility of two files with the same name
               * but different paths getting the same inode number is
               * minimized.
               * If we have a remote file, we invent inode even if there
               * is a non-zero number in the SFT, because it usually is
               * bogus (a left-over from last local file handle which used
               * the same SFT entry).
               * Note that we invent the inode even if is_remote is -1
               * (i.e., IOCTL Func 0Ah failed), because that should mean
               * some network redirector grabs IOCTL functions in an
               * incompatible way.
               */
              if ( (_djstat_flags & 1 /* should we bother getting inode numbers? */) == 0 &&
                   (stat_buf->st_ino == 0 || is_remote))
                {
                  static char name_pat[] = " :sft-   \\            ";
                  char name_buf[sizeof(name_pat)];
                  unsigned char *src_p = sft_buf + name_ofs + 7;
                  char *dst_p = name_buf + 17;
                  int j = 8;
                  char *name_end;
                  int first_digit = sft_idx / 100;
                  int second_digit = (sft_idx - first_digit * 100) / 10;
                  int third_digit = sft_idx - first_digit * 100
                                                        - second_digit * 10;

                  /* Initialize the name buffer with zeroes, then
                   * put in the drive letter and ``sft-XXX'', where
                   * XXX is the index of our file entry in the SFT.
                   */
                  strcpy(name_buf, name_pat);
                  memset(name_buf + 10, 0, sizeof(name_buf) - 10);
                  name_buf[0] = drv_no + 'A';
                  name_buf[6] = first_digit + '0';
                  name_buf[7] = second_digit + '0';
                  name_buf[8] = third_digit + '0';

                  /* Copy filename from SFT entry to local storage.
                   * It is stored there in the infamous DOS format:
                   * both name and extension are blank-padded, and no dot.
                   * We cannot use strcpy, because the name might
                   * include embedded blanks.  Therefore we move the
                   * characters from the end towards the beginning.
                   */
                  while (*src_p == ' ' && j--) /* skip traling blanks */
                    {
                      dst_p--;
                      src_p--;
                    }
                  name_end = dst_p + 1;

                  if (j >= 0) /* move whatever left */
                    while (j--)
                      *dst_p-- = *src_p--;

                  /* We've already got the extension.  If it is non-empty,
                   * insert a dot and copy the extension itself.
                   */
                  if (sft_extension[0])
                    {
                      *name_end++ = '.';
                      strcpy(name_end, sft_extension);
                    }
                  stat_buf->st_ino =
                    _invent_inode(name_buf, dos_ftime, sft_fsize);
                  _djstat_fail_bits |= 0x10 /* inode defaults to hashing */;
                }

              /* Size, date and time. */
              stat_buf->st_size = sft_fsize;
              stat_buf->st_atime = stat_buf->st_ctime = stat_buf->st_mtime =
                _file_time_stamp(dos_ftime);

       /* Additional time info for LFN platforms.  */
       set_fstat_times (fhandle, stat_buf);
              return 0;
            }

          _djstat_fail_bits |= 0x200 /* SFT entry found, but can't be trusted */;

        }

      /* Regular file, but DOS calls to find its length and time stamp
       * failed.  This must be an illegal file handle, or something
       * else very, very funny...
       */
      else
        return -1; /* errno set by filelength() or getftime() */

    }

  /* Can't get SFT itself or can't find SFT entry belonging to our file.
   * This is probably unsupported variety of DOS, or other (not-so-
   * compatible) OS.
   * For these we supply whatever info we can find by conventional calls.
   */
  if (have_trusted_values)
    {
      if (is_dev)
        {
          if (_djstat_flags & 1 /* should we bother getting inode numbers? */)
            {
              /* We need the name of the device to invent an inode for it.
               * We cannot get the REAL name, because SFT info is unavailable.
               * If IOCTL tells us this is one of the standard devices, we
               * can make an educated guess.  If not, we will invent inode
               * with no name.  This will at least ensure that no two calls
               * accidentally get the same inode number.
               * We will also pretend devices belong to a special drive
               * named `@'.
               */
              if (dev_info & (0x0001|0x0002|0x0004|0x0008))
                {
                  char dev_name[16];

                  strcpy(dev_name, "@:\\dev\\");
                  if (dev_info & (0x0001|0x0002))
                    strcat(dev_name, "CON     ");
                  else if (dev_info & 0x0004)
                    strcat(dev_name, "NUL     ");
                  else if (dev_info & 0x0008)
                    strcat(dev_name, "CLOCK$  ");

                  stat_buf->st_ino = _invent_inode(dev_name, 0, 0);
                }
              else
                stat_buf->st_ino = _invent_inode("", 0, 0);

              _djstat_fail_bits |= 0x10 /* inode defaults to hashing */;
            }

          stat_buf->st_dev = -1;




          stat_buf->st_mode |= (0x2000 | (00400 | 00040 | 00004) | 00200);

          stat_buf->st_atime = stat_buf->st_mtime = time(0);
          dos_ftime = 0;
          stat_buf->st_ctime = _file_time_stamp(dos_ftime);
        }
      else
        {
          /* Regular file.  We may have obtained this file's name
	   * from the file descriptor properties (fd_props).  Otherwise
	   * the inode will be arbitrary each time fstat is called.
	   * Sigh...
           */
          if ( (_djstat_flags & 1 /* should we bother getting inode numbers? */) == 0 )
            {
              _djstat_fail_bits |= 0x10 /* inode defaults to hashing */;
              stat_buf->st_ino
  = _invent_inode(filename, dos_ftime, trusted_fsize);
            }

          if (trusted_fsize == 510)
          {
            int old_errno = __dj_errno;
            char buf[2];
            int bytes_read = __internal_readlink(((void*)0), fhandle, buf, 1);
            if (bytes_read != -1)
            {
              stat_buf->st_mode = 0x8000;
              is_link = 1;
            }
            else
              __dj_errno = old_errno;
          }
          if (!is_link)
          {
            /* Return the minimum access bits every file has under DOS. */
            stat_buf->st_mode |= (0x0000 | (00400 | 00040 | 00004));
            if (_djstat_flags & (0x20 /* fstat() needs write bit? */ | (2 /* get execute bits from file extension? */ | 4 /* get execute bits from magic signature? */)))
              _djstat_fail_bits |= 0x80 /* fstat() failed to get write access bit */;

            /* If we are runing on Windows 9X, NT 4.0 with LFN or 2000 or XP
               with LFN is enabled, try harder. Note that we deliberately do
               NOT use this call when LFN is disabled, even if we are on
               Windows, because then we open the file with function 3Ch, and
               such handles aren't supported by 71A6h call we use here.  */
            if (_use_lfn(0) /* assume it's the same on ALL drives */)
            {
              __dpmi_regs r = {0};

              r.x.flags = 1; /* Always set CF before calling a 0x71NN function. */
              r.x.ax = 0x71a6; /* file info by handle */
              r.x.bx = fhandle;
              r.x.ds = _go32_info_block.linear_address_of_transfer_buffer >> 4;
              r.x.dx = 0;
              __dpmi_int(0x21, &r);
              if (!(r.x.flags & 1) && (r.x.ax != 0x7100) &&
                  !(_farpeekl_gs(_go32_info_block.selector_for_linear_memory, _go32_info_block.linear_address_of_transfer_buffer) & 0x07))
              {
                /*  Never assume that the complete LFN API is implemented,
                    so check that AX != 0x7100.  E.G.: MSDOS 6.22 and DOSLFN 0.40.  */
                stat_buf->st_mode |= 00200; /* no R, S or H bits set */
              }
            }

            /* Executables are detected if they have magic numbers.  */
            if (!(_djstat_flags & 4 /* get execute bits from magic signature? */) &&
                _is_executable((const char *)0, fhandle, (const char *)0))
              stat_buf->st_mode |= (00100 | 00010 | 00001);
          }
          /* Lower 6 bits of IOCTL return value give the device number. */
          stat_buf->st_dev = dev_info & 0x3f;




          /* Novell Netware does not return the drive number in the
           * lower 6 bits of dev_info.  But we cannot do anything with
           * that, since any value in these 6 bits could be correct...
           * In particular, 0 there means the A: drive.
           */
          stat_buf->st_size = trusted_fsize;
          stat_buf->st_atime = stat_buf->st_ctime = stat_buf->st_mtime =
            _file_time_stamp(dos_ftime);

          /* Additional time info for LFN platforms.  */
          set_fstat_times(fhandle, stat_buf);
        }
      return 0;
    }

    /* Don't have even values from conventional DOS calls.
     * Give up completely on this funny handle.  ERRNO is already
     * set by filelength() and/or getftime().
     */
  else
    return -1;
}

/*
 * Main entry point.  This is a substitute for library fstat() function.
 */

int
fstat(int handle, struct stat *statbuf)
{
  int e = __dj_errno; /* save previous value of errno */
  __FSEXT_Function* func;
  int rv;

  if (!statbuf)
  {
    __dj_errno = 11;
    return -1;
  }

  /* see if this is file system extension file */
  func = __FSEXT_get_function(handle);
  if (func && __FSEXT_func_wrapper(func, __FSEXT_fstat, &rv, handle, statbuf))
  {
     return rv;
  }

  /* See if this is a file descriptor for a directory. If so, just
   * use a normal stat call. */
  if (__get_fd_flags(handle) & 0x20)
  {
    const char *filename = __get_fd_name(handle);

    if (filename)
      return stat(filename, statbuf);
  }

  if (fstat_assist(handle, statbuf) == -1)
  {
    return -1; /* already have ERRNO set by fstat_assist() */
  }
  else
  {
    __dj_errno = e;
    return 0;
  }
}
