# 0 "lstat.c"
# 0 "<built-in>"
# 0 "<command-line>"
# 1 "lstat.c"
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2010 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
/* This is file LSTAT.C */
/*
 *   Almost a 100% U**X-compatible stat()/lstat() substitute.
 *
 * Note:
 *
 *   The actual function implemented here is lstat(), not stat(). The
 *   stat() now is just a lstat() wrapper.
 *
 * Rationale:
 *
 *   Many Unix-born programs make heavy use of stat() library
 *   function to make decisions on files' equality, size, access
 *   attributes etc.  In the MS-DOS environment, many implementations
 *   of stat() are crippled, because DOS makes it very hard to get to
 *   certain pieces of information about files and directories.  Thus
 *   porting a program to DOS is usually an exercise in #ifdef'ing.
 *   This implementation facilitates porting Unix programs to MS-DOS
 *   by providing stat() which is much more Unix-compatible than those
 *   of most DOS-based C compilers (e.g., Borland's).
 *   Specifically, the following issues are taken care of:
 *
 *      1. This stat() doesn't fail for root directories, returning
 *         valid information.
 *      2. Directory size is not reported zero; the number of used
 *         directory entries multiplied by entry size is returned instead.
 *      3. Mode bits are set for all 3 groups (user, group, other).
 *      4. Directories are NOT reported read-only, unless one of R, H or S
 *         attributes is set.
 *      5. Directories have their execute bit set, as they do under Unix.
 *      6. Device names (such as /dev/con, lpt1, aux etc.) are treated as
 *         if they were on a special drive called `@:' (st_dev = -1).
 *         The "character special" mode bit is set for these devices.
 *      7. The inode number (st_ino) is taken from the starting cluster
 *         number of the file.  If the cluster number is unavailable, it
 *         is invented using the file's name in a manner that minimizes
 *         the possibility of inventing an inode which already belongs
 *         to another file.  See below for details.
 *      8. Executable files are found based on files' extensions and
 *         magic numbers present at their beginning, and their execute
 *         bits are set.
 *
 *   Lossage:
 *
 *      Beautiful as the above sounds, this implementation does fail
 *      under certain circumstances.  The following is a list of known
 *      problems:
 *
 *      1. The time fields for a root directory cannot be obtained, so
 *         they are set to the beginning of the Epoch.
 *      2. For files which reside on networked drives, the inode number
 *         is invented, because network redirectors usually do not
 *         bring that info with them.  This is not a total lossage, but
 *         it could get us a different inode for each program run.
 *      3. Empty files do not have a starting cluster number, because
 *         DOS doesn't allocate one until you actually write something
 *         to a file.  For these the inode is also invented.
 *      4. If the st_ino field is a 16 bit number, the invented inode
 *         numbers are from 65535 and down, assuming that most disks have
 *         unused portions near their end.  Valid cluster numbers are 16-bit
 *         unsigned integers, so a possibility of a clash exists, although
 *         the last 80 or more cluster numbers are unused on all drives
 *         I've seen.  If the st_ino is 32 bit, then invented inodes are
 *         all greater than 64k, which totally eliminates a possibility
 *         of a clash with an actual cluster number.
 *      5. The method of computing directory size is an approximation:
 *         a directory might consume much more space, if it has many
 *         deleted entries.  Still, this is a close approximation, and
 *         it does follow the logic of reporting size for a regular file:
 *         only the actually used space is returned.
 *      6. As this implementation relies heavily on undocumented DOS
 *         features, it will fail to get actual file info in environments
 *         other than native DOS, such as DR-DOS, OS/2 etc.  For these,
 *         the function will return whatever info is available with
 *         conventional DOS calls, which is no less than any other
 *         implementation could do.  This stat() might also fail for
 *         future DOS versions, if the layout of internal DOS data
 *         area is changed; however, this seems unlikely.
 *
 * Copyright (c) 1994-96 Eli Zaretskii <eliz@is.elta.co.il>
 *
 * This software may be used freely so long as this copyright notice is
 * left intact.  There is no warranty on this software.
 *
 */

/*
 * Tested with DJGPP port of GNU C compiler, versions 1.11maint5 and 1.12,
 * under MS-DOS 3.3, 4.01, 5.0, 6.20 (with and without DoubleSpace) and
 * with networked drives under XFS 1.86, Novell Netware 3.22, and
 * TSoft NFS 0.24Beta.
 *
 */

# 1 "../../../../../include/libc/stubs.h" 1
/* Copyright (C) 2007 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 18 "../../../../../include/libc/stubs.h"
/* POSIX functions (for when compiling an ANSI function) */
# 89 "../../../../../include/libc/stubs.h"
# 1 "../../../../../include/stddef.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 14 "../../../../../include/stddef.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 23 "../../../../../include/sys/djtypes.h"
/* For GCC 3.00 or later we use its builtin va_list.                       */
# 34 "../../../../../include/sys/djtypes.h"
/* but remains a typedef in C */





/* Define internal type to be used for avoiding strict
   aliasing warnings */

typedef short __attribute__((__may_alias__)) __dj_short_a;
typedef int __attribute__((__may_alias__)) __dj_int_a;
typedef long __attribute__((__may_alias__)) __dj_long_a;
typedef long long __attribute__((__may_alias__)) __dj_long_long_a;
typedef unsigned short __attribute__((__may_alias__)) __dj_unsigned_short_a;
typedef unsigned int __attribute__((__may_alias__)) __dj_unsigned_int_a;
typedef unsigned long __attribute__((__may_alias__)) __dj_unsigned_long_a;
typedef unsigned long long __attribute__((__may_alias__)) __dj_unsigned_long_long_a;
typedef float __attribute__((__may_alias__)) __dj_float_a;
typedef double __attribute__((__may_alias__)) __dj_double_a;
typedef long double __attribute__((__may_alias__)) __dj_long_double_a;
# 15 "../../../../../include/stddef.h" 2

/* Some programs think they know better... */
# 26 "../../../../../include/stddef.h"
//#ifdef __cplusplus
//#define offsetof(s_type, mbr) ((std::size_t) &((s_type *)0)->mbr)
//#else

//#endif


typedef int ptrdiff_t;




typedef long unsigned int size_t;




typedef unsigned short wchar_t;
# 90 "../../../../../include/libc/stubs.h" 2
# 1 "../../../../../include/dpmi.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 40 "../../../../../include/sys/cdefs.h"
/* Support gcc's __attribute__ facility.  */




/* Ensure that always traditional GNU extern inline semantics are used
   (aka -fgnu89-inline) even if ISO C99 semantics have been specified.  */
# 9 "../../../../../include/dpmi.h" 2
# 1 "../../../../../include/sys/nearptr.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */




# 1 "../../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 17 "../../../../../include/stdint.h"
typedef signed char int_least8_t;
typedef unsigned char uint_least8_t;
typedef signed char int_fast8_t;
typedef unsigned char uint_fast8_t;
typedef signed char int8_t;
typedef unsigned char uint8_t;

typedef signed short int int_least16_t;
typedef unsigned short int uint_least16_t;
typedef signed int int_fast16_t;
typedef unsigned int uint_fast16_t;
typedef signed short int int16_t;
typedef unsigned short int uint16_t;

typedef signed int int_least32_t;
typedef unsigned int uint_least32_t;
typedef signed int int_fast32_t;
typedef unsigned int uint_fast32_t;
typedef signed int int32_t;
typedef unsigned int uint32_t;

__extension__ typedef signed long long int int_least64_t;
__extension__ typedef unsigned long long int uint_least64_t;
__extension__ typedef signed long long int int_fast64_t;
__extension__ typedef unsigned long long int uint_fast64_t;
__extension__ typedef signed long long int int64_t;
__extension__ typedef unsigned long long int uint64_t;

typedef long int intptr_t;
typedef unsigned long int uintptr_t;

__extension__ typedef signed long long int intmax_t;
__extension__ typedef unsigned long long int uintmax_t;

/* ANSI/ISO C99 says these should not be visible in C++ unless
   explicitly requested.  */

//#if !defined(__cplusplus) || defined(__STDC_LIMIT_MACROS)
# 110 "../../../../../include/stdint.h"
  /* These are defined by limits.h, so make them conditional.  */
# 124 "../../../../../include/stdint.h"
//#endif /* !__cplusplus || __STDC_LIMIT_MACROS */
# 23 "../../../../../include/libc/djthunks.h" 2

void crt1_startup(int handle);

uint8_t *djaddr2ptr(uint32_t addr);
uint8_t *djaddr2ptr2(uint32_t addr, uint32_t len);
uint32_t djptr2addr(const uint8_t *ptr);
uint32_t djthunk_get(const char *name);
uint32_t djthunk_get_h(int handle, const char *name);
void *djsbrk(int increment);
# 20 "../../../../../include/libc/asmobj.h" 2
# 8 "../../../../../include/sys/nearptr.h" 2
# 24 "../../../../../include/sys/nearptr.h"
/* Functions to enable "near" pointer access to DOS memory under DPMI
   CW Sandmann 7-95  NO WARRANTY: WARNING, since these functions disable
   memory protection, they MAY DESTROY EVERYTHING ON YOUR COMPUTER! */

int __djgpp_nearptr_enable(void); /* Returns 0 if feature not avail */
void __djgpp_nearptr_disable(void); /* Enables protection */
extern int *____djgpp_selector_limit(void);/*
*/
# 30 "../../../../../include/sys/nearptr.h"
       #define __djgpp_selector_limit (*____djgpp_selector_limit())/*
*/
# 30 "../../../../../include/sys/nearptr.h"
                                       ; /* Limit on CS and on DS if prot */
extern unsigned *____djgpp_base_address(void);/*
*/
# 31 "../../../../../include/sys/nearptr.h"
       #define __djgpp_base_address (*____djgpp_base_address())/*
*/
# 31 "../../../../../include/sys/nearptr.h"
                                          ; /* Used in calculation below */
# 10 "../../../../../include/dpmi.h" 2
# 1 "../../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../../include/libc/asmobj.h" 2
# 11 "../../../../../include/dpmi.h" 2
# 1 "../../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 12 "../../../../../include/dpmi.h" 2
# 26 "../../../../../include/dpmi.h"
//#ifndef _POSIX_SOURCE

extern unsigned short *____dpmi_error(void);/*
*/
# 28 "../../../../../include/dpmi.h"
       #define __dpmi_error (*____dpmi_error())/*
*/
# 28 "../../../../../include/dpmi.h"
                                        ;

typedef struct {
  unsigned short offset16;
  unsigned short segment;
} __dpmi_raddr;

typedef struct {
  unsigned int offset32;
  unsigned short selector;
} __dpmi_paddr;

typedef struct {
  unsigned int handle; /* 0, 2 */
  unsigned int size; /* or count */ /* 4, 6 */
  unsigned int address; /* 8, 10 */
} __dpmi_meminfo;

typedef union {
  struct {
    unsigned int edi;
    unsigned int esi;
    unsigned int ebp;
    unsigned int res;
    unsigned int ebx;
    unsigned int edx;
    unsigned int ecx;
    unsigned int eax;
  } d;
  struct {
    unsigned short di, di_hi;
    unsigned short si, si_hi;
    unsigned short bp, bp_hi;
    unsigned short res, res_hi;
    unsigned short bx, bx_hi;
    unsigned short dx, dx_hi;
    unsigned short cx, cx_hi;
    unsigned short ax, ax_hi;
    unsigned short flags;
    unsigned short es;
    unsigned short ds;
    unsigned short fs;
    unsigned short gs;
    unsigned short ip;
    unsigned short cs;
    unsigned short sp;
    unsigned short ss;
  } x;
  struct {
    unsigned char edi[4];
    unsigned char esi[4];
    unsigned char ebp[4];
    unsigned char res[4];
    unsigned char bl, bh, ebx_b2, ebx_b3;
    unsigned char dl, dh, edx_b2, edx_b3;
    unsigned char cl, ch, ecx_b2, ecx_b3;
    unsigned char al, ah, eax_b2, eax_b3;
  } h;
} __attribute__((packed)) __dpmi_regs;

typedef struct {
  unsigned char major;
  unsigned char minor;
  unsigned short flags;
  unsigned char cpu;
  unsigned char master_pic;
  unsigned char slave_pic;
} __dpmi_version_ret;

typedef struct {
  unsigned int largest_available_free_block_in_bytes;
  unsigned int maximum_unlocked_page_allocation_in_pages;
  unsigned int maximum_locked_page_allocation_in_pages;
  unsigned int linear_address_space_size_in_pages;
  unsigned int total_number_of_unlocked_pages;
  unsigned int total_number_of_free_pages;
  unsigned int total_number_of_physical_pages;
  unsigned int free_linear_address_space_in_pages;
  unsigned int size_of_paging_file_partition_in_pages;
  unsigned int reserved[3];
} __dpmi_free_mem_info;

typedef struct {
  unsigned int total_allocated_bytes_of_physical_memory_host;
  unsigned int total_allocated_bytes_of_virtual_memory_host;
  unsigned int total_available_bytes_of_virtual_memory_host;
  unsigned int total_allocated_bytes_of_virtual_memory_vcpu;
  unsigned int total_available_bytes_of_virtual_memory_vcpu;
  unsigned int total_allocated_bytes_of_virtual_memory_client;
  unsigned int total_available_bytes_of_virtual_memory_client;
  unsigned int total_locked_bytes_of_memory_client;
  unsigned int max_locked_bytes_of_memory_client;
  unsigned int highest_linear_address_available_to_client;
  unsigned int size_in_bytes_of_largest_free_memory_block;
  unsigned int size_of_minimum_allocation_unit_in_bytes;
  unsigned int size_of_allocation_alignment_unit_in_bytes;
  unsigned int reserved[19];
} __dpmi_memory_info;

typedef struct {
  unsigned int data16[2];
  unsigned int code16[2];
  unsigned short ip;
  unsigned short reserved;
  unsigned int data32[2];
  unsigned int code32[2];
  unsigned int eip;
} __dpmi_callback_info;

typedef struct {
  unsigned int size_requested;
  unsigned int size;
  unsigned int handle;
  unsigned int address;
  unsigned int name_offset;
  unsigned short name_selector;
  unsigned short reserved1;
  unsigned int reserved2;
} __dpmi_shminfo;

/* Unless otherwise noted, all functions return -1 on error, setting __dpmi_error to the DPMI error code */

void __dpmi_yield(void); /* INT 0x2F AX=1680 */

int __dpmi_allocate_ldt_descriptors(int _count); /* DPMI 0.9 AX=0000 */
int __dpmi_free_ldt_descriptor(int _descriptor); /* DPMI 0.9 AX=0001 */
int __dpmi_segment_to_descriptor(int _segment); /* DPMI 0.9 AX=0002 */
int __dpmi_get_selector_increment_value(void); /* DPMI 0.9 AX=0003 */
int __dpmi_get_segment_base_address(int _selector, unsigned int *_addr); /* DPMI 0.9 AX=0006 */
int __dpmi_set_segment_base_address(int _selector, unsigned int _address); /* DPMI 0.9 AX=0007 */
unsigned int __dpmi_get_segment_limit(int _selector); /* LSL instruction  */
int __dpmi_set_segment_limit(int _selector, unsigned int _limit); /* DPMI 0.9 AX=0008 */
int __dpmi_get_descriptor_access_rights(int _selector); /* LAR instruction  */
int __dpmi_set_descriptor_access_rights(int _selector, int _rights); /* DPMI 0.9 AX=0009 */
int __dpmi_create_alias_descriptor(int _selector); /* DPMI 0.9 AX=000a */
int __dpmi_get_descriptor(int _selector, unsigned char _buffer[8]); /* DPMI 0.9 AX=000b */
int __dpmi_set_descriptor(int _selector, unsigned char _buffer[8]); /* DPMI 0.9 AX=000c */
int __dpmi_allocate_specific_ldt_descriptor(int _selector); /* DPMI 0.9 AX=000d */


int __dpmi_get_multiple_descriptors(int _count, void *_buffer); /* DPMI 1.0 AX=000e */
int __dpmi_set_multiple_descriptors(int _count, void *_buffer); /* DPMI 1.0 AX=000f */

int __dpmi_allocate_dos_memory(int _paragraphs, int *_ret_selector_or_max); /* DPMI 0.9 AX=0100 */
int __dpmi_free_dos_memory(int _selector); /* DPMI 0.9 AX=0101 */
int __dpmi_resize_dos_memory(int _selector, int _newpara, int *_ret_max); /* DPMI 0.9 AX=0102 */

int __dpmi_get_real_mode_interrupt_vector(int _vector, __dpmi_raddr *_address); /* DPMI 0.9 AX=0200 */
int __dpmi_set_real_mode_interrupt_vector(int _vector, __dpmi_raddr *_address); /* DPMI 0.9 AX=0201 */
int __dpmi_get_processor_exception_handler_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0202 */
int __dpmi_set_processor_exception_handler_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0203 */
int __dpmi_get_protected_mode_interrupt_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0204 */
int __dpmi_set_protected_mode_interrupt_vector(int _vector, __dpmi_paddr *_address); /* DPMI 0.9 AX=0205 */

int __dpmi_get_extended_exception_handler_vector_pm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0210 */
int __dpmi_get_extended_exception_handler_vector_rm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0211 */
int __dpmi_set_extended_exception_handler_vector_pm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0212 */
int __dpmi_set_extended_exception_handler_vector_rm(int _vector, __dpmi_paddr *_address); /* DPMI 1.0 AX=0213 */

int __dpmi_simulate_real_mode_interrupt(int _vector, __dpmi_regs *_regs); /* DPMI 0.9 AX=0300 */
int __dpmi_int(int _vector, __dpmi_regs *_regs); /* like above, but sets ss sp fl */ /* DPMI 0.9 AX=0300 */
extern short __dpmi_int_ss, __dpmi_int_sp, __dpmi_int_flags; /* default to zero */
int __dpmi_simulate_real_mode_procedure_retf(__dpmi_regs *_regs); /* DPMI 0.9 AX=0301 */
int __dpmi_simulate_real_mode_procedure_retf_stack(__dpmi_regs *_regs, int stack_words_to_copy, const void *stack_data); /* DPMI 0.9 AX=0301 */
int __dpmi_simulate_real_mode_procedure_iret(__dpmi_regs *_regs); /* DPMI 0.9 AX=0302 */
int __dpmi_allocate_real_mode_callback(unsigned int _handler, unsigned int _regs, __dpmi_raddr *_ret); /* DPMI 0.9 AX=0303 */
int __dpmi_free_real_mode_callback(__dpmi_raddr *_addr); /* DPMI 0.9 AX=0304 */
int __dpmi_get_state_save_restore_addr( __dpmi_raddr *_rm, __dpmi_paddr *_pm); /* DPMI 0.9 AX=0305 */
int __dpmi_get_raw_mode_switch_addr( __dpmi_raddr *_rm, __dpmi_paddr *_pm); /* DPMI 0.9 AX=0306 */

int __dpmi_get_version( __dpmi_version_ret *_ret); /* DPMI 0.9 AX=0400 */

int __dpmi_get_capabilities( int *_flags, char vendor_info[128]); /* DPMI 1.0 AX=0401 */

int __dpmi_get_free_memory_information( __dpmi_free_mem_info *_info); /* DPMI 0.9 AX=0500 */
int __dpmi_allocate_memory(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0501 */
int __dpmi_free_memory(unsigned int _handle); /* DPMI 0.9 AX=0502 */
int __dpmi_resize_memory(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0503 */

int __dpmi_allocate_linear_memory(__dpmi_meminfo *_info, int _commit); /* DPMI 1.0 AX=0504 */
int __dpmi_resize_linear_memory(__dpmi_meminfo *_info, int _commit); /* DPMI 1.0 AX=0505 */
int __dpmi_get_page_attributes(__dpmi_meminfo *_info, short *_buffer); /* DPMI 1.0 AX=0506 */
int __dpmi_set_page_attributes(__dpmi_meminfo *_info, short *_buffer); /* DPMI 1.0 AX=0507 */
int __dpmi_map_device_in_memory_block(__dpmi_meminfo *_info, unsigned int _physaddr); /* DPMI 1.0 AX=0508 */
int __dpmi_map_conventional_memory_in_memory_block(__dpmi_meminfo *_info, unsigned int _linaddr); /* DPMI 1.0 AX=0509 */
int __dpmi_get_memory_block_size_and_base(__dpmi_meminfo *_info); /* DPMI 1.0 AX=050a */
int __dpmi_get_memory_information( __dpmi_memory_info *_buffer); /* DPMI 1.0 AX=050b */

int __dpmi_lock_linear_region(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0600 */
int __dpmi_unlock_linear_region(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0601 */
int __dpmi_mark_real_mode_region_as_pageable(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0602 */
int __dpmi_relock_real_mode_region(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0603 */
int __dpmi_get_page_size( unsigned int *_size); /* DPMI 0.9 AX=0604 */

int __dpmi_mark_page_as_demand_paging_candidate(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0702 */
int __dpmi_discard_page_contents(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0703 */

int __dpmi_physical_address_mapping(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0800 */
int __dpmi_free_physical_address_mapping(__dpmi_meminfo *_info); /* DPMI 0.9 AX=0801 */

/* These next four functions return the old state */
int __dpmi_get_and_disable_virtual_interrupt_state(void); /* DPMI 0.9 AX=0900 */
int __dpmi_get_and_enable_virtual_interrupt_state(void); /* DPMI 0.9 AX=0901 */
int __dpmi_get_and_set_virtual_interrupt_state(int _old_state); /* DPMI 0.9 AH=09   */
int __dpmi_get_virtual_interrupt_state(void); /* DPMI 0.9 AX=0902 */

int __dpmi_get_vendor_specific_api_entry_point(const char *_id, __dpmi_paddr *_api); /* DPMI 0.9 AX=0a00 */

int __dpmi_set_debug_watchpoint(__dpmi_meminfo *_info, int _type); /* DPMI 0.9 AX=0b00 */
int __dpmi_clear_debug_watchpoint(unsigned int _handle); /* DPMI 0.9 AX=0b01 */
int __dpmi_get_state_of_debug_watchpoint(unsigned int _handle, int *_status); /* DPMI 0.9 AX=0b02 */
int __dpmi_reset_debug_watchpoint(unsigned int _handle); /* DPMI 0.9 AX=0b03 */

int __dpmi_install_resident_service_provider_callback(__dpmi_callback_info *_info); /* DPMI 1.0 AX=0c00 */
int __dpmi_terminate_and_stay_resident(int return_code, int paragraphs_to_keep); /* DPMI 1.0 AX=0c01 */

int __dpmi_allocate_shared_memory(__dpmi_shminfo *_info); /* DPMI 1.0 AX=0d00 */
int __dpmi_free_shared_memory(unsigned int _handle); /* DPMI 1.0 AX=0d01 */
int __dpmi_serialize_on_shared_memory(unsigned int _handle, int _flags); /* DPMI 1.0 AX=0d02 */
int __dpmi_free_serialization_on_shared_memory(unsigned int _handle, int _flags); /* DPMI 1.0 AX=0d03 */

int __dpmi_get_coprocessor_status(void); /* DPMI 1.0 AX=0e00 */
int __dpmi_set_coprocessor_emulation(int _flags); /* DPMI 1.0 AX=0e01 */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Backwards compatibility stuff						       */



typedef struct {
  unsigned int available_memory;
  unsigned int available_pages;
  unsigned int available_lockable_pages;
  unsigned int linear_space;
  unsigned int unlocked_pages;
  unsigned int available_physical_pages;
  unsigned int total_physical_pages;
  unsigned int free_linear_space;
  unsigned int max_pages_in_paging_file;
  unsigned int reserved[3];
} _go32_dpmi_meminfo;







typedef struct {
  unsigned int size;
  unsigned int pm_offset;
  unsigned short pm_selector;
  unsigned short rm_offset;
  unsigned short rm_segment;
} _go32_dpmi_seginfo;

/* returns zero if success, else dpmi error and info->size is max size */
int _go32_dpmi_allocate_dos_memory(_go32_dpmi_seginfo *info);
 /* set size to bytes/16, call, use rm_segment.  Do not
	   change anthing but size until the memory is freed.
	   If error, max size is returned in size as bytes/16. */
int _go32_dpmi_free_dos_memory(_go32_dpmi_seginfo *info);
 /* set new size to bytes/16, call.  If error, max size
	   is returned in size as bytes/16 */
int _go32_dpmi_resize_dos_memory(_go32_dpmi_seginfo *info);
 /* uses pm_selector to free memory */

/* These both use the rm_segment:rm_offset fields only */
int _go32_dpmi_get_real_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
int _go32_dpmi_set_real_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);

/* These do NOT wrap the function in pm_offset in an iret handler.
   You must provide an assembler interface yourself, or alloc one below.
   You may NOT LONGjmp out of an interrupt handler. */
int _go32_dpmi_get_protected_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
 /* puts vector in pm_selector:pm_offset. */
int _go32_dpmi_set_protected_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
 /* sets vector from pm_offset and pm_selector */

/********** HELPER FUNCTIONS **********/

int _go32_dpmi_chain_protected_mode_interrupt_vector(int vector, _go32_dpmi_seginfo *info);
 /* sets up wrapper that calls function in pm_offset, chaining to old
	   handler when it returns */

/* These generate assember IRET-style wrappers for functions and set up stack */
int _go32_dpmi_allocate_iret_wrapper(_go32_dpmi_seginfo *info);
 /* Put function ptr in pm_offset, call, returns wrapper entry in pm_offset. */
int _go32_dpmi_free_iret_wrapper(_go32_dpmi_seginfo *info);
 /* assumes pm_offset points to wrapper, frees it */

/* RMCB functions, automatically restructure the real-mode stack for the
   proper return type and set up correct PM stack.  The callback
   (info->pm_offset) is called as (*pmcb)(_go32_dpmi_registers *regs); */
int _go32_dpmi_allocate_real_mode_callback_retf(_go32_dpmi_seginfo *info, unsigned int regs);
 /* points callback at pm_offset, returns seg:ofs of callback addr
	   in rm_segment:rm_offset.  Do not change any fields until freed.
	   Interface is added to simulate far return */
int _go32_dpmi_allocate_real_mode_callback_iret(_go32_dpmi_seginfo *info, unsigned int regs);
 /* same, but simulates iret */
int _go32_dpmi_free_real_mode_callback(_go32_dpmi_seginfo *info);
 /* frees callback */

/* The following two variables may be used to change the default stack size
   for interrupts and rmcb wrappers to a user defined size from the default
   of 32Kbytes.  Each RMCB and chain/iret wrapper gets its own stack. */

extern unsigned int _go32_interrupt_stack_size;
extern unsigned int _go32_rmcb_stack_size;

/* Convenience functions, the return value is *bytes* */
unsigned int _go32_dpmi_remaining_physical_memory(void);
unsigned int _go32_dpmi_remaining_virtual_memory(void);

/* locks memory from a specified offset within the code/data selector */
int _go32_dpmi_lock_code( unsigned int _lockaddr, unsigned int _locksize);
int _go32_dpmi_lock_data( unsigned int _lockaddr, unsigned int _locksize);

int __djgpp_set_page_attributes(void *our_addr, unsigned int num_bytes,
           unsigned short attributes);
int __djgpp_map_physical_memory(void *our_addr, unsigned int num_bytes,
           uint64_t phys_addr);

//#endif /* !_POSIX_SOURCE */
# 91 "../../../../../include/libc/stubs.h" 2
# 1 "../../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../../include/libc/asmobj.h" 2
# 92 "../../../../../include/libc/stubs.h" 2
int __spawnve(int mode, const char *path, char *const argv[],
              char *const envp[]);
int __chdir (const char *mydirname);
char * __getcwd(char *buf, size_t size);
int __brk(void *_heaptop);
unsigned int __sbrk(int delta);
void *sbrk(int delta);
long double __modfl(long double _x, long double *_pint);

extern unsigned *____djgpp_timer_countdown(void);/*
*/
# 101 "../../../../../include/libc/stubs.h"
       #define __djgpp_timer_countdown (*____djgpp_timer_countdown())/*
*/
# 101 "../../../../../include/libc/stubs.h"
                                             ;
extern int *____djgpp_timer_hdlr(void);/*
*/
# 102 "../../../../../include/libc/stubs.h"
       #define __djgpp_timer_hdlr (*____djgpp_timer_hdlr())/*
*/
# 102 "../../../../../include/libc/stubs.h"
                                   ;
# 106 "lstat.c" 2
# 1 "../../../../../include/stdlib.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 20 "../../../../../include/stdlib.h"
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../../include/sys/types.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../../include/sys/types.h" 2

typedef int blkcnt_t;
typedef int blksize_t;
typedef int dev_t;
typedef unsigned long fsblkcnt_t;
typedef unsigned long fsfilcnt_t;
typedef int ino_t;
typedef int mode_t;
typedef int nlink_t;


typedef int gid_t;



typedef int off_t;



__extension__ typedef long long off64_t;



__extension__ typedef long long offset_t;



typedef int pid_t;







typedef long signed int ssize_t;




typedef int uid_t;




/* Allow including program to override.  */




typedef struct fd_set {
  unsigned char fd_bits [((256) + 7) / 8];
} fd_set;







typedef unsigned int time_t;
# 21 "../../../../../include/stdlib.h" 2
# 1 "../../../../../include/stddef.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 22 "../../../../../include/stdlib.h" 2

/* Some programs think they know better... */
# 38 "../../../../../include/stdlib.h"
extern int __dj_mb_cur_max;

typedef struct {
  int quot;
  int rem;
} div_t;

typedef struct {
  long quot;
  long rem;
} ldiv_t;





void abort(void);
int abs(int _i);

int __dj_atexit(void (*_func)(void));
double atof(const char *_s);
int atoi(const char *_s);
long atol(const char *_s);
void * bsearch(const void *_key, const void *_base, size_t _nelem,
  size_t _size, int (*_cmp)(const void *_ck, const void *_ce));
void * calloc(size_t _nelem, size_t _size);
unsigned int calloc32(size_t _nelem, size_t _size);
div_t div(int _numer, int _denom);
void exit(int _status);
void free(void *_ptr);
void free32(unsigned int _addr);
char * getenv(const char *_name);
long labs(long _i);
ldiv_t ldiv(long _numer, long _denom);
void * malloc(size_t _size);
unsigned int malloc32(size_t _size);
int mblen(const char *_s, size_t _n);
size_t mbstowcs(wchar_t *_wcs, const char *_s, size_t _n);
int mbtowc(wchar_t *_pwc, const char *_s, size_t _n);
void qsort(void *_base, size_t _nelem, size_t _size,
       int (*_cmp)(const void *_e1, const void *_e2));
int rand(void);
void * realloc(void *_ptr, size_t _size);
unsigned int realloc32(unsigned int _addr, size_t _size);
void srand(unsigned _seed);
double strtod(const char *_s, char **_endptr);
long strtol(const char *_s, char **_endptr, int _base);
unsigned long strtoul(const char *_s, char **_endptr, int _base);
int system(const char *_s);
size_t wcstombs(char *_s, const wchar_t *_wcs, size_t _n);
int wctomb(char *_s, wchar_t _wchar);




typedef struct {
  long long int quot;
  long long int rem;
} lldiv_t;

void _Exit(int _status);
long long int atoll(const char *_s);
long long int llabs(long long int _i);
lldiv_t lldiv(long long int _numer, long long int _denom);
float strtof(const char *_s, char **_endptr);
long double strtold(const char *_s, char **_endptr);
long long int strtoll(const char *_s, char **_endptr, int _base);
unsigned long long int strtoull(const char *_s, char **_endptr, int _base);





long a64l(const char *_string);
char * l64a(long _value);
char * mktemp(char *_template);
char * mkdtemp(char *_template);
int mkstemp(char *_template);
int putenv(char *_val);
char * realpath(const char *_path, char *_resolved);
int setenv(const char *_var, const char *_val, int _overwrite);
int unsetenv(const char *_var);



void * alloca(size_t _size);
long double _atold(const char *_s);
void cfree(void *_ptr);
double drand48(void);
char * ecvtbuf(double _val, int _nd, int *_dp, int *_sn, char *_bf);
char * ecvt(double _val, int _nd, int *_dp, int *_sn);
double erand48(unsigned short _state[3]);
char * fcvtbuf(double _val, int _nd, int *_dp, int *_sn, char *_bf);
char * fcvt(double _val, int _nd, int *_dp, int *_sn);
char * gcvt(double _val, int _nd, char *_buf);
char * getpass(const char *_prompt);
int getlongpass(const char *_prompt, char *_buffer, int _max_len);
char * itoa(int _value, char *_buffer, int _radix);
long jrand48(unsigned short _state[3]);
void lcong48(unsigned short _param[7]);
unsigned long lrand48(void);
long mrand48(void);
unsigned long nrand48(unsigned short _state[3]);
unsigned short *seed48(unsigned short _state_seed[3]);
void srand48(long _seedval);
long double _strtold(const char *_s, char **_endptr);
void swab(const void *_from, void *_to, ssize_t _nbytes);
void * valloc (size_t _amt);





/* BSD Random Number Generator */
char * initstate (unsigned _seed, char *_arg_state, int _n);
char * setstate(char *_arg_state);
long random(void);
int srandom(int _seed);

/* These vary in expected prototype, so we just don't prototype them.
void		xfree(void *_ptr);
void *		xmalloc(size_t _size);
void *		xrealloc(void *ptr, size_t _size);
*/
# 173 "../../../../../include/stdlib.h"
extern int __system_flags;
# 107 "lstat.c" 2
# 1 "../../../../../include/stddef.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 108 "lstat.c" 2
# 1 "../../../../../include/unistd.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 19 "../../../../../include/unistd.h"
extern char **__dj_environ;
# 30 "../../../../../include/unistd.h"
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 31 "../../../../../include/unistd.h" 2
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 32 "../../../../../include/unistd.h" 2





/* `lockf' is a simpler interface to the locking facilities of `fcntl'.
   LEN is always relative to the current file position.
   The CMD argument is one of the following.  */






/* Some programs think they know better... */
# 112 "../../../../../include/unistd.h"
extern char *optarg;
extern int optind, opterr, optopt;

void __exit(int _status);
void _exit(int _status);
int access(const char *_path, int _amode);
unsigned int alarm(unsigned int _seconds);
int chdir(const char *_path);
int chown(const char *_path, uid_t _owner, gid_t _group);
int close(int _fildes);
size_t confstr(int _name, char *_buf, size_t _len);
char * ctermid(char *_s);
int dup(int _fildes);
int dup2(int _fildes, int _fildes2);
int execl(const char *_path, const char *_arg, ...);
int execle(const char *_path, const char *_arg, ...);
int execlp(const char *_file, const char *_arg, ...);
int execv(const char *_path, char *const _argv[]);
int execve(const char *_path, char *const _argv[], char *const _envp[]);
int execvp(const char *_file, char *const _argv[]);
int fchdir(int _fd);
pid_t fork(void);
long fpathconf(int _fildes, int _name);
char * getcwd(char *_buf, size_t _size);
gid_t getegid(void);
uid_t geteuid(void);
gid_t getgid(void);
int getgroups(int _gidsetsize, gid_t *_grouplist);
char * getlogin(void);
int getopt(int _argc, char *const _argv[], const char *_optstring);
pid_t getpgrp(void);
pid_t getpid(void);
pid_t getppid(void);
uid_t getuid(void);
int isatty(int _fildes);
int link(const char *_existing, const char *_new);
off_t lseek(int _fildes, off_t _offset, int _whence);
long pathconf(const char *_path, int _name);
int pause(void);
int pipe(int _fildes[2]);
ssize_t pwrite(int _fildes, const void *_buf, size_t _nbyte, off_t _offset);
ssize_t read(int _fildes, void *_buf, size_t _nbyte);
int rmdir(const char *_path);
int setgid(gid_t _gid);
int setpgid(pid_t _pid, pid_t _pgid);
pid_t setsid(void);
int setuid(uid_t _uid);
unsigned int sleep(unsigned int _seconds);
long sysconf(int _name);
pid_t tcgetpgrp(int _fildes);
int tcsetpgrp(int _fildes, pid_t _pgrp_id);
char * ttyname(int _fildes);
int unlink(const char *_path);
ssize_t write(int _fildes, const void *_buf, size_t _nbyte);



/* additional access() checks */


char * basename(const char *_fn);
int brk(void *_heaptop);
char * dirname(const char *_fn);
int __file_exists(const char *_fn);
int fchown(int fd, uid_t owner, gid_t group);
int fsync(int _fd);
int ftruncate(int, off_t);
int getdtablesize(void);
int gethostname(char *buf, int size);
int getpagesize(void);
char * getwd(char *__buffer);
int lchown(const char * file, int owner, int group);
int lockf(int _fildes, int _cmd, off_t _len);
int llockf(int _fildes, int _cmd, offset_t _len);
offset_t llseek(int _fildes, offset_t _offset, int _whence);
int nice(int _increment);
int readlink(const char * __file, char * __buffer, size_t __size);
void * sbrk(int _delta);
int symlink (const char *, const char *);
int sync(void);
int truncate(const char*, off_t);
unsigned int usleep(unsigned int _useconds);

pid_t vfork(void);
# 109 "lstat.c" 2
# 1 "../../../../../include/time.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 18 "../../../../../include/time.h"
/* 65536(tics/hour) / 3600(sec/hour) * 5(scale) = 91.02
   The 5 is to make it a whole number (18.2*5=91) so that
   floating point ops aren't required to use it. */


# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../../include/time.h" 2

/* Some programs think they know better... */
# 36 "../../../../../include/time.h"
typedef int clock_t;
# 48 "../../../../../include/time.h"
struct tm {
  int tm_sec;
  int tm_min;
  int tm_hour;
  int tm_mday;
  int tm_mon;
  int tm_year;
  int tm_wday;
  int tm_yday;
  int tm_isdst;
  char *__tm_zone;
  int __tm_gmtoff;
};

char * asctime(const struct tm *_tptr);
clock_t clock(void);
char * ctime(const time_t *_cal);
double difftime(time_t _t1, time_t _t0);
struct tm * gmtime(const time_t *_tod);
struct tm * localtime(const time_t *_tod);
time_t mktime(struct tm *_tptr);
size_t strftime(char * _s, size_t _n, const char * _format, const struct tm * _tptr);
time_t time(time_t *_tod);
# 81 "../../../../../include/time.h"
extern char *tzname[2];

char * asctime_r(const struct tm * __restrict__ _tptr, char * __restrict__ _buf);
char * ctime_r(const time_t *_cal, char *_buf);
struct tm * gmtime_r(const time_t * __restrict__ _tod, struct tm * __restrict__ _tptr);
struct tm * localtime_r(const time_t * __restrict__ _tod, struct tm * __restrict__ _tptr);
void tzset(void);






struct timeval {
  time_t tv_sec;
  long tv_usec;
};

struct timezone {
  int tz_minuteswest;
  int tz_dsttime;
};

# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 105 "../../../../../include/time.h" 2

typedef long long uclock_t;


int gettimeofday(struct timeval *_tp, struct timezone *_tzp);
unsigned long rawclock(void);
int select(int _nfds, fd_set *_readfds, fd_set *_writefds, fd_set *_exceptfds, struct timeval *_timeout);
int settimeofday(struct timeval *_tp, ...);
void tzsetwall(void);
uclock_t uclock(void);


/* GCC-4.8.0 and newer versions defines _rdtsc as macro in ia32intrin.h
   included from x86intrin.h (for MMX/SSE/AVX instructions. */



# 1 "../../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 123 "../../../../../include/time.h" 2

unsigned long long _rdtsc(void);
# 110 "lstat.c" 2
# 1 "../../../../../include/stdio.h" 1
/* Copyright (C) 2018 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2017 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 21 "../../../../../include/stdio.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 22 "../../../../../include/stdio.h" 2





/* Some programs think they know better... */
# 49 "../../../../../include/stdio.h"
typedef __builtin_va_list va_list;
# 63 "../../../../../include/stdio.h"
/* Note that the definitions of these fields are NOT guaranteed!  They
   may change with any release without notice!  The fact that they
   are here at all is to comply with ANSI specifictions. */

typedef struct {
  ssize_t _cnt;
  char *_ptr;
  char *_base;
  size_t _bufsiz;
  int _flag;
  int _file;
  char *_name_to_remove;
  size_t _fillsize;
} FILE;

typedef unsigned long fpos_t;

extern FILE __dj_stdin, __dj_stdout, __dj_stderr;




void clearerr(FILE *_stream);
int fclose(FILE *_stream);
int feof(FILE *_stream);
int ferror(FILE *_stream);
int fflush(FILE *_stream);
int fgetc(FILE *_stream);
int fgetpos(FILE *_stream, fpos_t *_pos);
char * fgets(char *_s, int _n, FILE *_stream);
FILE * fopen(const char *_filename, const char *_mode);
int fprintf(FILE *_stream, const char *_format, ...);
int fputc(int _c, FILE *_stream);
int fputs(const char *_s, FILE *_stream);
size_t fread(void *_ptr, size_t _size, size_t _nelem, FILE *_stream);
FILE * freopen(const char *_filename, const char *_mode, FILE *_stream);
int fscanf(FILE *_stream, const char *_format, ...);
int fseek(FILE *_stream, long _offset, int _mode);
int fsetpos(FILE *_stream, const fpos_t *_pos);
long ftell(FILE *_stream);
size_t fwrite(const void *_ptr, size_t _size, size_t _nelem, FILE *_stream);
int getc(FILE *_stream);
int getchar(void);
char * gets(char *_s);
void perror(const char *_s);
int printf(const char *_format, ...);
int putc(int _c, FILE *_stream);
int putchar(int _c);
int puts(const char *_s);
int remove(const char *_filename);
int rename(const char *_old, const char *_new);
void rewind(FILE *_stream);
int scanf(const char *_format, ...);
void setbuf(FILE *_stream, char *_buf);
int setvbuf(FILE *_stream, char *_buf, int _mode, size_t _size);
int sprintf(char *_s, const char *_format, ...);
int sscanf(const char *_s, const char *_format, ...);
FILE * tmpfile(void);
char * tmpnam(char *_s);
int ungetc(int _c, FILE *_stream);
int vfprintf(FILE *_stream, const char *_format, va_list _ap);
int vprintf(const char *_format, va_list _ap);
int vsprintf(char *_s, const char *_format, va_list _ap);




int snprintf(char *str, size_t n, const char *fmt, ...);
int vfscanf(FILE *_stream, const char *_format, va_list _ap);
int vscanf(const char *_format, va_list _ap);
int vsnprintf(char *str, size_t n, const char *fmt, va_list ap);
int vsscanf(const char *_s, const char *_format, va_list _ap);







/* #define STREAM_MAX	20 - DOS can change this */

int dprintf(int _fd, const char *_format, ...) __attribute__ ((__format__ (__printf__, 2, 3)));
int fileno(FILE *_stream);
FILE * fdopen(int _fildes, const char *_type);
int mkstemp(char *_template);
int pclose(FILE *_pf);
FILE * popen(const char *_command, const char *_mode);
char * tempnam(const char *_dir, const char *_prefix);
int vdprintf(int _fd, const char *_format, va_list _ap) __attribute__ ((__format__ (__printf__, 2, 0)));



extern FILE __dj_stdprn, __dj_stdaux;





void _djstat_describe_lossage(FILE *_to_where);
int _doprnt(const char *_fmt, va_list _args, FILE *_f);
int _doscan(FILE *_f, const char *_fmt, va_list _args);
int _doscan_low(FILE *, int (*)(FILE *_get), int (*_unget)(int, FILE *), const char *_fmt, va_list _args);
int fpurge(FILE *_f);
int getw(FILE *_f);
char * mktemp(char *_template);
int putw(int _v, FILE *_f);
void setbuffer(FILE *_f, void *_buf, int _size);
void setlinebuf(FILE *_f);
int _rename(const char *_old, const char *_new); /* Simple (no directory) */
int asprintf(char **_sp, const char *_format, ...) __attribute__((format (__printf__, 2, 3)));
char * asnprintf(char *_s, size_t *_np, const char *_format, ...) __attribute__((format (__printf__, 3, 4)));
int vasprintf(char **_sp, const char *_format, va_list _ap) __attribute__((format (__printf__, 2, 0)));
char * vasnprintf(char *_s, size_t *_np, const char *_format, va_list _ap) __attribute__((format (__printf__, 3, 0)));
# 185 "../../../../../include/stdio.h"
int fseeko(FILE *_stream, off_t _offset, int _mode);
off_t ftello(FILE *_stream);
int fseeko64(FILE *_stream, off64_t _offset, int _mode);
off64_t ftello64(FILE *_stream);
# 111 "lstat.c" 2
# 1 "../../../../../include/string.h" 1
/* Copyright (C) 2017 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 17 "../../../../../include/string.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 18 "../../../../../include/string.h" 2

/* Some programs think they know better... */
# 34 "../../../../../include/string.h"
void * memchr(const void *_s, int _c, size_t _n);
int memcmp(const void *_s1, const void *_s2, size_t _n);
void * memcpy(void * _dest, const void * _src, size_t _n);
void * memmove(void *_s1, const void *_s2, size_t _n);
void * memset(void *_s, int _c, size_t _n);
char * strcat(char * _s1, const char * _s2);
char * strchr(const char *_s, int _c);
int strcmp(const char *_s1, const char *_s2);
int strcoll(const char *_s1, const char *_s2);
char * strcpy(char * _s1, const char * _s2);
size_t strcspn(const char *_s1, const char *_s2);
char * strerror(int _errcode);
size_t strlen(const char *_s);
char * strncat(char * _s1, const char * _s2, size_t _n);
int strncmp(const char *_s1, const char *_s2, size_t _n);
char * strncpy(char * _s1, const char * _s2, size_t _n);
char * strpbrk(const char *_s1, const char *_s2);
char * strrchr(const char *_s, int _c);
size_t strspn(const char *_s1, const char *_s2);
char * strstr(const char *_s1, const char *_s2);
char * strtok(char * _s1, const char * _s2);
size_t strxfrm(char * _s1, const char * _s2, size_t _n);
# 64 "../../../../../include/string.h"
int strerror_r(int _errnum, char *_strerrbuf, size_t _buflen);
char * strtok_r(char * _s1, const char * _s2, char ** _s3);



# 1 "../../../../../include/sys/movedata.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../../include/sys/movedata.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../../include/sys/movedata.h" 2
# 1 "../../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../../include/sys/movedata.h" 2

# 1 "../../../../../include/sys/fmemcpy.h" 1



# 1 "../../../../../include/dpmi.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 5 "../../../../../include/sys/fmemcpy.h" 2

void fmemcpy1(__dpmi_paddr dst, const void *src, unsigned len);
void fmemcpy2(void *dst, __dpmi_paddr src, unsigned len);
void fmemcpy12(__dpmi_paddr dst, __dpmi_paddr src, unsigned len);
# 28 "../../../../../include/sys/movedata.h" 2
# 36 "../../../../../include/sys/movedata.h"
/* This header is intended to be included only by other
   headers, like <go32.h> and <string.h>.  You may
   include this directly, but it will be non-portable. */

/* These lengths are in bytes, optimized for speed */
void dosmemget(unsigned long _offset, size_t _length, void *_buffer);
void dosmemput(const void *_buffer, size_t _length, unsigned long _offset);

/* The lengths here are in TRANSFERS, not bytes! */
void _dosmemgetb(unsigned long _offset, size_t _xfers, void *_buffer);
void _dosmemgetw(unsigned long _offset, size_t _xfers, void *_buffer);
void _dosmemgetl(unsigned long _offset, size_t _xfers, void *_buffer);
void _dosmemputb(const void *_buffer, size_t _xfers, unsigned long _offset);
void _dosmemputw(const void *_buffer, size_t _xfers, unsigned long _offset);
void _dosmemputl(const void *_buffer, size_t _xfers, unsigned long _offset);
# 70 "../../../../../include/string.h" 2

int bcmp(const void *_ptr1, const void *_ptr2, size_t _length);
void bcopy(const void *_a, void *_b, size_t _len);
void bzero(void *ptr, size_t _len);
int ffs(int _mask);
char * index(const char *_string, int _c);
void * memccpy(void *_to, const void *_from, int _c, size_t _n);
int memicmp(const void *_s1, const void *_s2, size_t _n);
void * memmem(const void *l, size_t l_len, const void *s, size_t s_len);
char * rindex(const char *_string, int _c);
char * stpcpy(char *_dest, const char *_src);
char * stpncpy(char *_dest, const char *_src, size_t _n);
char * strdup(const char *_s);
char * strndup(const char *_s, size_t _n);
size_t strlcat(char *_dest, const char *_src, size_t _size);
size_t strlcpy(char *_dest, const char *_src, size_t _size);
char * strlwr(char *_s);
int strcasecmp(const char *_s1, const char *_s2);
int stricmp(const char *_s1, const char *_s2);
int strncasecmp(const char *_s1, const char *_s2, size_t _n);
int strnicmp(const char *_s1, const char *_s2, size_t _n);
size_t strnlen(const char *_s, size_t _n);
char * strsep(char **_stringp, const char *_delim);
char * strupr(char *_s);
# 112 "lstat.c" 2
# 1 "../../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 113 "lstat.c" 2
# 1 "../../../../../include/ctype.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 14 "../../../../../include/ctype.h"
int isalnum(int c);
int isalpha(int c);
int iscntrl(int c);
int isdigit(int c);
int isgraph(int c);
int islower(int c);
int isprint(int c);
int ispunct(int c);
int isspace(int c);
int isupper(int c);
int isxdigit(int c);
int tolower(int c);
int toupper(int c);




int isblank(int c);




# 1 "../../../../../include/inlines/ctype.ha" 1
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
# 18 "../../../../../include/inlines/ctype.ha"
extern unsigned short __dj_ctype_flags[];
extern unsigned char __dj_ctype_toupper[];
extern unsigned char __dj_ctype_tolower[];
# 37 "../../../../../include/ctype.h" 2
# 46 "../../../../../include/ctype.h"
int _tolower(int c);
int _toupper(int c);


# 1 "../../../../../include/inlines/ctype.hp" 1
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */



extern unsigned char __dj_ctype_toupper[];
extern unsigned char __dj_ctype_tolower[];
# 51 "../../../../../include/ctype.h" 2




int isascii(int c);
int toascii(int c);


# 1 "../../../../../include/inlines/ctype.hd" 1
# 60 "../../../../../include/ctype.h" 2
# 114 "lstat.c" 2
# 1 "../../../../../include/errno.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */







# 1 "../../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../../include/libc/asmobj.h" 2
# 16 "../../../../../include/errno.h" 2







extern int *____dj_errno(void);/*
*/
# 23 "../../../../../include/errno.h"
       #define __dj_errno (*____dj_errno())/*
*/
# 23 "../../../../../include/errno.h"
                             ;
# 77 "../../../../../include/errno.h"
/* EILSEQ		41 */

extern char * sys_errlist[];
extern int sys_nerr;
extern const char * __sys_errlist[];
extern int __sys_nerr;
extern int _doserrno;
# 115 "lstat.c" 2
# 1 "../../../../../include/fcntl.h" 1
/* Copyright (C) 2016 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2009 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 60 "../../../../../include/fcntl.h"
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 61 "../../../../../include/fcntl.h" 2

struct flock {
  off_t l_len;
  pid_t l_pid;
  off_t l_start;
  short l_type;
  short l_whence;
};

struct flock64 {
  offset_t l_len;
  pid_t l_pid;
  offset_t l_start;
  short l_type;
  short l_whence;
};

extern int _fmode; /* O_TEXT or O_BINARY */

int open(const char *_path, int _oflag, ...);
int creat(const char *_path, mode_t _mode);
int fcntl(int _fildes, int _cmd, ...);



/* Additional non-POSIX flags for open(). */
/* They are present on GNU libc. */
# 105 "../../../../../include/fcntl.h"
extern int __djgpp_share_flags;





/*
 *  For compatibility with other DOS C compilers.
 */
# 127 "../../../../../include/fcntl.h"
/*
 * Support for advanced filesystems (Windows 9x VFAT, NTFS, LFN etc.)
 */
# 138 "../../../../../include/fcntl.h"
unsigned _get_volume_info (const char *_path, int *_max_file_len, int *_max_path_len, char *_filesystype);
char _use_lfn (const char *_path);
char *_lfn_gen_short_fname (const char *_long_fname, char *_short_fname);
int _is_DOS83 (const char *_fname);




unsigned _lfn_get_ftime (int _handle, int _which);

char _preserve_fncase (void);
# 116 "lstat.c" 2
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 117 "lstat.c" 2
# 1 "../../../../../include/sys/stat.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 44 "../../../../../include/sys/stat.h"
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 45 "../../../../../include/sys/stat.h" 2
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 46 "../../../../../include/sys/stat.h" 2





struct stat {
  time_t st_atime;
  time_t st_ctime;
  dev_t st_dev;
  gid_t st_gid;
  ino_t st_ino;
  mode_t st_mode;
  time_t st_mtime;
  nlink_t st_nlink;
  off_t st_size;
  blksize_t st_blksize;
  uid_t st_uid;
  dev_t st_rdev;
};

int chmod(const char *_path, mode_t _mode);
int fchmod(int _fildes, mode_t _mode);
int fstat(int _fildes, struct stat *_buf);
int mkdir(const char *_path, mode_t _mode);
int mkfifo(const char *_path, mode_t _mode);
int stat(const char *_path, struct stat *_buf);
mode_t umask(mode_t _cmask);



/* POSIX.1 doesn't mention these at all */
# 93 "../../../../../include/sys/stat.h"
void _fixpath(const char *, char *);
char * __canonicalize_path(const char *, char *, size_t);
unsigned short _get_magic(const char *, int);
int _is_executable(const char *, int, const char *);
int lstat(const char * _path, struct stat * _buf);
int mknod(const char *_path, mode_t _mode, dev_t _dev);
char * _truename(const char *, char *);
char * _truename_sfn(const char *, char *);

/* Bit-mapped variable _djstat_flags describes what expensive
   f?stat() features our application needs.  If you don't need a
   feature, set its bit in the variable.  By default, all the
   bits are cleared (i.e., you get the most expensive code).  */







extern unsigned short _djstat_flags;

/* Bit-mapped variable _djstat_fail_bits describes which individual
   undocumented features f?stat() failed to use.  To get a human-
   readable description of the bits, call _djstat_describe_lossage(). */
# 131 "../../../../../include/sys/stat.h"
extern unsigned short _djstat_fail_bits;
# 118 "lstat.c" 2
# 1 "../../../../../include/dos.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 12 "../../../../../include/dos.h" 2
# 24 "../../../../../include/dos.h"
# 1 "../../../../../include/pc.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */







# 1 "../../../../../include/stdint.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
# 15 "../../../../../include/pc.h" 2
# 27 "../../../../../include/pc.h"
unsigned char inportb (unsigned short _port);
unsigned short inportw (unsigned short _port);
uint32_t inportl (unsigned short _port);
void inportsb (unsigned short _port, unsigned char *_buf, unsigned _len);
void inportsw (unsigned short _port, unsigned short *_buf, unsigned _len);
void inportsl (unsigned short _port, uint32_t *_buf, unsigned _len);
void outportb (unsigned short _port, unsigned char _data);
void outportw (unsigned short _port, unsigned short _data);
void outportl (unsigned short _port, uint32_t _data);
void outportsb(unsigned short _port, const unsigned char *_buf, unsigned _len);
void outportsw(unsigned short _port, const unsigned short *_buf, unsigned _len);
void outportsl(unsigned short _port, const uint32_t *_buf, unsigned _len);

unsigned char inp(unsigned short _port);
unsigned short inpw(unsigned short _port);
void outp(unsigned short _port, unsigned char _data);
void outpw(unsigned short _port, unsigned short _data);

int kbhit(void);

int getkey(void); /* ALT's have 0x100 set */
int getxkey(void); /* ALT's have 0x100 set, 0xe0 sets 0x200 */
const unsigned char * __get_extended_key_string(int _xkey_code);
void sound(int _frequency);



extern unsigned char ScreenAttrib;




int ScreenMode(void);
int ScreenRows(void);
int ScreenCols(void);
void ScreenPutChar(int _ch, int _attr, int _x, int _y);
void ScreenGetChar(int *_ch, int *_attr, int _x, int _y);
void ScreenPutString(const char *_ch, int _attr, int _x, int _y);
void ScreenSetCursor(int _row, int _col);
void ScreenGetCursor(int *_row, int *_col);
void ScreenClear(void);
void ScreenUpdate(const void *_virtual_screen);
void ScreenUpdateLine(const void *_virtual_screen_line, int _row);
void ScreenRetrieve(void *_virtual_screen);
void ScreenVisualBell(void);







//#include <inlines/pc.h>
# 25 "../../../../../include/dos.h" 2

extern int _8087;

int _detect_80387(void);

struct DWORDREGS {
  unsigned int edi;
  unsigned int esi;
  unsigned int ebp;
  unsigned int cflag;
  unsigned int ebx;
  unsigned int edx;
  unsigned int ecx;
  unsigned int eax;
  unsigned short eflags;
};

struct DWORDREGS_W {
  unsigned int di;
  unsigned int si;
  unsigned int bp;
  unsigned int cflag;
  unsigned int bx;
  unsigned int dx;
  unsigned int cx;
  unsigned int ax;
  unsigned short flags;
};

struct WORDREGS {
  unsigned short di, _upper_di;
  unsigned short si, _upper_si;
  unsigned short bp, _upper_bp;
  unsigned short cflag, _upper_cflag;
  unsigned short bx, _upper_bx;
  unsigned short dx, _upper_dx;
  unsigned short cx, _upper_cx;
  unsigned short ax, _upper_ax;
  unsigned short flags;
};

struct BYTEREGS {
  unsigned short di, _upper_di;
  unsigned short si, _upper_si;
  unsigned short bp, _upper_bp;
  unsigned int cflag;
  unsigned char bl;
  unsigned char bh;
  unsigned short _upper_bx;
  unsigned char dl;
  unsigned char dh;
  unsigned short _upper_dx;
  unsigned char cl;
  unsigned char ch;
  unsigned short _upper_cx;
  unsigned char al;
  unsigned char ah;
  unsigned short _upper_ax;
  unsigned short flags;
};

union REGS { /* Compatible with DPMI structure, except cflag */
  struct DWORDREGS d;






  struct DWORDREGS_W x;


  struct WORDREGS w;
  struct BYTEREGS h;
};

struct SREGS {
  unsigned short es;
  unsigned short ds;
  unsigned short fs;
  unsigned short gs;
  unsigned short cs;
  unsigned short ss;
};

struct ftime {
  unsigned ft_tsec:5; /* 0-29, double to get real seconds */
  unsigned ft_min:6; /* 0-59 */
  unsigned ft_hour:5; /* 0-23 */
  unsigned ft_day:5; /* 1-31 */
  unsigned ft_month:4; /* 1-12 */
  unsigned ft_year:7; /* since 1980 */
};

struct date {
  short da_year;
  char da_day;
  char da_mon;
};

struct time {
  unsigned char ti_min;
  unsigned char ti_hour;
  unsigned char ti_hund;
  unsigned char ti_sec;
};

struct dfree {
  unsigned df_avail;
  unsigned df_total;
  unsigned df_bsec;
  unsigned df_sclus;
};





extern unsigned short _osmajor, _osminor;
extern unsigned short _os_trueversion;
extern const char * _os_flavor;
extern int _doserrno;

unsigned short _get_dos_version(int);

int _get_fat_size(const int _drive);
int _get_fs_type(const int _drive, char *const _result_str);
int _is_cdrom_drive(const int _drive);
int _is_fat32(const int _drive);
int _is_ram_drive(const int _drive);
int _media_type(const int _drive);

int int86(int _ivec, union REGS *in, union REGS *out);
int int86x(int _ivec, union REGS *_in, union REGS *_out, struct SREGS *_seg);
int intdos(union REGS *_in, union REGS *_out);
int intdosx(union REGS *_in, union REGS *_out, struct SREGS *_seg);
int bdos(int _func, unsigned _dx, unsigned _al);
int bdosptr(int _func, void *_dx, unsigned _al);

int _int86(int ivec, union REGS *in, union REGS *out);





int enable(void);
int disable(void);

int getftime(int _handle, struct ftime *_ftimep);
int setftime(int _handle, struct ftime *_ftimep);

int getcbrk(void);
int setcbrk(int _new_value);

void getdate(struct date *);
void gettime(struct time *);
void setdate(struct date *);
void settime(struct time *);

void getdfree(unsigned char _drive, struct dfree *_ptr);

void delay(unsigned _msec);


/*
 *  For compatibility with other DOS C compilers.
 */
# 204 "../../../../../include/dos.h"
struct _dosdate_t {
  unsigned char day; /* 1-31 */
  unsigned char month; /* 1-12 */
  unsigned short year; /* 1980-2099 */
  unsigned char dayofweek; /* 0-6, 0=Sunday */
};


struct _dostime_t {
  unsigned char hour; /* 0-23 */
  unsigned char minute; /* 0-59 */
  unsigned char second; /* 0-59 */
  unsigned char hsecond; /* 0-99 */
};


struct _find_t {
  char reserved[21];
  unsigned char attrib;
  unsigned short wr_time;
  unsigned short wr_date;
  unsigned int size;
  char name[256];
} __attribute__((packed));

/* make sure that structure packing is correct */
typedef int _DJCHK_FIND_T[(sizeof(struct _find_t)==286)*3 - 1];

struct _diskfree_t {
  unsigned short total_clusters;
  unsigned short avail_clusters;
  unsigned short sectors_per_cluster;
  unsigned short bytes_per_sector;
};


struct _DOSERROR {
  int exterror;
  char _class;
  char action;
  char locus;
};


struct _DOSERROR_STR {
  char *exterror_str;
  char *class_str;
  char *action_str;
  char *locus_str;
};


unsigned int _dos_creat(const char *_filename, unsigned int _attr, int *_handle);
unsigned int _dos_creatnew(const char *_filename, unsigned int _attr, int *_handle);
unsigned int _dos_open(const char *_filename, unsigned int _mode, int *_handle);
unsigned int _dos_write(int _handle, const void *_buffer, unsigned int _count, unsigned int *_result);
unsigned int _dos_read(int _handle, void *_buffer, unsigned int _count, unsigned int *_result);
unsigned int _dos_close(int _handle);
unsigned int _dos_commit(int _handle);

unsigned int _dos_findfirst(const char *_name, unsigned int _attr, struct _find_t *_result);
unsigned int _dos_findnext(struct _find_t *_result);

void _dos_getdate(struct _dosdate_t *_date);
unsigned int _dos_setdate(struct _dosdate_t *_date);
void _dos_gettime(struct _dostime_t *_time);
unsigned int _dos_settime(struct _dostime_t *_time);

unsigned int _dos_getftime(int _handle, unsigned int *_p_date, unsigned int *_p_time);
unsigned int _dos_setftime(int _handle, unsigned int _date, unsigned int _time);
unsigned int _dos_getfileattr(const char *_filename, unsigned int *_p_attr);
unsigned int _dos_setfileattr(const char *_filename, unsigned int _attr);

void _dos_getdrive(unsigned int *_p_drive);
void _dos_setdrive(unsigned int _drive, unsigned int *_p_drives);
unsigned int _dos_getdiskfree(unsigned int _drive, struct _diskfree_t *_diskspace);

int _dosexterr(struct _DOSERROR *_p_error);

int _dostrerr(struct _DOSERROR *_p_error, struct _DOSERROR_STR *_p_str);
# 119 "lstat.c" 2
# 1 "../../../../../include/dir.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 10 "../../../../../include/dir.h" 2
# 26 "../../../../../include/dir.h"
/* ffblk is also enhanced for LFNs; the dos 21 byte reserved area is used to
   hold the extra information.  Fields marked LFN are only valid if the magic
   is set to LFN32 */

struct ffblk {
  char lfn_magic[6]; /* LFN */
  short lfn_handle; /* LFN */
  unsigned short lfn_ctime; /* LFN */
  unsigned short lfn_cdate; /* LFN */
  unsigned short lfn_atime; /* LFN */
  unsigned short lfn_adate; /* LFN */
  char _ff_reserved[5];
  unsigned char ff_attrib;
  unsigned short ff_ftime;
  unsigned short ff_fdate;
  unsigned int ff_fsize;
  char ff_name[260];
} __attribute__((packed));

struct ffblklfn {
  unsigned int fd_attrib;
  unsigned long long fd_ctime;
  unsigned long long fd_atime;
  unsigned long long fd_mtime;
  unsigned int fd_sizehi;
  unsigned int fd_size;
  unsigned long long fd_reserved;
  char fd_longname[260];
  char fd_name[14];
} __attribute__((packed));

/* make sure that structure packing is correct */
typedef int _DJCHK_FFBLK0[(sizeof(struct ffblk) ==290)*3 - 1];
typedef int _DJCHK_FFBLK1[(sizeof(struct ffblklfn)==318)*3 - 1];
# 68 "../../../../../include/dir.h"
/* for fnmerge/fnsplit */
# 81 "../../../../../include/dir.h"
int __file_tree_walk(const char *_dir, int (*_fn)(const char *_path, const struct ffblk *_ff));
int findfirst(const char *_pathname, struct ffblk *_ffblk, int _attrib);
int findnext(struct ffblk *_ffblk);

int findclose(int handle);
void fnmerge (char *_path, const char *_drive, const char *_dir, const char *_name, const char *_ext);
int fnsplit (const char *_path, char *_drive, char *_dir, char *_name, char *_ext);
int getdisk(void);
char * searchpath(const char *_program);
int setdisk(int _drive);
# 120 "lstat.c" 2
# 1 "../../../../../include/sys/fsext.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../../include/sys/fsext.h"
# 1 "../../../../../include/stdarg.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 18 "../../../../../include/stdarg.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 19 "../../../../../include/stdarg.h" 2
# 26 "../../../../../include/sys/fsext.h" 2

typedef enum {
  __FSEXT_nop,
  __FSEXT_open,
  __FSEXT_creat,
  __FSEXT_read,
  __FSEXT_write,
  __FSEXT_ready,
  __FSEXT_close,
  __FSEXT_fcntl,
  __FSEXT_ioctl,
  __FSEXT_lseek,
  __FSEXT_link,
  __FSEXT_unlink,
  __FSEXT_dup,
  __FSEXT_dup2,
  __FSEXT_fstat,
  __FSEXT_stat,
  __FSEXT_llseek,
  __FSEXT_readlink,
  __FSEXT_symlink,
  __FSEXT_fchown,
  __FSEXT_chmod,
  __FSEXT_chown,
  __FSEXT_fchmod
} __FSEXT_Fnumber;

/* _ready gets passed a fd and should return a mask of these,
   as if we were emulating "int ready(int fd)" */




/* The return value is nonzero if the function has overridden the
   caller's functionality. */
typedef int (__FSEXT_Function)(__FSEXT_Fnumber _function_number,
          int *_rv, va_list _args);

int __FSEXT_alloc_fd(__FSEXT_Function *_function);
int __FSEXT_set_function(int _fd, __FSEXT_Function *_function);
__FSEXT_Function *__FSEXT_get_function(int _fd);
void *__FSEXT_set_data(int _fd, void *_data);
void *__FSEXT_get_data(int _fd);

int __FSEXT_add_open_handler(__FSEXT_Function *_function);
int __FSEXT_call_open_handlers(__FSEXT_Fnumber _function_number,
          int *rv, va_list _args);
# 121 "lstat.c" 2
# 1 "../../../../../include/libc/fsexthlp.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */



# 1 "../../../../../include/stdarg.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 8 "../../../../../include/libc/fsexthlp.h" 2
# 1 "../../../../../include/sys/fsext.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 9 "../../../../../include/libc/fsexthlp.h" 2
# 25 "../../../../../include/libc/fsexthlp.h"
/* This is a wrapper for an FSEXT function that allows it to be called
 * with a variable number of arguments. */

static inline int
__FSEXT_func_wrapper (__FSEXT_Function *func,
        __FSEXT_Fnumber fnum,
        int *rv,
        ...)
{
  va_list args;
  int ret;

  __builtin_va_start(args,rv);
  ret = func(fnum, rv, args);
  __builtin_va_end(args);

  return(ret);
}

/* This is a wrapper for __FSEXT_call_open_handlers that allows it
 * to be called with a variable number of arguments. */

static inline int
__FSEXT_call_open_handlers_wrapper (__FSEXT_Fnumber fnum, int *rv, ...)
{
  va_list args;
  int ret;

  __builtin_va_start(args,rv);
  ret = __FSEXT_call_open_handlers(fnum, rv, args);
  __builtin_va_end(args);

  return(ret);
}
# 122 "lstat.c" 2
# 1 "../../../../../include/dpmi.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 123 "lstat.c" 2
# 1 "../../../../../include/go32.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../../include/go32.h"
# 1 "../../../../../include/sys/version.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../../include/go32.h" 2
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 26 "../../../../../include/go32.h" 2
# 1 "../../../../../include/libc/asmobj.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

# 1 "../../../../../include/libc/djthunks.h" 1
/*
 *  dj64 - 64bit djgpp-compatible tool-chain
 *  Copyright (C) 2021-2024  @stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# 20 "../../../../../include/libc/asmobj.h" 2
# 27 "../../../../../include/go32.h" 2
# 1 "../../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 28 "../../../../../include/go32.h" 2






typedef struct {
  unsigned int size_of_this_structure_in_bytes;
  unsigned int linear_address_of_primary_screen;
  unsigned int linear_address_of_secondary_screen;
  unsigned int linear_address_of_transfer_buffer;
  unsigned int size_of_transfer_buffer; /* >= 4k */
  unsigned int pid;
  unsigned char master_interrupt_controller_base;
  unsigned char slave_interrupt_controller_base;
  unsigned short selector_for_linear_memory;
  unsigned int linear_address_of_stub_info_structure;
  unsigned int linear_address_of_original_psp;
  unsigned short run_mode;
  unsigned short run_mode_info;
} __Go32_Info_Block;

extern __Go32_Info_Block *___go32_info_block(void);/*
*/
# 50 "../../../../../include/go32.h"
       #define _go32_info_block (*___go32_info_block())/*
*/
# 50 "../../../../../include/go32.h"
                                                 ;







# 1 "../../../../../include/sys/movedata.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 59 "../../../../../include/go32.h" 2
# 1 "../../../../../include/sys/segments.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "../../../../../include/sys/segments.h"
int _my_cs(void);
int _my_ds(void);
int _my_ss(void);
# 60 "../../../../../include/go32.h" 2
# 70 "../../../../../include/go32.h"
/* returns number of times hit since last call. (zero first time) */
unsigned _go32_was_ctrl_break_hit(void);
void _go32_want_ctrl_break(int yes); /* auto-yes if call above function */
# 124 "lstat.c" 2
# 1 "../../../../../include/libc/farptrgs.h" 1
/* special version for libc - uses %gs instead of %fs.  Ignore comments */

/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
/* Copyright (c) 1995 DJ Delorie.  Permission granted to use for any
   purpose, provided this copyright remains attached and unmodified.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

ͻ
		Far Pointer Simulation Functions			
ͼ

This file attempts to make up for the lack of a "far" keyword in GCC.
Although it doesn't provide access to far call APIs (like Windows), it
does allow you to do far pointer data access without the overhead of
movedata() or dosmemget/dosmemput().

You should *always* include this file when using these functions and
compile with optimization enabled.  They don't exist as normal functions
in any library, and they compile down to only a few opcodes when used
this way.  They are almost as fast as native pointer operations, and
about as fast as far pointers can get.

If you don't use optimization, this file becomes prototypes for
farptr.c, which generates real functions for these when not optimizing.
When optimizing, farptr.c compiles to nothing.

There are two types of functions here - standalone and invariant.  The
standalone functions take a selector and offset.  These are used when
you need only a few accesses, time isn't critical, or you don't know
what's in the %gs register.  The invariant ones don't take a selector,
they only take an offset.  These are used inside loops and in
time-critical accesses where the selector doesn't change.  To specify
the selector, use the farsetsel() function.  That selector is used for
all farns*() functions until changed.  You can use _fargetsel() if you
want to temporary change the selector with _farsetsel() and restore
it afterwards.

The farpoke* and farpeek* take selectors.

The farnspoke* and farnspeek* don't (note the `ns' for `no selector').

Warning: These routines all use the %gs register for their accesses.
GCC normally uses only %ds and %es, and libc functions (movedata,
dosmemget, dosmemput) use %gs.  Still, you should be careful about
assumptions concerning whether or not the value you put in %gs will be
preserved across calls to other functions.  If you guess wrong, your
program will crash.  Better safe than sorry.

*/
# 74 "../../../../../include/libc/farptrgs.h"
# 1 "../../../../../include/sys/cdefs.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 75 "../../../../../include/libc/farptrgs.h" 2

void _farpokeb_gs(unsigned short, unsigned int, unsigned char);
void _farpokew_gs(unsigned short, unsigned int, unsigned short);
void _farpokel_gs(unsigned short, unsigned int, unsigned int);
unsigned char _farpeekb_gs(unsigned short, unsigned int);
unsigned short _farpeekw_gs(unsigned short, unsigned int);
unsigned int _farpeekl_gs(unsigned short, unsigned int);
void _farsetsel_gs(unsigned short);
unsigned short _fargetsel_gs(void);
void _farnspokeb_gs(unsigned int, unsigned char);
void _farnspokew_gs(unsigned int, unsigned short);
void _farnspokel_gs(unsigned int, unsigned int);
unsigned char _farnspeekb_gs(unsigned int);
unsigned short _farnspeekw_gs(unsigned int);
unsigned int _farnspeekl_gs(unsigned int);
# 125 "lstat.c" 2
# 1 "../../../../../include/libc/bss.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 23 "../../../../../include/libc/bss.h"
/* this gets *incremented* at startup, so that if a program dumps
   itself (emacs) you still have a way of telling.  Compare this value
   with a static, and if it's different, you've been restarted.
   First time through on a non-unexec'd program, it's greater than 0 */

extern int __bss_count;
# 126 "lstat.c" 2
# 1 "../../../../../include/libc/dosio.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 25 "../../../../../include/libc/dosio.h"
/* set to O_BINARY or O_TEXT */
extern char *__file_handle_modes;
extern void __file_handle_set(int fd, int mode);

int __doserr_to_errno(int doserr);

/* puts "path" in the transfer buffer, fixing
   unix-allowed multi-slashes */
int _put_path(const char *_path);
int _put_path2(const char *_path, int _offset);

/* Convenience functions for setting up transfers */




/* Functions used by write and _write.  */
int _write_fill_seek_gap(int fd);
int _write_int(int fd, const char *buffer, unsigned long long count);
# 127 "lstat.c" 2
# 1 "../../../../../include/libc/symlink.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */

/* Written by Laurynas Biveinis                                */
/* This file contains some internal info related to symlinks   */
/* Note: symlink file format is still in internal include file */
/* because I don't think it's required for user apps           */
# 28 "../../../../../include/libc/symlink.h"
# 1 "../../../../../include/sys/djtypes.h" 1
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1994 DJ Delorie, see COPYING.DJ for details */
# 29 "../../../../../include/libc/symlink.h" 2






/* Symlink files have fixed length - 510 bytes. Why this value? Why not? */
/* It is big enough to hold longest possible path                        */


/* Semi-internal library function which reads symlink contents given  */
/* either a file name or its handle.  Used by readlink(), fstat() and */
/* user supplied (if any) file fstat handler.                         */

int __internal_readlink(const char * __path, int __fhandle, char * __buf,
                        size_t __max);

/* A prototype for internal library function for fully resolving symlink   */
/* chain. Standard function readlink() solves only one symlink level.      */
/* If path name passed appears to be not a symlink, it is copied to result */
/* string. Return code 1 means success, 0 - failure (to many links - errno */
/* is set too).                                                            */

int __solve_symlinks(const char * __symlink_path, char * __real_path);

/* Internal library function for fully resolving symlinks in `__symlink_path' */
/* except the last its component. Otherwise it is similar to __solve_symlinks */
int __solve_dir_symlinks(const char * __symlink_path, char * __real_path);
# 128 "lstat.c" 2

# 1 "xstat.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
/*
 * Header for internal stat()/fstat()/lstat() assist functions.
 *
 */




# 1 "../../../../../include/stdio.h" 1
/* Copyright (C) 2018 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2017 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2014 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 16 "xstat.h" 2
# 1 "../../../../../include/time.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2013 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2011 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 17 "xstat.h" 2
# 1 "../../../../../include/sys/types.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 18 "xstat.h" 2
# 1 "../../../../../include/sys/stat.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2002 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 19 "xstat.h" 2
# 1 "../../../../../include/dpmi.h" 1
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 20 "xstat.h" 2

/* Some errno.h headers do not define EFAULT.  Here the value is
   taken from POSIX IEEE 1003.1.  */
# 1 "../../../../../include/errno.h" 1
/* Copyright (C) 2015 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2012 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2003 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 2000 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1998 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
# 24 "xstat.h" 2




/* Under MS-DOS, file access permissions are shared by all, except for
   Write permission. */




/* Macro to convert a segment and an offset to a "far offset" suitable
   for _farxxx() functions of DJGPP.  */






extern unsigned short _osmajor, _osminor;
extern const char * _os_flavor;

extern time_t _file_time_stamp(unsigned int);
extern ino_t _invent_inode(const char *, unsigned, unsigned long);
extern unsigned short _get_magic(const char *, int);
extern unsigned short _get_dos_version(int);
extern char * _truename(const char *, char *);
extern int _is_remote_drive(int);
extern int _is_executable(const char *, int, const char *);
extern short _get_dev_info(int);
extern long __filelength(int);
extern int _is_remote_handle(int);
extern void _djstat_describe_lossage(FILE *);
extern int _getftime(int, unsigned int *);
extern blksize_t _get_cached_blksize (const char *path);
# 130 "lstat.c" 2

int __getdisk(void);
int __findfirst(const char *, struct ffblk *, int);
int __findnext(struct ffblk *);



/* Should we bother about executables at all? */


/* The structure of the full directory entry.  This is the 32-byte
   record present for each file/subdirectory in a DOS directory.
   Although the ``packed'' attribute seems to be unnecessary, I use
   it to be sure it will still work for future versions of GCC.  */

struct full_dirent {
  char fname[8];
  char fext[3];
  unsigned char fattr;
  unsigned char freserved[10];
  unsigned short ftime;
  unsigned short fdate;
  unsigned short fcluster;
  unsigned int fsize;
}
__attribute__ ((packed));


/* Static variables to speed up SDA DOS Swappable Data Area access on
   subsequent calls.  */

/* The count of number of SDA's we have.  It is more than 1 for DOS
   4.x only.  If it has a value of 0, the function init_dirent_table()
   will be called to compute the addresses where we are to look for
   directory entry of our file.  A value of -1 means this method is
   unsupported for this version of DOS.  */
static int dirent_count;

/* The table of places to look for our directory entry.
   Each entry in the table is a linear offset from the beginning of
   conventional memory which points to a particular location within
   one of the SDA's, where the entry of a file being stat()'ed could
   appear.  The offsets are computed once (when the routine is first
   called) and then reused for other calls.  The actual storage for
   the table is malloc()'ed when this function is first called.  */
static unsigned int * dirent_table;

/* When we have only one SDA, this is where its only place to look for
   directory entry is stored.  */
static unsigned int dirent_place;

/* This holds the fail bits from the last call to init_dirent_table(),
   so we can return them every time get_inode_from_sda() is called.  */
static unsigned short init_dirent_table_bits;

/* Holds the last seen value of __bss_count, to be safe for
   restarted programs (emacs).  */
static int stat_count = -1;

/*
 * Parts of the following code is derived from file DOSSWAP.C,
 * which came with ``Undocumented DOS'', 1st edition.
 */

/* Compute table of pointers to look for directory entry of a file.  */
static int
init_dirent_table (void)
{
  short get_sda_func;
  unsigned short dirent_offset;
  unsigned short true_dos_version;
  unsigned short dos_major, dos_minor;
  __dpmi_regs regs = {0};

  if (dirent_count == -1) /* we already tried and found we can't */
    return 0;

  /* Compute INT 21h function number and offset of directory entry
     from start of SDA.  These depend on the DOS version.
     We need exact knowledge about DOS internals, so we need the
     TRUE DOS version (not the simulated one by SETVER), if that's
     available.  */
  true_dos_version = _os_trueversion;
  dos_major = true_dos_version >> 8;
  dos_minor = true_dos_version & 0xff;

  if ((dos_major == 3) && (dos_minor >= 10))
    {
      get_sda_func = 0x5d06;
      dirent_offset = 0x1a7;
    }
  else if (dos_major == 4)
    {
      /* According to ``Undocumented DOS, 2nd edition'', I could have
         used 5d06 here, as for DOS 5 and above, but I like to be
         defensive.  In fact, the above book itself uses 5d0b, contrary
         to its own recommendation.  */
      get_sda_func = 0x5d0b;
      dirent_offset = 0x1b3;
    }
  else if (dos_major >= 5)
    {
      get_sda_func = 0x5d06;
      dirent_offset = 0x1b3;
    }
  else
    {
      _djstat_fail_bits |= 2 /* Unsupported DOS version */;
      dirent_count = -1;
      return 0;
    }

  _djstat_fail_bits &= ~2 /* Unsupported DOS version */; /* version is OK */

  /* Get the pointer to SDA by calling undocumented function 5dh of INT 21. */
  regs.x.ax = get_sda_func;
  __dpmi_int(0x21, &regs);
  if (regs.x.flags & 1)
    {
      _djstat_fail_bits |= 1 /* Get SDA call failed */;
      dirent_count = -1; /* if the call failed, never try this later */
      return 0;
    }

  _djstat_fail_bits &= ~1 /* Get SDA call failed */; /* Get SDA succeeded */

  /* DOS 4.x might have several SDA's, which means we might have more
     than one place to look into.  (It is typical of DOS 4 to complicate
     things.)
     Compute all the possible addresses where we will have to look.  */
  if (dos_major == 4)
    {
      /* The pointer returned by INT 21h, AX=5D0b points to a header
         which holds a number of SDA's and then an array of that number
         of records each one of which includes address of an SDA (DWORD)
         and its length and type (encoded in a WORD).
         While walking this list of SDA's, we add to each pointer the
         offset of directory entry and stash the resulting address in
         our table for later use.  */

      int sda_list_walker = ((int)((((unsigned long)(regs.x.ds)) << 4) + (unsigned short)(regs.x.si)));
      int i;
      int *tbl;

      dirent_count = _farpeekw_gs(_go32_info_block.selector_for_linear_memory, sda_list_walker); /* number of SDA's */

      /* Allocate storage for table.  */
      dirent_table = (unsigned *)malloc(dirent_count*sizeof(int));
      tbl = (int *) dirent_table;
      if (!dirent_table)
        {
          /* If malloc() failed, maybe later it will succeed, so don't
             store -1 in dirent_count.  */
          dirent_count = 0;
          _djstat_fail_bits |= 0x40 /* dirent_count ridiculously large */;
          return 0;
        }

      memset(dirent_table, 0, dirent_count*sizeof(int));
      _djstat_fail_bits &= ~0x40 /* dirent_count ridiculously large */; /* dirent_count seems OK */

      /* Walk the array of pointers, computing addresses of directory
         entries and stashing them in our table.  */
      _farsetsel_gs(_go32_info_block.selector_for_linear_memory);
      for (i = dirent_count, sda_list_walker += 2; i--; sda_list_walker += 6)
        {
          int sda_start = _farnspeekl_gs(sda_list_walker);
          unsigned short sda_len = _farnspeekw_gs(sda_list_walker + 4) & 0x7fff;

          /* Let's be defensive here: if this SDA is too short to have
             place for directory entry, we won't use it.  */
          if (sda_len > dirent_offset)
            *tbl++ = sda_start + dirent_offset;
          else
            dirent_count--;
        }
    }

  /* DOS 3.1 and 5.0 or later.  We have only one SDA pointed to by
     whatever INT 21h, AH=5d returns.  */
  else
    {
      dirent_count = 1;
      dirent_place = ((int)((((unsigned long)(regs.x.ds)) << 4) + (unsigned short)(regs.x.si))) + dirent_offset;
      dirent_table = &dirent_place;
    }

  return 1;
}

/* Get inode number by searching DOS Swappable Data Area.
   The entire directory entry for a file found by FindFirst/FindNext
   appears at a certain (version-dependent) offset in the SDA after
   one of those function is called.
   Should be called immediately after calling DOS FindFirst function,
   before the info is overwritten by somebody who calls it again.  */
static unsigned int
get_inode_from_sda(const char *mybasename)
{
  int count = dirent_count;
  unsigned int * dirent_p = dirent_table;
  unsigned short dos_mem_base = _go32_info_block.selector_for_linear_memory;
  char * dot = strchr(mybasename, '.');
  size_t total_len = strlen(mybasename);
  int name_len;
  int ext_len;
  int cluster_offset = ((size_t) &((struct full_dirent *)0)->fcluster);

  if( dot )
    {
      name_len = dot - mybasename;
      ext_len = total_len - name_len - 1;
    }
  else
    {
      name_len = total_len;
      ext_len = 0;
    }

  /* Restore failure bits set by last call to init_dirent_table(), so
     they will be reported as if it were called now.  */
  _djstat_fail_bits |= init_dirent_table_bits;

  /* Force reinitialization in restarted programs (emacs).  */
  if (stat_count != __bss_count)
    {
      stat_count = __bss_count;
      dirent_count = 0;
    }

  /* Initialize the table of SDA entries where we are to look for
     our file.  */
  if (!dirent_count && !init_dirent_table())
    {
      /* Don't save the truename failure bit.  */
      init_dirent_table_bits = (_djstat_fail_bits & ~8 /* _truename() failed */);
      return 0;
    }
  init_dirent_table_bits = (_djstat_fail_bits & ~8 /* _truename() failed */);
  if (dirent_count == -1)
    return 0;

  count = dirent_count;
  dirent_p = dirent_table;

  _farsetsel_gs(dos_mem_base);

  /* This is DOS 4.x lossage: this loop might execute many times.
     For other DOS versions it is executed exactly once.  */
  while (count--)
    {
      unsigned int src_address = *dirent_p;
      char cmp_buf[sizeof(struct full_dirent)];

      /* Copy the directory entry from the SDA to local storage.
         The filename is stored there in infamous DOS format: name and
         extension blank-padded to 8/3 characters, no dot between them.  */
      fmemcpy2(cmp_buf, (__dpmi_paddr){ .selector = dos_mem_base, .offset32 = src_address, },
               sizeof(struct full_dirent));

      /* If this is the filename we are looking for, return
         its starting cluster. */
      if (!strncmp(cmp_buf, mybasename, name_len) &&
          (ext_len == 0 || !strncmp(cmp_buf + 8, dot + 1, ext_len)))
        return (unsigned int)_farnspeekw_gs(*dirent_p + cluster_offset);

      /* This is not our file.  Search more, if more addresses left. */
      dirent_p++;
    }

  /* If not found, give up.  */
  _djstat_fail_bits |= 4 /* Bad pointer to SDA */;
  return 0;
}

static int _ioctl_get_first_cluster(const char *);

/* Get the number of the first cluster of PATHNAME using
   the IOCTL call Int 21h/AX=440Dh/CX=0871h, if that call
   is supported by the OS.  Return the cluster number, or
   a negative number if this service isn't supported.  */

static int
_ioctl_get_first_cluster(const char *pathname)
{
  __dpmi_regs r = {0};

  /* See if the IOCTL GetFirstCluster call is supported.  */
  r.x.ax = 0x4411; /* query generic IOCTL capability by drive */
  r.h.bl = pathname[0] & 0x1f; /* drive number (1=A:) */
  r.x.cx = 0x871;
  __dpmi_int(0x21, &r);
  if ((r.x.flags & 1) == 0 && r.x.ax == 0)
    {
      r.x.ax = 0x440d; /* Generic IOCTL */
      r.x.cx = 0x0871; /* category code 08h, minor code 71h */
      r.x.bx = 1; /* pathname uses current OEM character set */
      r.x.ds = _go32_info_block.linear_address_of_transfer_buffer >> 4;
      r.x.dx = _go32_info_block.linear_address_of_transfer_buffer & 0x0f;
      _put_path(pathname);
      __dpmi_int(0x21, &r);
      if ((r.x.flags & 1) == 0)
 return ( ((int)r.x.dx << 16) + r.x.ax );
    }
  return -1;
}

static char blanks_8[] = "        ";

static int
stat_assist(const char *path, struct stat *statbuf)
{
  struct ffblk ff_blk;
  char canon_path[260];
  char pathname[260];
  short drv_no;
  unsigned dos_ftime;

  _djstat_fail_bits = 0;

  memset(statbuf, 0, sizeof(struct stat));
  memset(&dos_ftime, 0, sizeof(dos_ftime));

  /* Fields which are constant under DOS.  */
  statbuf->st_uid = getuid();
  statbuf->st_gid = getgid();
  statbuf->st_nlink = 1;

  /* Make the path explicit.  This makes the rest of our job much
     easier by getting rid of some constructs which, if present,
     confuse `_truename' and/or `findfirst'.  In particular, it
     deletes trailing slashes, makes "d:" explicit, and allows us
     to make an illusion of having a ".." entry in root directories.  */
  _fixpath (path, pathname);

  /* Get the block size for the device associated with `pathname'. */

  statbuf->st_blksize = _get_cached_blksize(pathname);
  if (statbuf->st_blksize == -1)
    return -1; /* errno set by _get_cached_blksize() */


  /* Get the drive number.  It is always explicit, since we
     called `_fixpath' on the original pathname.  */
  drv_no = toupper((unsigned char)pathname[0]) - 'A';

  /* Produce canonical pathname, with all the defaults resolved and
     all redundant parts removed.  This calls undocumented DOS
     function 60h.  */
  if (_truename(path, canon_path) || _truename(pathname, canon_path))
    {
      /* Detect character device names which must be treated specially.
         We could simply call FindFirst and test the 6th bit, but some
         versions of DOS have trouble with this (see Ralph Brown's
         Interrupt List, ``214E'', under `Bugs').  Instead we use
         truename() which calls INT 21/AX=6000H.  For character devices
         it returns X:/DEVNAME, where ``X'' is the current drive letter
         (note the FORWARD slash!).  E.g., for CON or \dev\con it will
         return C:/CON.
         We will pretend that devices all reside on a special drive
         called `@', which corresponds to st_dev = -1.  This is because
         these devices have no files, and we must invent inode numbers
         for them; this scheme allows to lower a risk of clash between
         invented inode and one which belongs to a real file.  This is
         also compatible with what our fstat() does.
      */
    char_dev:
      if (canon_path[2] == '/')
        {
          char dev_name[9]; /* devices are at most 8 characters long */

          strncpy(dev_name, canon_path + 3, 8); /* the name without `X:/' */
          dev_name[8] = '\0';
          strcpy(canon_path, "@:\\dev\\");
          strcat(canon_path, dev_name);
          strncat(canon_path, blanks_8, 8 - strlen(dev_name)); /* blank-pad */
          canon_path[15] = '\0'; /* ensure zero-termination */

          /* Invent inode */
          statbuf->st_ino = _invent_inode(canon_path, 0, 0);

          /* Device code. */
          statbuf->st_dev = -1;




          /* Set mode bits, including character special bit.
             Should we treat printer devices as write-only?  */
          statbuf->st_mode |= (0x2000 | (00400 | 00040 | 00004) | 00200);

          /* We will arrange things so that devices have current time in
             the access-time and modified-time fields of struct stat, and
             zero (the beginning of times) in creation-time field.  This
             is consistent with what DOS FindFirst function returns for
             character device names (if it succeeds--see above).  */
          statbuf->st_atime = statbuf->st_mtime = time(0);
          statbuf->st_ctime = _file_time_stamp(dos_ftime);

          return 0;
        }
      else if (canon_path[0] >= 'A' && canon_path[0] <= 'z' &&
               canon_path[1] == ':' && canon_path[2] == '\\')
        {
          /* _truename() returned a name with a drive letter.  (This is
             always so for local drives, but some network redirectors
             also do this.)  We will take this to be the TRUE drive
             letter, because _truename() knows about SUBST and JOIN.
             If the canonicalized path returns in the UNC form (which
             means the drive is remote), it cannot be SUBSTed or JOINed,
             because SUBST.EXE and JOIN.EXE won't let you do it; so, for
             these cases, there is no problem in believing the drive
             number we've got from the original path (or is there?).  */
          drv_no = toupper((unsigned char)canon_path[0]) - 'A';
        }
    }
  else
    {
      /* _truename() failed.  (This really shouldn't happen, but who knows?)
         At least uppercase all letters, convert forward slashes to backward
         ones, and pray... */
      const char *src = pathname;
            char *dst = canon_path;

      while ( (*dst = (*src > 'a' && *src < 'z'
                       ? *src++ - ('a' - 'A')
                       : *src++)) != '\0')
        {
          if (*dst == '/')
            *dst = '\\';
          dst++;
        }

      _djstat_fail_bits |= 8 /* _truename() failed */;
    }

  /* Call DOS FindFirst function, which will bring us most of the info.  */
  if (!__findfirst(pathname, &ff_blk, (1|2|4|16|32)))
    {
      /* Time fields. */
      dos_ftime =
        ( (unsigned short)ff_blk.ff_fdate << 16 ) +
          (unsigned short)ff_blk.ff_ftime;

      /* If the IOCTL GetFirstCluster call is available, try it first.  */
      if ( (_djstat_flags & 1 /* should we bother getting inode numbers? */) == 0
    && (statbuf->st_ino = _ioctl_get_first_cluster(pathname)) <= 0)
        {

          /* For networked drives, don't believe the starting cluster
             that the network redirector feeds us; always invent inode.
             This is because network redirectors leave bogus values there,
             and we don't have enough info to decide if the starting
             cluster value is real or just a left-over from previous call.
             For local files, try first using DOS SDA to get the inode from
             the file's starting cluster number; if that fails, invent inode.
             Note that the if clause below tests for non-zero value returned
             by is_remote_drive(), which includes possible failure (-1).
             This is because findfirst() already succeeded for our pathname,
             and therefore the drive is a legal one; the only possibility that
             is_remote_drive() fails is that some network redirector takes
             over IOCTL functions in an incompatible way, which means the
             drive is remote.  QED.  */
          if (statbuf->st_ino == 0 /* don't try SDA if IOCTL call succeeded */
       || _is_remote_drive(drv_no)
              || (statbuf->st_ino = get_inode_from_sda(ff_blk.ff_name)) == 0)
            {
              _djstat_fail_bits |= 0x10 /* inode defaults to hashing */;
              statbuf->st_ino =
                _invent_inode(canon_path, dos_ftime, ff_blk.ff_fsize);
            }
   else if (toupper ((unsigned char)canon_path[0]) != toupper ((unsigned char)pathname[0])
     && canon_path[1] == ':'
     && canon_path[2] == '\\'
     && canon_path[3] == '\0')
     /* The starting cluster in SDA for the root of JOINed drive
	       actually belongs to the directory where that drive is
	       ``mounted''.  This can potentially be the cluster of
	       another file on the JOINed drive.  We cannot allow this.  */
     statbuf->st_ino = 1;
        }

      /* File size. */
      statbuf->st_size = ff_blk.ff_fsize;

      /* Mode bits. */
      statbuf->st_mode |= (00400 | 00040 | 00004);
      if ( !(ff_blk.ff_attrib & 0x07) ) /* no R, H or S bits set */
        statbuf->st_mode |= 00200;

      /* Windows 2000 and XP - device attributes not set correctly */
      if( _os_trueversion == 0x532 && !ff_blk.ff_fsize &&
          (_djstat_fail_bits & 8 /* _truename() failed */) &&
          !ff_blk.ff_ftime && (ff_blk.ff_fdate == 33) &&
          !ff_blk.lfn_ctime && (ff_blk.lfn_cdate == 33) &&
          !ff_blk.lfn_atime && (ff_blk.lfn_adate == 33) &&
          !strcmp(ff_blk.lfn_magic,"LFN32") &&
          /* strlen(ff_blk.ff_name) == 3 && */ ff_blk.ff_attrib == 0x20)
        ff_blk.ff_attrib = 0x40;

      /* Sometimes `_truename' doesn't return X:/FOO for character
	 devices.  However, FindFirst returns attribute 40h for them.  */
      if (ff_blk.ff_attrib == 0x40)
 {
   size_t cplen = strlen (canon_path);
   char *pslash = canon_path + cplen - 1;

   while (pslash > canon_path + 2 && *pslash != '\\')
     pslash--;

   /* Force it into X:/FOO form.  */
   if (canon_path[1] == ':')
     {
       if (pslash > canon_path + 2)
  memmove (canon_path + 2, pslash,
    cplen - (pslash - canon_path) + 1);
       canon_path[2] = '/';
       goto char_dev;
     }
 }

      /* Directories should have Execute bits set. */
      if (ff_blk.ff_attrib & 0x10)
        statbuf->st_mode |= (0x3000 | (00100 | 00010 | 00001));

      else
        {
          int bytes_read;
          int old_errno;
          char buf[2];

          /* Check if it is symlink.  */
          old_errno = __dj_errno;
          bytes_read = readlink(pathname, buf, 1);
          if (bytes_read != -1)
          {
              /* Yes, it is.  */
              statbuf->st_mode |= 0x8000;
              __dj_errno = old_errno;
          }
          else
          {
              /* This is a regular file. */
              char *extension = strrchr(ff_blk.ff_name, '.');

              __dj_errno = old_errno;

              /* Set regular file bit.  */
              statbuf->st_mode |= 0x0000;

              if ((_djstat_flags & (2 /* get execute bits from file extension? */ | 4 /* get execute bits from magic signature? */)) != (2 /* get execute bits from file extension? */ | 4 /* get execute bits from magic signature? */))
                {
                  /* Set execute bits based on file's extension and
                     first 2 bytes. */
                  if (extension)
                    extension++; /* get past the dot */
                  if (_is_executable(pathname, -1, extension))
                    statbuf->st_mode |= (00100 | 00010 | 00001);
                }
          }
        }
    }
  else if ((_djstat_fail_bits & 8 /* _truename() failed */))
    {
      /* If both `findfirst' and `_truename' failed, this must
	 be a non-existent file or an illegal/inaccessible drive.  */
      if (__dj_errno == 38)
 __dj_errno = 21;
      return -1;
    }
  else if (pathname[3] == '\0' && (pathname[2] == '/' || pathname[2] == '\\') )
    {
      /* Detect root directories.  These are special because, unlike
	 subdirectories, FindFirst fails for them.  We look at PATHNAME
	 because a network redirector could tweak what `_truename'
	 returns to be utterly unrecognizable as root directory.  PATHNAME
	 begins with "d:/" (unless relative, in which case "d:.").  */

      /* Mode bits. */
      statbuf->st_mode |= (0x3000|(00400 | 00040 | 00004)|00200|(00100 | 00010 | 00001));

      /* Root directory will have an inode = 1.  Valid cluster numbers
         for real files under DOS start with 2. */
      statbuf->st_ino = 1;

      /* Simulate zero size.  This is what FindFirst returns for every
         sub-directory.  Later we might compute a better approximation
         (see below).  */
      ff_blk.ff_fsize = 0L;

      /* The time fields are left to be zero, unless the user wants us
         to try harder.  In the latter case, we check if the root has
         a volume label entry, and use its time if it has. */

      if ( (_djstat_flags & 0x10 /* try to get root dir time stamp? */) == 0 )
 {
   char buf[7];
   int volume_found = 0;

   strcpy(buf, pathname);
   strcat(buf, "*.*");
   /* Floppies written by Windows 9X and NT include entries
	     that have volume label bit set, but are actually parts
	     of an LFN entry.  Non-LFN platforms might be fooled to
	     take them as volume labels, and report bogus time stamps.  */
   volume_found = __findfirst(buf, &ff_blk, 8) == 0;
   while (volume_found
   && (ff_blk.ff_attrib & (2|4)) != 0)
     volume_found = __findnext(&ff_blk) == 0;
   if (volume_found)
     dos_ftime = ( (unsigned)ff_blk.ff_fdate << 16 ) + ff_blk.ff_ftime;
   else
     _djstat_fail_bits |= 0x20 /* Root dir, but no volume label */;
 }
    }
  else
    {
      int e = __dj_errno; /* errno value from original FindFirst on PATHNAME */
      int i = 0;
      int j = strlen (pathname) - 1;

      /* Check for volume labels.  We did not mix FA_LABEL with
	 other attributes in the call to `__findfirst' above,
	 because some environments will return bogus info in
	 that case.  For instance, Win95 and WinNT seem to
	 ignore `pathname' and return the volume label even if it
	 doesn't fit the name in `pathname'.  This fools us to
	 think that a non-existent file exists and is a volume
	 label.  Hence we test the returned name to be PATHNAME.  */
      if (!__findfirst(pathname, &ff_blk, 8))
 {
   i = strlen (ff_blk.ff_name) - 1;

   if (j >= i)
     {
       for ( ; i >= 0 && j >= 0; i--, j--)
  if (toupper ((unsigned char)ff_blk.ff_name[i]) != toupper ((unsigned char)pathname[j]))
    break;
     }
 }

      if (i < 0 && pathname[j] == '/')
 {
   /* Indeed a label.  */
   statbuf->st_mode = (00400 | 00040 | 00004);

   statbuf->st_mode |= 0x5000;

   statbuf->st_ino = 1;
   statbuf->st_size = 0;
   dos_ftime = ( (unsigned)ff_blk.ff_fdate << 16 ) + ff_blk.ff_ftime;
 }
      else
 {
   /* FindFirst on volume labels might set errno to ENMFILE or even
	     to something more strange like EINVAl; correct that.  */
   __dj_errno = e; /* restore errno from the original FindFirst */
   if (__dj_errno == 38)
     __dj_errno = 22;
   return -1;
 }
    }

  /* Device code. */
  statbuf->st_dev = drv_no;




  /* Time fields. */
  statbuf->st_atime = statbuf->st_mtime = statbuf->st_ctime =
    _file_time_stamp(dos_ftime);

  if ( ! strcmp(ff_blk.lfn_magic,"LFN32") )
    {
      unsigned xtime;
      xtime = ff_blk.lfn_ctime | (((unsigned)ff_blk.lfn_cdate) << 16);
      if(xtime) /* May be zero if file written w/o lfn active */
        statbuf->st_ctime = _file_time_stamp(xtime);
      xtime = ff_blk.lfn_atime | (((unsigned)ff_blk.lfn_adate) << 16);
      if(xtime > dos_ftime) /* Accessed time is date only, no time */
        statbuf->st_atime = _file_time_stamp(xtime);
    }

  if ( (statbuf->st_mode & 0xf000) == 0x3000
       && (_djstat_flags & 8 /* compute directory size? */) == 0 )
    {
      /* Under DOS, directory entries for subdirectories have
         zero size.  Therefore, FindFirst brings us zero size
         when called on a directory.  (Some network redirectors
         might do a better job, thus below we also test for zero size
         actually being returned.)  If we have zero-size directory,
         we compute here the actual directory size by reading its
         entries, then multiply their number by 32 (the size of a
         directory entry under DOS).  This might lose in the case
         that many files were deleted from a once huge directory,
         because AFAIK, directories don't return unused clusters to
         the disk pool.  Still, it is a good approximation of the
         actual directory size.

         We also take this opportunity to return the number of links
         for directories as Unix programs expect it to be: the number
         of subdirectories, plus 2 (the directory itself and the ``.''
         entry).

         The (max) size of the root directory could also be taken from
         the disk BIOS Parameter Block (BPB) which can be obtained
         by calling IOCTL (INT 21/AH=44H), subfunction 0DH, minor
         function 60H.  But we will treat all directories the same,
         even at performance cost, because it's more robust for
         networked drives.  */

      size_t pathlen = strlen (pathname);
      char lastc = pathname[pathlen - 1];
      char *search_spec = (char *)__builtin_alloca (pathlen + 10); /* need only +5 */
      int nfiles = 0, nsubdirs = 0, done;
      size_t extra = 0;
      int add_extra = 0;

      strcpy(search_spec, pathname);
      if (lastc == '/')
        strcat(search_spec, "*.*");
      else
        strcat(search_spec, "/*.*");

      if (statbuf->st_size == 0 && _use_lfn(0) /* assume it's the same on ALL drives */)
 {
   /* VFAT filesystems use additional directory entries to
	     store the long filenames.  */
   char fstype[40];

   if ((_get_volume_info(pathname,0,0,fstype) & 0x4000)
       && strncmp(fstype, "FAT", 4) == 0)
     add_extra = 1;
 }

      /* Count files and subdirectories.  */
      for (done = __findfirst(search_spec, &ff_blk, (1|2|4|16|32));
    !done;
    done = __findnext(&ff_blk))
 {
   char *fname = ff_blk.ff_name;

   /* Don't count "." and ".." entries.  This will show empty
	     directories as size 0.  */
   if (! (fname[0] == '.'
   && (fname[1] == '\0'
       || (fname[1] == '.'
    && fname[2] == '\0'))))
     {
       nfiles++;
       if (ff_blk.ff_attrib & 0x10)
  nsubdirs++;
       /* For each 13 characters of the long filename, a
		 32-byte directory entry is used.  */
       if (add_extra && !_is_DOS83(fname))
  extra += (strlen(fname) + 12) / 13;
     }
 }

      statbuf->st_nlink = nsubdirs + 2;
      if (statbuf->st_size == 0)
 statbuf->st_size = (nfiles + extra) * sizeof(struct full_dirent);
    }

  return 0;
}

/* Main entry point.  This is library lstat() function.
 */

int
lstat(const char *path, struct stat *statbuf)
{
  int e = __dj_errno;
  int pathlen, ret;
  char real_path[260];

  if (!path || !statbuf)
    {
      __dj_errno = 11;
      return -1;
    }

  if ((pathlen = strlen (path)) >= 260)
    {
      __dj_errno = 19;
      return -1;
    }

  /* Handle symlinks */
  if (!__solve_dir_symlinks(path, real_path))
    return -1;

  /* Fail if PATH includes wildcard characters supported by FindFirst,
     or if it is empty.  */
  if (memchr(real_path, '*', pathlen) || memchr(real_path, '?', pathlen)
      || real_path[0] == '\0')
    {
      __dj_errno = 22; /* since no such filename is possible */
      return -1;
    }

  if (__FSEXT_call_open_handlers_wrapper(__FSEXT_stat, &ret,
      real_path, statbuf))
    return ret;

  if (stat_assist(real_path, statbuf) == -1)
    {
      return -1; /* errno set by stat_assist() */
    }
  else
    {
      __dj_errno = e;
      return 0;
    }
}
